#!/bin/bash 
# goa_functions.sh
# Functions used for GPS processing with GIPSY OASIS II 5.0
# Version: 2.5
# 
# This software is not part of the official GIPSY distribution and it is not
# supported by JPL. Contact the GIPSY group at JPL to obtain a GIPSY license.
#
# goa_functions.sh is provided 'AS IS' to be used together with a licensed
# GIPSY distribution.
# The author and Central Washington University are not responsible 
# for the use of this software or any  derivative work from this software
# outside of the Geodesy Laboratory and PANGA data analysis facility 
# at Central Washington University. 
#
# For any publication or derivative work please aknowledge:
#
# Marcelo Santillan (marcelo@geology.cwu.edu)
# Tim Melbourne     (tim@geology.cwu.edu)
# Geodesy Laboratory & PANGA Data Analysis Facility (www.panga.cwu.edu)
# Geology Department
# Central Washington University
# 400 E University Way.
# Ellensburg, WA 98926
#
# Driver scripts: ppp.sh, sm_amb.sh
#
# Main Functions are: 
# Point Positioning: estimation()
# Bias Fixing:       bias_fixing() 
#
# Notes:
# The estimation() strategy is based on Willy Bertiger's gd2p.pl,
# All the needed gd2p routines were translated from perl to sh by Marcelo Santillan. 
# You can think of estimation() as a simpler and more limited inside-out gd2p.pl. 
# While gd2p.pl's uses perl to launch sh to run external executables, 
# goa_functions.sh is designed  to run the same instance of sh (or bash) for
# all excutables. The use of perl is limited to a few cases were non integer
# math was needed or to avoid the use of the system's grep and awk programs.  
# 
# Limitations:
# estimation() only works for Ground GPS Receivers.
#
# The Bias Fixing strategy uses smapper for the bias fixing and 
# needs the output of estimation() to work although other simpler estimation
# strategies may work as well.
# bias_fixing() was written based on the smapper's userguide.txt,
# and it uses the userguide's examples for the ambigon.nml.
# Note that bias_fixing() may not be the most efficient bias fixing
# strategy. See also GOA's ambigon2 and ambigon_p1.
# 
# Dependencies:
# goa_functions.sh, ppp.sh and sm_amb.sh use non standard GOA scripts
# and will not work with an 'out-of-the-box' GOA install and without 
# the rest of PANGA's local toolbox.
# However that doesn't mean that cannot be adapted to be used with
# a standard GOA distribution.
# Variables:
#   All the scripts assume that the enviroment variable GPS_HOME_DIR exists
#   and points to the local toolbox directory (~/processing) and 
#   the PERL5LIB variable is set to include the directory (~/processing/perl_lib).
# Scripts (non-standard):
#      - get_proc_defaults : get default variables that specify input and
#                            output directories and station_info files set in
#                            the $GPS_HOME_DIR/etc/defaults, such as:
#                            ($STA_INFO_DIR, $WORK_DIR_HOME, $PPP_DIR,
#                            $AMB_DIR)
#      - gpsdate           : very flexible date and time conversion utility
#      - get_rnx           : rinex download program
#      - get_orbits.pl     : orbits download program
#      - get_brdc          : broadcasted orbits download program
#      - gps_sat_offsets   : get satellite phase center offsets.
#      - fix_rnx_header    : overrides rinex using local metadata
#      - qregres.pl        : perlized qregres, can be replaced by
#                            qregres -i qmfile \
#                                    -o rgfile \
#                                    -n qregres.nml \
#                                    -scr ./SCRATCH_REGRES \
#                                    -shad Shadow \
#                                    -time_dep TDPfile \
#                                    -sc peci \
#                                    -pr qregres.log 
#
#      - proc_summary     : gather statistics
#      - make_process_log : gather statistics (deprecated)
#
# 
#
# Author:
# Marcelo Santillan
# GPS Data Analyst 
# Geodesy Lab. and PANGA Data Analysis Facility.
# Central Washington University
# February 2009
# 
# History:
# Version 1.0: Jan 2007 (ppp_functions.sh).
#
# Copyright (c) 2009
# Central Washington University


#GLOBALS
# define executable's system call for the main GOA programs.
perl='/opt/perl/bin/perl'

preprefilter='preprefilter wash.nml rgfile prefilter.txt'

prefilter='prefilter prefilter.txt rgfile batch.txt phase.txt'

filter="filter batch.txt wash.nml '' rgfile accume.nio smooth.nio uinv.nio '' "

smapper="smapper wash.nml '' accume.nio smooth.nio uinv.nio smsol.nio smcov.nio '' "

smapper2="smapper wash.nml '' accume.nio smooth.nio uinv.nio smsol2.nio smcov2.nio '' "

smapper_wlpb="smapper amb_res.nml '' accume.nio smooth.nio uinv.nio smsol2.nio smcov2.nio '' "

postfit='postfit wash.nml rgfile smsol.nio uinv.nio postfit.nio point.txt'

postfit_wlpb="postfit wash.nml rgfile smsol2.nio uinv.nio postfit2.nio point.txt"

postbreak="postbreak -p postfit.nio -n postbreak.nml -b batch.txt"

edtpnt2="edtpnt2 point.txt wash.nml rgfile accume.nio smooth.nio uinv.nio"

edtpnt2_wlpb="edtpnt2 '' edtpnt2_ambig.nml aregres.nio accume.nio smooth.nio uinv.nio"

flag_qm="flag_qm -p point.txt -i qmfile_old -o qmfile "


RATE=300               # secs
PROC_TYPE='s'          # s: static ; k: kinematic
WEED_SAT=1             # iterate to remove bad sats
MIN_SATS=4             # 10
MIN_QM_EPOCHS=2000     # 
ELEV_MASK=15           # degrees
DEBUG=0                # debug mode
ECLIP_SATS=( )

# useful functions for general scripting
#####################################################################################
uc() {
  echo $@ | tr '[a-z]' '[A-Z]' 
}

lc() {
  echo $@ | tr '[A-Z]' '[a-z]' 
}

echoc() {
  printf "%s" "$@";
}

echo2()
{
  echo $@ 1>&2 
}

join_with()
{
$perl -e 'print join (shift,@ARGV),"\n"' $@
}

sgrep()
{
 $perl -MGOA::Util -e '$re=shift; $f=shift; print grep { /$re/ } slurp($f);' $1 $2
}

#####################################################################################


load_defaults()
{
  for v in $@; do
   v=$(uc $v)
   eval $v=$( get_proc_defaults $v )
  done
}

mk_dirs()
{
  local clean_out_dir=$1
  work_dir=$2
  out_dir=$3
  
  rm $work_dir  2> /dev/null
  if [  -d $work_dir ] ; then
    \rm -rf $work_dir
  fi

  if [[  -d $out_dir  &&  $clean_out_dir -ne 0 ]]  ; then
    \rm -rf $out_dir
  fi

  if [ ! -d $out_dir ]; then
    mkdir -p $out_dir 
  fi
  
  if [ ! -d $work_dir ]; then
    mkdir -p $work_dir 
  fi
}

yy2y()
{
$perl <<'EOT' '' $1
$yy=shift;
$yy+=0;
print $yy > 50 ? 1900+$yy : 2000 + $yy,"\n";
EOT
}

rnx2qm() {
  local site=$1
  local rnx=$2
  local rt="$GPS_HOME_DIR/etc/GPS_Receiver_Types.user"
  ninja -i $rnx \
        -t $RATE \
        -PC -LC \
        -F -a 20 -ion 30 \
        -ca2p_override \
        -q $site.qm \
        -n $(uc $site)  \
        -scr $SCRATCH_DIR  \
	-urt $rt

  sats=($(qm_sat $site.qm))
  nsat=${#sats[@]}
  if [ $nsat -lt $MIN_SATS ]; then
      echo "ERASING $site.qm ; $nsat < $MIN_SATS sats" 
      \rm $site.qm 2> /dev/null
  fi
  
}

ok_info()
{
  local site=$( uc $1) ; shift;
  local ok=1
  local dir=$STA_INFO_DIR
  local info=(pos id svec)
  if [ $# -ne 0 ]; then
    dir=$1; shift;
  fi
  if [ $# -ne 0 ]; then
    info=($@)
  fi
  for i in ${info[@]} ; do
    if [[  $( sgrep $site "$dir/sta_$i" |wc -l) -eq 0 ]]; then
       ok=0
       echo2 "No sta_$i entry for $site in $dir/sta_$i"
    fi
  done
  echo $ok
}


mk_qmfile_from_pp() {
   copy_pp_sol ${sites[@]}
   merge_qm ????/$date.????.qm qmfile_in
   weed_qm qmfile_in peci qmfile
}

combine_tdp_dry() {
    local s=''
    local i=''
    for i in ${sites[@]}
    do
	    s_lc=$(lc $i)
	    s_uc=$(uc $i)
	    if [[ -e $s_lc/$date.$s_lc.$s_uc.TDPdry ]]; then
		    cat $s_lc/$date.$s_lc.$s_uc.TDPdry >> TDPdry
	    fi
    done
}

make_tropo_table() {
	local tdp=$1
	local net=$2
	local out=$3
	local qr_nml=$4
	local wash_nml=$5
	local tdry=$6
	if [ -e $tdry ]
	then
	    tropo_table.py --tdp ${tdp} --net ${net} --out ${out} --qr_nml ${qr_nml} --wash_nml ${wash_nml} --dryz ${tdry}
        else
	    tropo_table.py --tdp ${tdp} --net ${net} --out ${out} --qr_nml ${qr_nml} --wash_nml ${wash_nml}
        fi
}

mk_qfront_from_pp() {
  local s=''
  local i=''
  for i in ${sites[@]} ; do
    s=$(lc $i)
    if [[ -e $s/$date.$s.qfront.log ]]; then
      ln -s $s/$date.$s.qfront.log qfront.$s.log
    fi
  done
}

mk_qmfile_from_rnx()
{
   local sats=()
   local nsat=0
   for site  in ${sites[@]} ; do
     ok_info=$(ok_info $site $STA_INFO_DIR pos id svec)
     ok_rnx=$(get_rnx -site $site -date $cal $rnx_local)
     rnx=$( ls $site$doy*.${yy}o 2> /dev/null )
     pp_qm="$ppp_dir/$site/$cal.$site.qm.Z"
     if [[ $ok_rnx -eq 1 && $ok_info -eq 1 ]]; then
          fix_rnx $(uc $site) $rnx $yyyy $doy $yy $ELEV_MASK
          rnx2qm $site $rnx 1> qfront.$site.log \
                            2> qfront.$site.error
          if [ ${#del_sat[@]} -ne 0 ]; then
              for i in ${del_sat[@]}; do
                del_qm_sat GPS$i $site.qm 
              done
          fi
     fi
   done
   merge_qm *.qm qmfile_in
   weed_qm qmfile_in peci qmfile
   sites=( $( qm_sta qmfile ) )
   rm -f qmfile_in
}

mk_stacov() 
{
  stacov -i smcov_fix.nio \
         -o $cal.fix.stacov \
         -q qregres.nml -tp tpeo.nml
}

del_qm_sat() {
  local gps=$1; shift;
  local qmfile='qmfile'
  if [ $# -ne 0 ]; then
    qmfile=$1
  fi
  del_qm -i $qmfile -o $qmfile.tmp -s $gps
  \mv $qmfile.tmp $qmfile

}

wlpb_nml() {

  cat <<EOT > wlpb.nml
   ! Default seting for ground station ( kinematic or static ) April 2009
 ! tested on static with -post_wind 2.5e-3, 2.5e-5
 \$SMINPUT
    OUTNAME = 'OASIS'
    READAMBIGONNML = .TRUE.
    UDCOPYAMBIGON = .FALSE.
 \$END

  \$AMBIGON

! General settings
    DEBUG = .FALSE.
    WIDELANE = 'CODE'
    READWLPB = .TRUE.  
    OVERLAP = 3600    ! Double differences must last this long
    AMBWEIGHT = 1.0D7 ! 1/sigma for applied "data point" from resolved amb
    MAXUPDATEDSIGMAGE = 100.0 ! Check against ridiculous resolution
    ALTERNATE_NL_CONF_CALC = .TRUE. ! Assume that incorrect widelanes off by
                                    ! only 1, and compute NL conf accordingly.
                                    ! The nice thing is, there is not nearly as
                                    ! large a penalty for being off by two as
                                    ! by one, anyway, and being off by three
                                    ! should be caught by the test for off by
                                    ! one, so what the heck.

! WL settings
    NINT_MAX_WL = 0.30    ! Widelane at least this close to integer
    WL_INFLATION = 0.10   ! Assumed background variance in double difference
    BACKGROUNDLEVEL = 0.5 ! Multiply by formal sigma to estimate level of
                          ! assumed near flat distribution of bad widelane
                          ! errors.
    CONFIDENCE_WL = 0.75 ! Cut-off for confidence in wide-lane

! NL settings
    GLOBAL_NL_SIGMA_FACTOR = 6.0D0 ! Sigma in cycles = GNSF/sqrt(DD Time (s))
    NINT_MAX_NL = 0.20             ! Narrow-lane at least this close to integer
    CONFIDENCE_NL = 0.999          ! Cut-off for confidence in narrow-lane

  \$END

  \$APRIORINML
  \$END
 
EOT
cat wlpb.nml datawght.nml > amb_res.nml
}

postbreak_nml() {
   local size=$1
   local auto=$2
   cat <<EOT > postbreak.nml  
 \$options
   dtyp = 120
   min_slip_size = $size
   max_auto_slip = $auto
   min_arc = 5
   delta = 5E-6
   auto = .true.
 \$end
EOT
 let pb_size++
 if [ ! -e postbreak_$pb_size.nml ]; then
   \cp postbreak.nml postbreak_$pb_size.nml
 fi
 
}

new_slips() {
 $perl -MGOA::Util -e '$n=0;$n=grep {/No new slips/} slurp(shift); printf "%d", ! $n ;' $1
}

wlpb_loop() {
  local iterations=2

  # EDTPNT2 namelist file
  cat <<EOT > edtpnt2_ambig.nml

  \$DATAWGHT
  \$end
  \$EDTINIT
      REJFLAGONLY = .FALSE.
      ALLPOINTS = .TRUE.
  \$END

EOT

  export AREGRES="aregres.nio"
  export REGRES="rgfile"
  export GLOBALWLPBLIST="wlpb"

  while [ $iterations -ne 3 ]; do
    \rm smcov2.nio smsol2.nio smsig2.nio ambigon.log aregres.nio tdp_final
    echo "Running smapper with wlpb iteration $iterations"
    $smapper_wlpb > amb_res.log
    $edtpnt2_wlpb
    let iterations++
  done

  \rm smcov2.nio smsol2.nio smsig2.nio TDPTABLE CORRELATION
  $smapper2 > smapper2.log

  $postfit_wlpb > postfit.log
}

postbreak_loop() {
  local new_slips=1
  local pc=0

  while [ $new_slips -ne 0 ]; do
     \mv qmfile qmfile_old 
     $postbreak -qi qmfile_old -qo qmfile -o batch.out.$pc > postbreak.$pc.log
     \cp postbreak.$pc.log postbreak.log
     new_slips=$( new_slips postbreak.log )
     \cp batch.out.$pc batch.txt
     let pc++
     if [ $new_slips -ne 0 ]; then
         \rm accume.nio smooth.nio uinv.nio smsol.nio smcov.nio \
             point.txt postfit.nio TDPTABLE APVALSOUT CORRELATION 
         $filter  > filter.log
         $smapper > smapper.log
         $postfit > postfit.log
         flag_qmfile
     fi
     if [ $pc -eq 10 ]; then
        echo "Postbreak Analysis Failed to Converge 10 iterations"
        break
     fi
  done
}


window() {
  $perl -e '$w=shift;@w=(200,20,5*$w,4*$w,2*$w,$w);print "@w";' $1
}

change_postfit_window() {
 local rhs="  WINDOW = $1,$2"
 local lhs='^\s\+WINDOW\s\+=.*$'
 local cmd="s/$lhs/$rhs/"
 sed -i "$cmd" wash.nml
}

edtpnt2_loop() {
  local ec_max=6
  local ec=0
  while [ -e point.txt ]; do
      echo "editing point.txt"
      cat point.txt >> point.all
      if [ $ec -eq $ec_max ]; then
         echo "edit point cycle failed to converge after $ec cycles,"
         echo "THERE ARE PROBABLY LARGE OUTLIERS LEFT"
         return 0
      fi
      $edtpnt2
      if [ $? ];then
        printf "edtpnt2 exit status: %d  RE-RUNNING FILTER\n" $?;
        \rm accume.nio smooth.nio uinv.nio 2> /dev/null
        $filter
      fi
      rm point.txt \
      smsol.nio smcov.nio postfit.nio TDPTABLE APVALSOUT CORRELATION 2> /dev/null
      $smapper > smapper.log
      $postfit > postfit.log
      flag_qmfile
      let ec++
  done

}

ambigon_nml()
{
  #ambigon_nml_global
  #ambigon_nml_region
  #ambigon_nml_test
  ambigon_nml_2sta
}

#For one two station run on a short, static baseline, 
ambigon_nml_2sta()
{
 cat <<EOT > ambigon.nml 
 \$AMBIGON
     CONFIDENCE_WL = 0.95
     CONFIDENCE_NL = 0.999
     NINT_MAX_WL = 0.1
     NINT_MAX_NL = 0.02
     WL_INFLATION = 0.06
     NL_INFLATION = 2.50
     BACKGROUNDLEVEL = 0.15
     OVERLAP = 3600
     CUTALL = 0.0D0
     WIDELANE = 'CODE'
 \$END
EOT
}

ambigon_nml_test()
{
 cat <<EOT > ambigon.nml 
 \$AMBIGON
     CONFIDENCE_WL = 0.90
     CONFIDENCE_NL = 0.99
     NINT_MAX_WL = 0.1
     NINT_MAX_NL = 0.02
     WL_INFLATION = 0.06
     NL_INFLATION = 2.50
     BACKGROUNDLEVEL = 0.15
     OVERLAP = 3600
     CUTALL = 0.0D0
     WIDELANE = 'CODE'
 \$END
EOT
}



#Global network
ambigon_nml_global()
{
 cat <<EOT > ambigon.nml 
 \$AMBIGON
     CONFIDENCE_WL = 0.95
     CONFIDENCE_NL = 0.999
     NINT_MAX_WL = 0.1
     NINT_MAX_NL = 0.02
     WL_INFLATION = 0.06
     NL_INFLATION = 2.50
     BACKGROUNDLEVEL = 0.15
     OVERLAP = 3600
     CUTALL = 0.0D0
     WIDELANE = 'CODE'
 \$END
EOT
}

#Tighter Global
ambigon_nml_region()
{
 cat <<EOT > ambigon.nml 
 \$AMBIGON
     CONFIDENCE_WL = 0.96465
     CONFIDENCE_NL = 0.99929
     NINT_MAX_WL = 0.07071
     NINT_MAX_NL = 0.01414
     WL_INFLATION = 0.06
     NL_INFLATION = 2.5
     BACKGROUNDLEVEL = 0.15
     OVERLAP = 3600
     CUTALL = 0.0D0
     WIDELANE = 'CODE'
 \$END
EOT
}


sdelt() #sec2deltaT from gd2p
{

$perl <<'EOT' '' $1 
  my $sec=shift;
  my ($dddd,$hh,$mm,$ss,$ffff);
  $dddd = int $sec/86400;
  if ( $dddd > 9999 ) {
    warn "Input to sdelt too large: $sec\n";
  }
  $hh   = int ( ($sec - 86400*$dddd)/3600 );
  $mm   = int( ($sec -  86400*$dddd - 3600*$hh)/60 );
  $ss   = int( ($sec -  86400*$dddd - 3600*$hh - $mm*60) );
  $ffff = ($sec -  86400*$dddd - 3600*$hh - $mm*60 - $ss);
  $ffff *= 1E4;

  printf "%4.4d/%2.2d:%2.2d:%2.2d\.%4.4d\n", $dddd,$hh, $mm, $ss,$ffff;

EOT

}


prep_k_nml()
{
local secs=0
local sdelt=$(sdelt $RATE)
local spsig='1.0E-3'
local smtau='RANDOMWALK'   # 'WHITENOISE'
    # $smtau=$(sdelt $secs)
local sta=$(uc $site)

cat  <<EOT 

  SSTCH =    'STA X   $sta', 'STA Y   $sta', 'STA Z   $sta'
  SDELT = 3*'$sdelt'
  SPSIG = 3*$spsig
  SMTAU = 3*'$smtau'

EOT

}

prep_nml()
{
local type=$1
local rate=$RATE
local tropdrift='5e-8' # 1.7e-7 km/sqr(sec) 
local wetztrop='.TRUE.'
local wetzgrad='.TRUE.'
local graddrift='5.0D-9'
local prep_k=''

if [[ $type == 'k' ]]; then
      #wetztrop='.FALSE.' 
      prep_k=$( prep_k_nml )
fi

cat <<EOT 

  \$PREP                       !namelist input for preprefilter
  
  FORCE_TIME_LESS_PT1 = .TRUE.
  REFCLOCK = 'ALLGPS' 
  YADDSG    = 1E6,
  CLOCKPSIGMA = 3.0D6,
  WETZTROP  = $wetztrop
  TROPDRIFT = $tropdrift
  WETZGRAD  = $wetzgrad
  GRADDRIFT = $graddrift
  
  $prep_k

 \$END
           
EOT

}

init_nml()
{
local type=$1; shift
local amb=''
local grounddelete=''

if [ $# -eq 1 ]; then
  amb='ambigon'
fi

if [[ $type == 'k' ]]; then
  grounddelete="GROUNDDELETE =  '*'  ! don't solve for station position"
fi

cat<<EOT 

  \$INIT                       !namelist input for filter
    STATEDELETE = 'GPS*' 
    IDIGIT    = 9
    YPARAM = 'DUMMY',   ! Needs at least 1 or defaults to all; everything after 
    SOLPRT    = .TRUE.
    SOLVE     = .TRUE.
    DEBUG     = 1
    OUTNAME   = 'OASIS'
    DBGROUP = 'XMT'

    $grounddelete

  \$END
  
EOT

}

apriori_nml()
{
 local type=$1
 local apsig='2.0D3'
 if [[ $type == 'k' ]]; then
    apsig=$2
 fi

 cat<<EOT

 \$APRIORI                    !namelist input for filter 
  APALL     = .TRUE.
  APNAMS(1)=      'PHAS*',            APSIGS(1)= 1E6 ! km should be same YADDSG
  APNAMS(2)=      'STAX*',            APSIGS(2)= $apsig
  APNAMS(3)=      'STAY*',            APSIGS(3)= $apsig
  APNAMS(4)=      'STAZ*',            APSIGS(4)= $apsig
  APNAMS(5) =     'STABIAS*',         APSIGS(5)= 3.0D2
  APNAMS(6)=      'WETZTROP*',        APSIGS(6)= 5.0D-4   !half meter
  APNAMS(7) =      'TRPAZ*',          APSIGS(7) = 5.0D-4 !half meter
  APNAMS(8) =      'DUMMY',           APSIGS(8) = 1.0D0
 \$END

EOT

}

base_nml()
{
   local amb=$1
   local apsig='2.0E3'
   local type=$PROC_TYPE  
   local prep=$( prep_nml $type )
   local init=$( init_nml $type $amb )
   local apriori=$( apriori_nml $type $apsig )

cat <<EOT  > base.nml

   $prep

   $init

   $apriori

EOT

}


datawght_sites_nml()
{
  
  local PHASE_N=120
  local RANGE_N=110

  local drec=$( join_with "','" $@ ) 

  cat <<EOT
  
 \$DATAWGHT                 !namelist input for filter, edtpnt2, and postfit
    DATSIG    = 1.0D-5, 1.0D-3, 
    DATYPE    = $PHASE_N, $RANGE_N,   
    DATREC(1,1) = '$drec'
    DATREC(1,2) = '$drec'
    ELMINSTA  = $ELEV_MASK
 \$END

EOT

}

datawght_no_sites_nml()
{

  cat <<EOT 
    \$DATAWGHT                 !namelist input for filter, edtpnt2, and postfit

    \$END
EOT

}

datawght_nml()
{
   local datawght=$(datawght_no_sites_nml)
   if [ $# -ne 0 ]; then
     datawght=$(datawght_sites_nml $@)
   fi
   cat<<EOT > datawght.nml
   $datawght
EOT
}

edtinit_nml()
{
  local allpoints=''
  if [[ $1 == 'allpoints' ]]; then
     allpoints='     ALLPOINTS = .TRUE. '
  fi
  cat <<EOT > edtinit.nml

 \$EDTINIT                !namelist input for edtpnt2
  DBGROUP = 'XMT'
  DEBUG  = .TRUE.
  REJFLAGONLY = .FALSE.
  $allpoints
 \$END

EOT
 
}

rdsmsol_nml()
{
  local sta=$(uc $site)
  cat<<EOT > rdsmsol.nml
    \$INPUT
        NAMES   = 'STA X   $sta', 'STA Y   $sta', 'STA Z   $sta'
	    OUTNAME = 'OASIS'
    \$END
EOT
}

sminput_nml()
{
 
  local read_ambigon='';
  if [[ $1 == 'ambigon' ]]; then
    read_ambigon='     READAMBIGONNML = .TRUE.'
  fi
  local mapstr=$(gpsdate -i $date char)
  local mapend=$(gpsdate -i ${date}:23:59:45.00 char)
  local delta=$(sdelt $RATE)
  cat <<EOT > sminput.nml

 \$SMINPUT                !namelist input for smapper
  $read_ambigon
  SAVESIGMA = .TRUE.      ! true to get sig's on tdp
  SAVEUD =  'LAST'
  OUTNAME = 'OASIS'
  TDPORDER = 3          ! interpolation order applied to an input 
                        ! nominal tdp file
  MAPSTR  = '$mapstr'
  MAPEND  = '$mapend'
  MAPUD   = .TRUE.

  TDPTYP  = 'CONSTANT'
  DELTA   = '$delta'
  WRTCOR  = 'STAX*', 'STAY*', 'STAZ*'
  WRTTDP  = 'PHASE*','STAX*', 'STAY*', 'STAZ*', 'STABIAS*', 'WETZTROP*','SATBIAS*',
             'SAT1BIAS*', 'SAT2BIAS*', 'SAT3BIAS*', 'SAT4BIAS*',  'TRPAZ*'
  TDPFULLPRECISION = .TRUE. 
! READAPDELTANML = .TRUE.
 \$END

 \$APDELTANML
 \$END


 \$APRIORINML  !namelist input for smapper not used if apriori in filter
  APNAMS(1)=      'STAX*',            APSIGS(1)= 1.0D-5   !1 cm
  APNAMS(2)=      'STAY*',            APSIGS(2)= 1.0D-5   !1 cm
  APNAMS(3)=      'STAZ*',            APSIGS(3)= 1.0D-5   !1 cm
 \$END

EOT
 
}

wash_amb_nml()
{
  base_nml  
  sminput_nml ambigon
  datawght_nml  $@         # ${sites[@]}
  edtinit_nml 
  ambigon_nml
  cp wash.nml wash.nml.tmp
  wash_nml 
  \mv wash.nml wash_amb.nml
}

edt_amb_nml()
{
  edtinit_nml allpoints
  datawght_nml
  wash_nml 
  \mv wash.nml edt2fix.nml 
  cp wash.nml.tmp wash.nml

}

init_wash_nml()
{
  base_nml
  sminput_nml
  datawght_nml $@  # ${sites[@]}
  edtinit_nml
   
}
wash_nml()
{
  local range_win=5.0D-3
  local phase_win=50.0D-5
  local PHASE_N=120
  local RANGE_N=110

  if [ $# -eq 2 ]; then
      range_win=$1
      phase_win=$2
  fi
  if [[ ! -e base.nml ]]; then
    init_wash_nml ${sites[@]}
  fi
  
  local drec=$( join_with "','" ${sites[@]})  
  
  cat <<EOT > limits.nml

  \$LIMITS                 !namelist input for postfit
    DBGROUP = 'XMT'
    WINDOW = $range_win, $phase_win 
    DATYPE    =    $RANGE_N,    $PHASE_N
    DATREC(1,1) = '$drec'
    DATREC(1,2) = '$drec'
  \$END

EOT
 
 cat  base.nml \
      datawght.nml \
      sminput.nml \
      limits.nml \
      edtinit.nml \
      tpeo.nml  > wash.nml
  
  if [[ -e ambigon.nml ]]; then
   cat ambigon.nml >> wash.nml   
  fi
  
 
}

rnx2start_end()
{
$perl <<'EOT' '' $1
$rnx=shift;
($start,$end)= grep { !/^trans / } grep { /date . time/ } `teqc -plot +meta $rnx`;
chomp $start;
chomp $end;
$start =join(':', (split(' ',$start))[4..5]);
$end =join(':', (split(' ',$end))[4..5]);
print "$start $end\n";
EOT
}

fix_rnx()
{
  local site=$1
  local rnx=$2
  local yyyy=$3
  local doy=$4
  local yy=$5
  local elev_mask=$6
  local smooth=''
  if [[ $PROC_TYPE == 'k' ]]; then
      teqc_flags=( ${teqc_flags[@]} '+smooth' )
  fi 
  get_brdc -date $yyyy,$doy 


  # Remove mis-used "RCV CLOCK OFFS APPL" RINEX header tags
  grep -q "RCV CLOCK OFFS APPL" $rnx
  if [ $? -eq 0 ] 
  then
	 value=$(grep "RCV CLOCK OFFS APPL" $rnx | awk '{print $1}')
	 if [ $value -eq 1 ]
	 then
	 	echo2 "Removing 'RCV CLOCK OFFS APPL' tags"
	 	sed '/RCV CLOCK OFFS APPL/d' $rnx > $SCRATCH_DIR/$rnx.tmp
	 	\mv $SCRATCH_DIR/$rnx.tmp $rnx
	fi
  fi
  teqc_flags=( $( fix_rnx_header.dev -rnx $rnx -date $yyyy,$doy -site $(uc $site) ) ${teqc_flags[@]} )
  
  
  teqc +relax \
       -phc \
       -nav brdc${doy}0.${yy}n  \
       -R  \
       -O.dec 30s \
       -O.obs L1L2C1P2 ${teqc_flags[*]} \
       -set_mask $elev_mask $rnx > $SCRATCH_DIR/$rnx.tmp  2> /dev/null
  echo "teqc flags: ${teqc_flags[@]}" >> estimation.log  
  #teqc -phc -nav brdc${doy}0.${yy}n  -G32 -R -set_mask $ELEV_MASK $rnx > $rnx.tmp  2> /dev/null
  #   -R -O.mo $SITE -set_mask $ELEV_MASK $rnx > $rnx.tmp 2> /dev/null
  \mv $SCRATCH_DIR/$rnx.tmp $rnx
  #cat $rnx.tmp |sed 's/\s\+4 1\s\+$//;s/other.*COMMENT\s*$//' > $rnx

}

flag_qmfile() {
  if [ -e point.txt ]; then
     cat point.txt >> point.all
     \mv qmfile qmfile_old
     $flag_qm
  fi
}

#almost verbatim from gd2p.pm
qregres_errors()
{
$perl <<'EOT' '' $1
use Scalar::Util qw(looks_like_number);
my $qrlog = shift;
if ( ! open(FILE,'<',$qrlog) ) {
   print "1\n";
   exit;
}
while( <FILE> ) {
   if ( /TROUBLE TROUBLE/ ) {
        print "1\n"; 
        exit;
   }
   if ( /FORTRAN I.0 ERROR/ ) {
        print "1\n"; 
        exit;
   }
   if ( /S:AL:ALL     :ALL/ ) {
      @sum=split(':');
      if ( $sum[4] !~ /^\*+$/ ) {
         if ( $sum[4] == 0 ) {
           print "1\n";
           exit;
         }
         else {
            for my $n ( @sum[4..9] ) {
               if ( ! looks_like_number($n) ) {
                  print "1\n";
                  exit;
               }
            }
         }
      }
   }
}
print "0\n";
EOT
}

weed_qmfile() {
  local i=0; 
  if [ $# -ne 0 ]; then
     echoc "# Deleting Satellites: " >> del_sats.log
     for i in $@; do
         echoc "GPS$i " >> del_sats.log
         del_qm_sat GPS$i
     done
     echo "" >> del_sats.log
  fi

 mv qmfile qmfile_in
 weed_qm qmfile_in peci qmfile 1> weed_qm.log 2> weed_qm.err

}

qm_regres() {
  echo "Tropo apriori = ${tropo0[@]}"
  echo "VMF1 = ${vmf1[@]}"
  echo "IONEX = ${ionex[@]}"
  qregres.pl -qm  qmfile -date $date ${vmf1[@]} ${tropo0[@]} ${ionex[@]} 1>> estimation.log 2>> qr_err.log
  if [ $(qregres_errors qregres.log) -eq 1 ]; then
     echo2 "ERROR: qregres FAILED. check qregres.log , qr_err.log\n";
     exit 1;
  fi
  cat qregres.log >> qregres_log
}

wc_files() {
$perl <<'EOT' '' $1
 $file = shift;
 $n= grep { /$file/ } glob("*");
 print $n;
EOT
}

pre_estimation() {

  wash_nml 

  \rm -f  prefilter.txt \
          batch.txt \
          phase.txt \
          accume.nio \
          smooth.nio \
          uinv.nio \
          smsol.nio \
          smcov.nio \
          point.txt \
          postfit.nio \
          APVALSOUT \
	      CORRELATION \
	      TDPTABLE 
        
  $preprefilter 
  $prefilter    
  $filter      

  export EPHFIL="$GOA_EPH/$ARCH/de405s.nio"
  export TDPTABLE='TDPTABLE'

  $smapper > smapper.log
  
  local i=$(wc_files pre_smapper)
  let i++
  
  \cp smapper.log pre_smapper.$i.log
  
  \rm -f postfit.nio 2> /dev/null

  wash_nml 20E-3 20E-3 

  $postfit  > postfit.log
  
  i=$(wc_files pre_postfit)
  let i++
  
  \cp postfit.nio postfit.$i.nio 2> /dev/null
  
  \cp postfit.log pre_postfit.$i.log
 
}

estimation_wlpb() {
 
  pre_estimation 
 
  del_sat=( )
  if [ $WEED_SAT -ne 0 ]; then
    del_sat=($(bad_sats postfit.log $date ))
  fi
  
  if [ ${#del_sat} -ne 0 ]; then
    weed_qmfile ${del_sat[@]}

    sites=($(qm_sta qmfile) )
    sats=($(qm_sat qmfile))
  
    if [ ${#sats[@]} -lt $MIN_SATS ];then
      echo2 "ERROR: qmfile has < $MIN_SATS sats, bye"
      exit 1
    fi
  
    qm_regres
  
    pre_estimation
  fi    
  
  flag_qmfile
  
  \rm batch.out* postbreak.log* point.txt 2> /dev/null

  
  touch point.all 
  touch point.log
  #             min_slip_size  max_auto_slip 
  postbreak_nml 20E-3          200E-3

  postbreak_loop 

  # prepare postfit cycle

  range_window=( $(window 10E-3) )
  phase_window=( $(window 200E-5) )
  i=0
  while [ $i -lt ${#range_window[@]} ]
  do 
  
     win_r=${range_window[$i]}
     win_p=${phase_window[$i]}
     \rm postfit.nio point.txt 2> /dev/null
     wash_nml $win_r $win_p
     $postfit > postfit.log
     flag_qmfile
     if [ -e point.txt ]; then
       edtpnt2_loop
     fi
     let i++
  done

  postbreak_nml 20E-3 20E-3

  \rm batch.out.* postbreak.log* point.txt 2> /dev/null

  postbreak_loop

  sats=($(postfit_sat))
  if [ ${#sats[@]} -lt $MIN_SATS ];then
      echo2 "ERROR: after postfit solution has < $MIN_SATS sats, bye"
      exit 1
  fi
  
  # Resolve ambiguities with wide-lane phase bias file
  wlpb_nml
  wlpb_loop

  \cp TDPTABLE tdp_final

  \cp CORRELATION tdp_corr_final

  \cp smcov2.nio smcov.nio
 
  if [ $PROC_TYPE == 'k' ]; then
    rdsmsol_nml $site
    rdsmsol rdsmsol.nml smsol.nio  xyz_smsol.txt
  elif [ $PROC_TYPE == 's' ]; then
    stacov -i smcov.nio \
           -o estimation.stacov \
           -q qregres.nml -tp tpeo.nml
  fi

}

copy_pp_sol() {
  local s=''
  local dir="$PPP_DIR/$ac/$yyyy/$date"
  echo "Incorrect copy_pp_sol() called"
  for s in $@ ; do
     if [ -d "$dir/$s" ]; then
       if [ $(ls $dir/$s|wc -l) -ne 0 ]; then
         mkdir -p $work_dir/$s
         \cp $dir/$s/* $work_dir/$s
         gzip -df $work_dir/$s/*.Z 
       fi
     fi
  done
}

estimation() {
 
  pre_estimation 
 
  del_sat=( )
  if [ $WEED_SAT -ne 0 ]; then
    del_sat=($(bad_sats postfit.log $date ))
  fi
  
  if [ ${#del_sat} -ne 0 ]; then
    weed_qmfile ${del_sat[@]}

    sites=($(qm_sta qmfile) )
    sats=($(qm_sat qmfile))
  
    if [ ${#sats[@]} -lt $MIN_SATS ];then
      echo2 "ERROR: qmfile has < $MIN_SATS sats, bye"
      exit 1
    fi
  
    qm_regres
  
    pre_estimation
  fi    
  
  flag_qmfile
  
  \rm batch.out* postbreak.log* point.txt 2> /dev/null
  
  touch point.all 
  touch point.log
  #             min_slip_size  max_auto_slip 
  postbreak_nml 20E-3          200E-3

  postbreak_loop 

  # prepare postfit cycle

  range_window=( $(window 10E-3) )
  phase_window=( $(window 200E-5) )
  i=0
  while [ $i -lt ${#range_window[@]} ]
  do 
  
     win_r=${range_window[$i]}
     win_p=${phase_window[$i]}
     \rm postfit.nio point.txt 2> /dev/null
     wash_nml $win_r $win_p
     $postfit > postfit.log
     flag_qmfile
     if [ -e point.txt ]; then
       edtpnt2_loop
     fi
     let i++
  done

  postbreak_nml 20E-3 20E-3

  \rm batch.out.* postbreak.log* point.txt 2> /dev/null

  postbreak_loop

  sats=($(postfit_sat))
  if [ ${#sats[@]} -lt $MIN_SATS ];then
      echo2 "ERROR: after postfit solution has < $MIN_SATS sats, bye"
      exit 1
  fi
  
  \cp TDPTABLE tdp_final

  \cp CORRELATION tdp_corr_final
 
  if [ $PROC_TYPE == 'k' ]; then
    rdsmsol_nml $site
    rdsmsol rdsmsol.nml smsol.nio  xyz_smsol.txt
  elif [ $PROC_TYPE == 's' ]; then
    stacov -i smcov.nio \
           -o estimation.stacov \
           -q qregres.nml -tp tpeo.nml
  fi

}

copy_pp_sol() {
  local s=''
  local dir="$PPP_DIR/$ac/$yyyy/$date"
  for s in $@ ; do
     if [ -d "$dir/$s" ]; then
       # See if directory has files in it
       if [ "$(ls -A $dir/$s)" ]; then
         mkdir -p $work_dir/$s
         \cp $dir/$s/* $work_dir/$s
         gzip -df $work_dir/$s/*.Z 
       fi
     fi
  done
}

bias_fixing()
{
  
  wash_amb_nml ${sites[@]}
  
  export REAL=rgfile
  export REGRES=rgfile
  export AREGRES=aregres.nio
  export AMBIGONLOG=ambigon.log

  \rm point.txt \
      smsol.nio \
      smcov.nio \
      TDPTABLE \
      CORRELATION \
      APVALSOUT 2> /dev/null
  smapper wash_amb.nml '' accume.nio \
          smooth.nio uinv.nio \
          smsol.nio smcov.nio '' 2> bias_fixing.error
  
  edt_amb_nml
  
  edtpnt2 '' edt2fix.nml aregres.nio accume.nio \
          smooth.nio uinv.nio 2>> bias_fixing.error  
 
  \rm TDPTABLE CORRELATION 2> /dev/null
  smapper wash.nml '' accume.nio \
          smooth.nio uinv.nio \
          smsol_fix.nio smcov_fix.nio \
          smsig_fix.nio 1> smap_fixed.log 2>> bias_fixing.error
 
   stacov -i smcov_fix.nio \
          -o fix.stacov \
          -q qregres.nml -tp tpeo.nml
  
}

save_solution()
{
  local work_dir=$1
  local out_dir=$2 
  if [ $DEBUG -eq 0 ]; then
       \cp $work_dir/$cal.* $out_dir #2> /dev/null
       if [ $? -ne 0 ]; then
           echo "FAILED: cp $work_dir/$cal.* $out_dir"
       fi
  fi
}

stacov4fix() {
  \rm smcov.nio estimation.stacov 2> /dev/null
  smap4fix accume.nio smooth.nio \
           SMCOV.nio 1> smap4fix.log 2> smap4fix.err
  stacov -i SMCOV.nio -o estimation.stacov \
         -tp tpeo.nml -q qregres.nml 1> stacov.log
 # stacov -i smcov.nio -o station.stacov \
 #        -tp tpeo.nml -q qregres.nml > stacov.log
}


file_size()
{
 $perl -e 'print -s shift,"\n";' $1
}

ok_stacov()
{
  local f=$1
  local ok=1
  local np=
  if [ ! -e $f ]; then
    ok=0
  fi
  if [ $(file_size $f) -lt 362 ]; then
    echo2 "$f is too small" 
    ok=0
  fi
  if [ $ok -eq 1 ]; then
    np=$(head -1 $f)
    np=${np:1:4};
    if [ $np -eq 0 ]; then
      echo2 "$f has 0 parameters"
      ok=0
    fi
  fi
  echo $ok
}

stacov_name()
{
  local stacov_file=$1
  local cal=$2
  local s=$3
  local ac=$4
  local orb_type=$5
  local frame=$6
  
  if [[ $ac == 'jpl' || $ac == 'jgp' ]]; then
       if [[ $orb_type == 'final' ]]; then
          \mv $stacov_file $cal.$s.raw.stacov
          if [[ -e XFILE ]] ; then
              apply -x XFILE -i $cal.$s.raw.stacov \
                    -o $cal.$s.$frame.stacov \
                    -r -t -s 1> apply.log 2> apply.error
          fi
       elif [[ $orb_type == 'rapid' ]]; then
           \mv $stacov_file $cal.$s.$frame.r.stacov
       elif [[ $orb_type == 'final_fid' ]] ; then
          \mv $stacov_file $cal.$s.$frame.stacov
          ln -sf $cal.$s.$frame.stacov $cal.$s.raw.stacov
       else
         echo "Invalid orbit type $orb_type"
       fi
       
  elif [[ $ac == "igs" || $ac == 'ig1' ]]; then
      if [[ $orb_type == "rapid" ]]; then
          \mv $stacov_file $cal.$s.$frame.r.stacov
      elif [[ $orb_type == "final" ]]; then
         \mv $stacov_file $cal.$s.$frame.stacov
         ln -sf $cal.$s.$frame.stacov $cal.$s.raw.stacov
      fi
  
  fi
}

pack_solution()
{
   local cal=$1; shift; 
   local name=$1; shift;
   for f in $@; 
   do
      if [ -e $f ]; then
         compress -c $f > $cal.$name.$f.Z    
      fi
   done
}


clean_dirs() {
  local work_dir=$1
  local out_dir=$2
  local files=()
  local nfiles=1 
  if [ -d $out_dir ];then
    files=($(ls $out_dir))
    nfiles=${#files[@]}
  fi
  if [[  -d $work_dir && $DEBUG -eq 0 ]] ; then
     \rm -rf $work_dir
  fi
  if [[  -d $out_dir && $nfiles -eq 0 ]]  ; then
    \rm -rf $out_dir
  fi

}

leave() {
  local result=$1
  local archive_sol=$2
  local work_dir=$3 
  local out_dir=$4
  
  if [ $archive_sol -eq 1 ]; then
    if [[ $result == 'SUCCESS' ]]; then
      save_solution $work_dir $out_dir
      clean_dirs $work_dir $out_dir
    fi
  fi
  
  if [[ $result == 'FAILED' ]]; then
      clean_dirs $work_dir $out_dir
  fi
 
}

show_result() {
  local pa=$1
  local result=$2
  local s=$3
  local date=$4
  
  echo "RESULT: $result $(uc $pa): $(uc $site)  $date"

}

