#!/bin/bash 

. $GPS_HOME_DIR/bin/goa_functions.sh

progname=$( basename $0 )


usage() {
  more <<EOT
   
  $progname -site site -date date -orb [rapid|final]  -ac [igs|jgp|code] [options]

  options: (required)
  -site          : 4 char site name. (required unless using -rnx rnxfile)
  -date          : proc. date. (examples: yyyy-mm-dd, yyyymmdd, yyyy,doy, wwwwd )
                   Use 'gpsdate -help' to see other formats.
                   (required unless using -rnx rnxfile)
  -ac            : agency providing the orbits (Analysis Center)
                   jgp : JPL GPS Products
                   igs : IGS
                   code: CODE (Center for Orb. Determination Europe)
 
                   Note that jpl ( JPL legacy products ) are not longer 
                   mantained.  Use jgp instead.
  -orb           : orbit type 
                   rapid: 
                   final: by default the combination -ac jgp -orb final
                          provides non-fiducial orbits. fiducial orbits
                          are provided using: -ac jgp -orb final_fid
                     

  options:   (not required)
   -out_dir dir  :  save solution files in 'dir'
   -debug        :  don't erase processing files and don't archive.
                    processing files are kept in dir yyyy-mm-dd.site.ac 
                    under the current dir.
   -proc_type p  :  processing type p='s' static, p='k' kinematic ( stochastic station) 
   -k            :  process site in kinematic mode
   -rate secs    :  processing rate (default 300)  
   -local        :  only search for rinex in local archive
   -rnx rnxfile  :  use rnxfile as input. 
                    Use rnxfile name to get 'site' and 'date'
   -del_sat nn   : delete satellite[s] giving their GPSnn number.
                   For multiple sats use integers separated by commas: nn1,nn2,... 
   -all_sat      : don't iterate to remove bad satellites
   -mask         : elevation mask ( default is 15 degrees )
   -dev          : run developing version                  
   -wlpb         : Use wide-lane phase biases to solve for single-station 
                   ambiguities
   -help         : This help.
   -H            : This help.

  See also:   sm_amb.sh

  Author:
  Marcelo Santillan
  GPS Data Analyst 
  Geodesy Lab. and PANGA Data Analysis Facility.
  Central Washington University

  Copyright (c) 2009
  Central Washington University

EOT

exit 1

}
site=
sites=()
date=
cal=
orb_type=
ac=
SITE=
proj='dummy'
clean=0
rnx_local=''
rnx=''
args=$@
here=$( pwd )
frame="igs08"
del_sat=( ) 
#del_sat=( 62 63 )  # delete Block IIF GPS62 GPS63
ecl_sat=( )
out_dir=''
save4fix=0
archive_sol=1
result='FAILED'
start_epoch=''
end_epoch=''
delta_epoch=''
orb_opt=( )
vmf1=()
tropo0=()
ionex=()
ion_2nd=0
wlpb=0

if [ $# -eq 0 ] ;then   usage; fi

cmd_line=`echo "$progname $@"`

while [ $# -gt 0 ]
do
     case "$1" in
         -site) site=$( lc $2 )
                shift;;
         -date) date=$(gpsdate -i $2 cal )
                cal=$date 
                shift;;
         -rnx) rnx=$2
               shift;;
         -orb) orb_type=$(lc $2)
               shift;;
         -ac)  ac=$2
               shift;;
         -proj) proj=$2
                shift;;
         -rate) RATE=$2  
                shift;;
         -start_epoch) start_epoch=$2
                       shift;;
         -end_epoch) end_epoch=$2
                     shift;;
         -delta_epoch) delta_epoch=$2
                       shift;;
         -proc_type) PROC_TYPE=$2
                     shift;;
	      -k)    PROC_TYPE='k'
		         ;;
         -mask) ELEV_MASK=$2
                shift;;
         -del_sat) del_sat=($( echo $2|tr ',' ' '))
                shift;;
         -local) rnx_local='-local'
                 ;;
         -frame) frame=$2
                 shift;;
         -debug) DEBUG=1
                 ;;
         -all_sat) WEED_SAT=0
                 ;;
         -clear_orb) orb_opt=(${orb_opt[@]} '-purge')
                 ;;
	 -noftp) orb_opt=(${orb_opt[@]} '-noftp')
	         ;;
         -t1) orb_opt=(${orb_opt[@]} '-t1' $2 )
              shift;;
         -t2) orb_opt=(${orb_opt[@]} '-t2' $2 )
              shift;;
	 -tropo0) tropo0=('-tropo0' $2)
	      shift;;
         -vmf1) vmf1=('-vmf1')
	 	echo "Using VMF1"
                ;;
	 -ion_2nd) ion_2nd=1
	        ;;
	 -wlpb) wlpb=1
	        ;;
         -clean) clean=1
                 ;;
         -out_dir) out_dir=$2
                   shift;;
         -nofix) save4fix=0
                 ;;
         -no_archive) archive_sol=0
                 ;;
	     -dev)  .  $GPS_HOME_DIR/bin/goa_functions_dev.sh
		         ;;
         -help|-H) usage
                   ;;
     esac
     shift
done

load_defaults sta_info_dir ppp_dir work_dir_home scratch_dir ionex_dir

if [[ ! -d $SCRATCH_DIR ]]; then
  SCRATCH_DIR="/tmp"
fi
if [[ $rnx != ''  ]];then
   dname=$(dirname $rnx)
   name=$(basename $rnx)
   
   if [ $dname == '.' ];then
      rnx="$here/$name"
   fi
   site=${name:0:4}
   doy=${name:4:3}
   yy=${name:9:2}
   yyyy=$(yy2y $yy)
   #date="$yyyy,$doy"
fi

if [[ $date != '' ]]; then
  ydoy=( $(gpsdate -i $date doy) )
  doy=${ydoy[1]}
  yyyy=${ydoy[0]}
  yy=${yyyy:2:2}
  gwd=$(gpsdate -i $date gwd)
  cal=$(gpsdate -i $date cal)
fi

if [[ $ion_2nd == 1 ]]; then
	ionex=('-ionex_dir' "$IONEX_DIR")
fi

if [[  $site == '' || $date == '' ||  $ac == '' || $orb_type == '' ]];then
    echo "Insufficient arguments !"
    usage    
fi

 wd="$WORK_DIR_HOME/pp_work/$proj"
 od="$PPP_DIR/$ac/$yyyy/$cal/$site"
 if [ $DEBUG -eq 1 ]; then
    wd=$here
    clean=0
 fi
 
 wd="$wd/$cal.$site.$ac"

 hostname=`hostname` 
 echo "Processing: site: $(uc $site) ; date: $date $doy $gwd on $hostname"
 result="FAILED"
 mk_dirs $clean $wd $od
 cd $work_dir
  #echo "cd $work_dir"  > junk.log
  #echo "here = `pwd`" >> junk.log
  #export >> junk.log
 
 echo $cmd_line > estimation.log
 
 if [[ $ac == 'jgp' || $ac == 'igs' || $ac == 'code'  || $ac == 'ig1' ]]; then
   #ok_orb=$(get_orbits.pl -ac $ac -type $orb_type -date $date ${orb_opt[@]} -frame $frame ) 
   ok_orb=$(get_orbits.pl -ac $ac -type $orb_type -date $date ${orb_opt[@]} ) 
   ok_orb=$?
   if [ -e pos ]; then
     ln -sf pos peci
   fi
 else
   echo2 "Invalid agency $ac"
   leave $result $archive_sol $work_dir $out_dir
   show_result "PPP" "$result |Invalid agency $ac|" $(uc $site) "$cal , $doy , $gwd"
   usage
 fi
 
 
 if [ $ok_orb -eq 1 ]; then
    #echo "No orbits, bye..."
    leave $result $archive_sol $work_dir $out_dir
    show_result "PPP" "$result |No orbits, Bye ...|" $(uc $site) "$cal , $doy , $gwd"
    exit 1
 fi
 
 frame=$(cat frame)
 frame=$(lc $frame)


 if [[ $rnx == '' ]]; then  
   echo "Getting RINEX Files"
   ok_rnx=$( get_rnx -site $site -date $date $rnx_local -sr $RATE -save )
 fi
 if [[ -e $rnx ]] ; then
   echo "RINEX Files already exist"
   cp $rnx $work_dir
   ok_rnx=1
 fi
 
 if [ $ok_rnx -ne 1 ]; then
  #echo "No rinex, bye..."
  #\rm -rf $out_dir
   leave $result $archive_sol $work_dir $out_dir
   show_result "PPP" "$result |No RINEX, Bye ...|" $(uc $site) "$cal , $doy , $gwd"
   exit 1
 fi
 
 rnx=$(ls $site$doy*.${yy}o)

 if [ -z "$rnx" ]; then
   leave $result $archive_sol $work_dir $out_dir
   show_result "PPP" "$result |No RINEX, Bye ...|" $(uc $site) "$cal , $doy , $gwd"
   exit 1
 fi

 if [[ $delta_epoch != ''  ]]; then
    epochs=($(rnx2start_end $rnx))
    start_epoch=$(gpsdate -i "${epochs[0]}" time,sp,int|tr -d ' ' )
    end_epoch=$(gpsdate -i "${epochs[1]}" time,sp,int|tr -d ' ' )
    if [[ $delta_epoch != '' ]]; then
      end_epoch=$(gpsdate -i "${epochs[0]}" -delta "$delta_epoch" time,sp,int|tr -d ' ' )
    fi
    teqc_flags=( -start_window $start_epoch -end_window $end_epoch )
 fi

 
 fix_rnx $site $rnx $yyyy $doy $yy $ELEV_MASK
 
 rnx2qm $site $rnx 1> qfront.log 2> qfront.err
 
 if [ -e $site.qm ] && [ ! -z $site.qm ]; then
   mv $site.qm qmfile
 else
   leave $result $archive_sol $work_dir $out_dir
   show_result "PPP" "$result |ERROR: qmfile doesn't exists or zero length|" $(uc $site) "$cal , $doy , $gwd"
   exit 1
 fi
 
 #if [[ $ac == 'igs' || $ac == 'code' ]] && [[ ${#del_sat[@]} -eq 0 ]]; then
 if [[ $ac == 'igs' || $ac == 'code' || $ac == 'ig1' ]];  then
    del_sat=( ${del_sat[@]} $( igs_sum $date $orb_type $ac x | tr ',' ' ' ))
 fi
 
 weed_qmfile ${del_sat[@]}
 
 sites=($(qm_sta qmfile))
 sats=($(qm_sat qmfile))

 if [ ${#sats[@]} -lt $MIN_SATS ];then
   leave $result $archive_sol $work_dir $out_dir
   show_result "PPP" "$result |ERROR: qmfile has < $MIN_SATS sats, bye...|" $(uc $site) "$cal , $doy , $gwd"
   #echo2 "ERROR: qmfile has < $MIN_SATS sats, bye"
   exit 1
 fi
 
 qm_regres

 if [ $wlpb -eq 1 ]; then 
	if [ -e wlpb ]; then
		echo "Resolving Single-station Ambiguities"
 		estimation_wlpb  1>> estimation.log 2> estimation.err
	else
		leave $result $archive_sol $work_dir $out_dir
		show_result "PPP" "$result |No wlpb file, Bye ...|" $(uc $site) "$cal , $doy , $gwd"
   		exit 1
	fi
 else
 	estimation  1>> estimation.log 2> estimation.err
 fi
  

  if [ $save4fix -eq 1 ];then
     stacov4fix 
  fi 
  
  if [[ $PROC_TYPE == 's' ]]; then

     if [ $(ok_stacov estimation.stacov) -eq 1 ]; then
          stacov_name estimation.stacov $cal $site $ac $orb_type $frame
          proc_summary -o $cal.$site.summary.log -name $(uc $site) -date "$cal $doy $gwd"
	  tdry=`echo $site | tr '[:lower:]' '[:upper:]'`.TDPdry
	  # Create tropospheric files
	  make_tropo_table "tdp_final" ${site} "tropotable" "qregres.nml" "wash.nml" ${tdry}
          #make_process_log $cal $site
          compress -c qmfile > $cal.$site.qm.Z
          files=(smapper.log qregres.nml qfront.log postfit.log smcov.nio tdp_final wash.nml tropotable $tdry )
          pack_solution $cal $site ${files[@]}
          #save_solution  $work_dir $out_dir
          result='SUCCESS'
     fi
  
  elif [[ $PROC_TYPE == 'k' ]]; then 
     if [ -e xyz_smsol.txt ]; then
         tdp_sol 
         compress -c $site.lat > $cal.$site.lat.Z
         compress -c $site.lon > $cal.$site.lon.Z
         compress -c $site.rad > $cal.$site.rad.Z
         compress -c $site.envcov > $cal.$site.envcov.Z
         compress -c xyz_smsol.txt > $cal.$site.$frame.$RATE.smsol.Z  
         compress -c tdp_corr_final > $cal.$site.$frame.$RATE.corr.Z  
         proc_summary -o $cal.$site.summary.log -name $(uc $site) -date "$cal $doy $gwd"
         compress -c qmfile > $cal.$site.qm.Z
         files=(smapper.log qregres.nml qfront.log postfit.log smcov.nio tdp_final wash.nml )
         pack_solution $cal $site ${files[@]}
         result='SUCCESS'
     fi
  fi
  
  leave $result $archive_sol $work_dir $out_dir
  show_result "PPP" $result $(uc $site) "$cal , $doy , $gwd"
  #echo "RESULT: $result PPP site: $(uc $site) , date: $cal , $doy , $gwd"
 


