#!/usr/bin/env python

import re
import sys
import numpy

def read_nml(filename):
    """
    Read in some values from a namelist file.  This does not correctly 
    capture most values, but just the ones we need.
    """
    try:
        with open(filename,'r') as afile:
            nml = []
            for line in afile:
                if (line.strip().startswith('!')):
                    continue
                items = line.split('=')
                if (len(items) == 2):
                    key = items[0].strip()
                    value = items[1].strip()
                    value = value.strip("'")
                    nml.append({'key':key,'value':value})
    except IOError as e:
        print e
        sys.exit(1)

    # Unpack parenthesis with numbers
    unpacked_nml = []
    for item in nml:
        m = re.search('(\w+)\(\s*([\d, ]+)\s*\)',item['key'])
        if (m):
            index = map(int,m.group(2).split(','))
            if (len(index) == 1):
                index = index[0]
            unpacked_nml.append({'key':m.group(1),'index': index,'value':item['value']})
        else:
            unpacked_nml.append({'key':item['key'],'value': item['value']})


    return unpacked_nml

def read_tdp(filename):
    try:
        with open(filename,'r') as afile:

            wtrop = []
            for line in afile:
                items = line.split()
                seconds = float(items[0])
                apriori = float(items[1])
                value = float(items[2])
                sigma = float(items[3])
                parameter = items[4][0:8]
                site = items[4][8:]
                wtrop.append({'site':site,'epoch':seconds, 'parameter':parameter,'value': value,'sigma':sigma})
    except IOError as e:
        print e
        sys.exit(1)

    return wtrop

def get_sites(nml):
    sites = [x['value'] for x in nml if (x['key'] == 'NMLname')]
    return sites

def getNMLParameter(parameter,site,nml):
    sites = [x for x in nml if (x['key'] == 'NMLname')]
    index = [x['index'] for x in sites if (x['value'] == site)]
    item = [x for x in nml if (x['key'] == parameter and x['index'] == index[0])]
    return item[0]['value']

def getTDPParameter(parameter,epoch,tdp):
    values = [(x['value'],x['sigma']) for x in tdp if (x['epoch'] == epoch and x['parameter'] == parameter)]
    return values[0]

def fitPolynomial(site,tdp):
    # Extract the time series for this component
    t = [x['epoch'] for x in tdp if (x['parameter'] == 'DRYZTROP' and x['site'] == site)]
    y = [x['value'] for x in tdp if (x['parameter'] == 'DRYZTROP' and x['site'] == site)]

    p = numpy.polyfit(t,y,2)

    return p


def main(args):
    wash_nml = read_nml(args.wash_nml)
    qr_nml = read_nml(args.qr_nml)
    wtrop = read_tdp(args.tdp)
    dtrop = None
    if (args.dryz):
        dtrop = read_tdp(args.dryz)

    tropdrift = [x for x in wash_nml if (x['key'] == 'TROPDRIFT')]
    graddrift = [x for x in wash_nml if (x['key'] == 'GRADDRIFT')]

    fp = open(args.out,'w')
    fp.write("""#For more information about Troposphere Parameters, in a system with GOA 6.2 type: gd2p.pl -h_wash 
# GRADDRIFT = {0} km/sqrt(sec)
# TROPDRIFT = {1} km/sqrt(sec)
# SECS[J2000]  ZD[m] TROP_DRY[m] WETZ[m] WETZ_SIG[m] TRPAZSIN[km] TRPAZCOS[km]  TRPAZSIN_SIG[km] TRPAZCOS_SIG[km] SITE SNET  MAP_FUNC
""".format(graddrift[0]['value'],tropdrift[0]['value']))
    sites = get_sites(qr_nml)
    for site in sites:
        if (dtrop):
            pfit = fitPolynomial(site,dtrop)
        dry = float(getNMLParameter('NMLTROP_DRY',site,qr_nml))
        mapname = getNMLParameter('NMLTropMap',site,qr_nml)
        tdp = [x for x in wtrop if (x['site'] == site)]
        epochs = [x['epoch'] for x in tdp]
        epochs = sorted(set(epochs))
        for epoch in epochs:
            wetz = getTDPParameter('WETZTROP',epoch,tdp)
            sine = getTDPParameter('TRPAZSIN',epoch,tdp)
            cosine = getTDPParameter('TRPAZCOS',epoch,tdp)
            if (dtrop):
                dry = numpy.polyval(pfit,epoch)
            z = dry + wetz[0]
            out_string = " ".join(
                    map(str,(epoch,z,dry,wetz[0],wetz[1],sine[0],sine[1],cosine[0],cosine[1],site,args.net,mapname,"\n"))
                    )
            fp.write(out_string)

    fp.close()
              
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description='Interpolate DRYZ from a TDP file')
    parser.add_argument('--tdp',required=True,help='TDP file')
    parser.add_argument('--wash_nml',required=True,help='Wash Namelist File')
    parser.add_argument('--qr_nml',required=True,help='QRegres Namelist File')
    parser.add_argument('--out',required=True,help='Output file')
    parser.add_argument('--net',required=True,help='Network Name')
    parser.add_argument('--dryz',required=False,help='Dry Troposphere Estimate')

    args = parser.parse_args()
    
    main(args)
