#!/usr/bin/env python

import math
import copy

def _distance_and_bearing(point1,point2):
    """
    Calculate the ground distance between two Point objects

    @param point1 a point object
    @param point2 a point object
    @return the distance between the point objects
    """
    lon1 = point1.longitude
    lon2 = point2.longitude
    lat1 = point1.latitude
    lat2 = point2.latitude

    a = 6378137.0
    b = 6356752.314245  
    f = 1/298.257223563  # WGS-84 ellipsoid params
    L = math.radians((lon2-lon1))
    U1 = math.atan((1.0-f) * math.tan(math.radians(lat1)))
    U2 = math.atan((1.0-f) * math.tan(math.radians(lat2)))
    sinU1 = math.sin(U1)
    cosU1 = math.cos(U1)
    sinU2 = math.sin(U2)
    cosU2 = math.cos(U2)
  
    lambdaL = L
    lambdaP = 100
    iterLimit = 100
    while ((math.fabs(lambdaL-lambdaP) > 1e-12) and (iterLimit>0)):
        sinLambda = math.sin(lambdaL)
        cosLambda = math.cos(lambdaL)
        sinSigma = math.sqrt((cosU2*sinLambda) * (cosU2*sinLambda) + (cosU1*sinU2-sinU1*cosU2*cosLambda) * (cosU1*sinU2-sinU1*cosU2*cosLambda))
        if (sinSigma==0):
            return (0.0,0.0,0.0)  # co-incident points
        cosSigma = sinU1*sinU2 + cosU1*cosU2*cosLambda
        sigma = math.atan2(sinSigma, cosSigma)
        sinAlpha = cosU1 * cosU2 * sinLambda / sinSigma
        cosSqAlpha = 1.0 - sinAlpha*sinAlpha
        cos2SigmaM = cosSigma - 2.0*sinU1*sinU2/cosSqAlpha
        if (math.isnan(cos2SigmaM)): 
            cos2SigmaM = 0.0  # equatorial line: cosSqAlpha=0
        C = f/16.0*cosSqAlpha*(4.0+f*(4.0-3.0*cosSqAlpha))
        lambdaP = lambdaL
        lambdaL = L + (1.0-C) * f * sinAlpha * (sigma + C*sinSigma*(cos2SigmaM+C*cosSigma*(-1+2*cos2SigmaM*cos2SigmaM)))
        iterLimit = iterLimit - 1

    if (iterLimit==0): 
        return (float('nan'),0.0,0.0) # formula failed to converge

    uSq = cosSqAlpha * (a*a - b*b) / (b*b)
    A = 1.0 + uSq/16384.0*(4096.0+uSq*(-768.0+uSq*(320.0-175.0*uSq)))
    B = uSq/1024.0 * (256.0+uSq*(-128.0+uSq*(74.0-47.0*uSq)))
    deltaSigma = B*sinSigma*(cos2SigmaM+B/4.0*(cosSigma*(-1.0+2.0*cos2SigmaM*cos2SigmaM)-B/6.0*cos2SigmaM*(-3.0+4.0*sinSigma*sinSigma)*(-3.0+4.0*cos2SigmaM*cos2SigmaM)))
    s = b*A*(sigma-deltaSigma)
  
    #s = s.toFixed(3) # round to 1mm precision
    #return s
  
    # note: to return initial/final bearings in addition to distance, use something like:
    fwdAz = math.atan2(cosU2*sinLambda,  cosU1*sinU2-sinU1*cosU2*cosLambda)
    revAz = math.atan2(cosU1*sinLambda, -sinU1*cosU2+cosU1*sinU2*cosLambda)
    return (s,math.degrees(fwdAz),math.degrees(revAz))

def distance(point1,point2):
    (dist,fwdAz,revAz) = __distance_and_bearing(point1,point2)

    return dist

def bearing(point1,point2):
    (dist,fwdAz,revAz) = __distance_and_bearing(point1,point2)
    fwdAz = (fwdAz+360.0) % 360.0
    return fwdAz

class Point(object):

    def __init__(self,latitude=0.0,longitude=0.0,height=0.0):
        self.latitude = latitude
        self.longitude = longitude
        self.height = height 

    def __str__(self):
        return "%s %s %s" % (self.longitude,self.latitude,self.height)

    def __repr__(self):
        return "geodetic.Point(latitude=%f,longitude=%f,height=%f)" % (self.latitude,self.longitude,self.height)

    def toECEF(self):
        """
        Convert this point to a Cartesian Earth-centered Earth-fixed
        coordinate.

        @todo this is hard-coded to the WGS84 ellipsoid
        @return tuple of (x,y,z) in Cartesian coordinates relative to
                the center of the Earth [m]
        """
        # First convert the latitude and longitude to radians
        longitude = math.radians(self.longitude)
        latitude = math.radians(self.latitude)
        height = self.height

        a = 6378137.0 # Semi-major axis
        b = 6356752.3142 # Semi-minor axis
        e1 = 8.1819190842622e-2 # First eccentricity

        # Calculate the radius of curvature in meters
        N = a / (math.sqrt(1.0 - math.pow(e1, 2.0) * math.pow(math.sin(latitude), 2.0)))

        x = (N + height) * (math.cos(latitude) * math.cos(longitude))
        y = (N + height) * (math.cos(latitude) * math.sin(longitude))
        z = (N * math.pow(b, 2.0) / math.pow(a, 2.0) + height) * math.sin(latitude)

        return (x,y,z)

    def easting(self,point2,units='m'):
        """
        Calcuate the eastward component of distance between two points

        @return the eastward distance [m]
        """
        tmp = copy.copy(point2)
        tmp.latitude = self.latitude
        dist = self.distance(tmp,units)
        dist = math.copysign(dist,(self.longitude-point2.longitude))
        return dist

    def northing(self,point2,units='m'):
        """
        Calculate the northward component of distance between two points

        @return the northward distance [m]
        """
        tmp = copy.copy(point2)
        tmp.longitude = self.longitude
        dist = self.distance(tmp,units)
        dist = math.copysign(dist,(self.latitude-point2.latitude))
        return dist

    def distance(self,point2,units='m'):
        (dist,fwdAz,revAz) = _distance_and_bearing(self,point2)
        if (units == 'km'):
            dist = dist/1e3
        return dist

    def bearing(self,point2):
        (dist,fwdAz,revAz) = _distance_and_bearing(self,point2)
        fwdAz = (fwdAz+360.0) % 360.0
        return fwdAz

    def at(self,distance=None,azimuth=None):
        """
        Calculate a new coordinate at a given distance and azimuth
        """
        a = 6378137.0
        f = 1/298.257223563  # WGS-84 ellipsoid params
        r = 1.0-f
        lon1 = math.radians(self.longitude)
        lat1 = math.radians(self.latitude)

        tu = r*math.sin(lat1)/math.cos(lat1)
        faz = math.radians(azimuth)
        s = distance
        sf = math.sin(faz)
        cf = math.cos(faz)

        baz = 0.0
        if cf != 0.0:
            baz = 2.0*math.atan2(tu,cf)

        cu = 1.0/math.sqrt(1.0 + tu**2)
        su = tu*cu
        sa = cu*sf
        c2a = 1.0 - sa**2
        x = 1.0 + math.sqrt((((1.0/(r**2)) - 1.0 )*c2a) +1.0)
        x = (x-2.0)/x
        c = 1.0 - x
        c = ((x**2/4.0) + 1.0)/c
        d = x*((0.375*x**2)-1.0)
        tu = ((s/r)/a)/c
        y = tu

        while True:
            sy = math.sin(y)
            cy = math.cos(y)
            cz = math.cos(baz+y)
            e = (2.0*cz**2)-1.0
            c = y
            x = e*cy
            y = (2.0*e) - 1.0
            y = (((((((((sy*sy*4.0)-3.0)*y*cz*d)/6.0)+x)*d)/4.0)-cz)*sy*d) + tu
            if (math.fabs(y-c) < 1e-13):
                break

        baz = (cu*cy*cf) - (su*sy)
        c = r*math.sqrt((sa*sa) + (baz*baz))
        d = su*cy + cu*sy*cf
        lat2 = math.atan2(d,c)
        c = cu*cy - su*sy*cf
        x = math.atan2(sy*sf,c)
        c = (((((-3.0*c2a)+4.0)*f)+4.0)*c2a*f)/16.0
        d = ((((e*cy*c) + cz)*sy*c)+y)*sa
        lon2 = lon1 + x - (1.0-c)*d*f

        lat2 = math.degrees(lat2)
        lon2 = math.degrees(lon2)
        return lat2,lon2

    @staticmethod
    def fromECEF(x,y,z):
        """
        Construct a geodetic point from an Earth-centered Earth-fixed coordinate
        """
        # WGS84 ellipsoid constants:
        A = 6378137.0
        e = 8.1819190842622e-2

        # calculations:
        B   = math.sqrt(A**2*(1-e**2))
        ep  = math.sqrt((A**2-B**2)/B**2)
        p   = math.sqrt(x**2+y**2)
        th  = math.atan2(A*z,B*p)
        lon = math.atan2(y,x)
        lat = math.atan2((z+ep**2*B*math.sin(th)**3),(p-e**2.*A*math.cos(th)**3))
        N   = A/math.sqrt(1-e**2*math.sin(lat)**2)
        alt = p/math.cos(lat)-N

        # return lon in range [0,2*pi)
        lon = math.fmod(lon,2*math.pi)

        # correct for numerical instability in altitude near exact poles:
        # (after this correction, error is about 2 millimeters, which is about
        # the same as the numerical precision of the overall function)

        #my $k=abs(x)<1 & abs(y)<1;
        #$alt(k) = abs(z(k))-b;

        lat = math.degrees(lat)
        lon = math.degrees(lon)

        return Point(latitude=lat,longitude=lon,height=alt)
