#!/usr/bin/env python

import fnmatch
import os
import os.path
import time
import env
import drmaa
import logging
import tarfile
import subprocess

def prependPath(environment,newDirectory):
    environment['PATH'] = newDirectory + ":" + environment['PATH']
    return environment

class Gipsy(object):

    def __init__(self):
        self.session = None
        self.gipsyEnvironment = None
        self.workingDirectory = '/home/gnet/logs'
        self.logger = logging.getLogger(__name__)

    def initialize(self):
        """
        Initialize both the environment variables for GIPSY and the DRMAA session
        """
        self.gipsyEnvironment = env.readEnvironmentFile("/share/apps/cwu/etc/defaults")
        self.gipsyEnvironment['HOSTTYPE'] = 'x86_64'
        prependPath(self.gipsyEnvironment,self.gipsyEnvironment['GPS_HOME_DIR'] + "/bin")
        prependPath(self.gipsyEnvironment,self.gipsyEnvironment['GOA'] + "/bin/" + self.gipsyEnvironment['ARCH'])
        prependPath(self.gipsyEnvironment,self.gipsyEnvironment['GOA'] + "/bin")

        self.session = drmaa.Session()
        self.session.initialize()
        self.logger.info("Session started")

    def shutdown(self):
        """
        Shutdown the DRMAA session
        """
        self.session.exit()
        self.logger.info("Exited from session")

    def ppp(self,sites,date,orb,ac,proj,local=False,wlpb=False):
        """
        Run precise point positioning for a list of sites
        :params local: Tell ppp.sh that all data are local and do not need to be downloaded
        :params wlpb: Include wide-lane phase-bias for ppp ambiguity resolution
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        tarFileName = os.path.join(self.workingDirectory,'{0}-{1}-ppp.tar.gz'.format(gipsyDate,proj))
        tar = tarfile.open(tarFileName,'w:gz')
        joblist = []
        jobmap = {}
        for site in sites:
            jt = self.session.createJobTemplate()
            jt.jobEnvironment = self.gipsyEnvironment
            jt.workingDirectory = self.workingDirectory
            jt.remoteCommand = '/share/apps/cwu/bin/ppp.sh'
            arguments = ['-site',site,'-date',gipsyDate,'-orb',orb,'-ac',ac,'-proj',proj,'-noftp']
            if (local):
                arguments.append('-local')
            if (wlpb):
                arguments.append('-wlpb')
            if (proj in ('pbo-week','pbo-reanalysis','pbo-supplemental','pbo-supplemental6','panga')):
                self.logger.info('Adding VMF1 flag for site %s' % (site))
                arguments.extend(('-vmf1','-tropo0','vmf1'))
                arguments.append('-ion_2nd')
            jt.joinFiles = True
            jt.args = arguments
            self.logger.info("Running {0} {1}".format(jt.remoteCommand,jt.args)) 
            jobid = self.session.runJob(jt)
            joblist.append(jobid)
            jobmap[jobid] = site
            self.logger.info("Precise Point Positioning job for site %s submitted with id %s" % (site,jobid))
            self.session.deleteJobTemplate(jt)

        # Iterate over the list so that we can ballpark when
        # sites are finished
        while (len(joblist) > 0):
            toDelete = []
            for curjob in joblist:
                site = jobmap[curjob]
                status = self.session.jobStatus(curjob)
                if (status == drmaa.JobState.DONE):
                    retval = self.session.wait(curjob, drmaa.Session.TIMEOUT_WAIT_FOREVER)
                    self.logger.info("Precise Point Positioning Job for site %s (id %s) finished (exit status %s)" % (site,curjob,retval.exitStatus))
                    toDelete.append(curjob)
                elif (status == drmaa.JobState.FAILED):
                    retval = self.session.wait(curjob, drmaa.Session.TIMEOUT_WAIT_FOREVER)
                    self.logger.info("Precise Point Positioning Job for site %s (id %s) failed (exit status %s)" % (site,curjob,retval.exitStatus))
                    toDelete.append(curjob)
            # Remove finished jobs from the joblist
            for job in toDelete:
                if (job in joblist):
                    # Package up the log files
                    jobStdout = os.path.join(self.workingDirectory,'ppp.sh.o{0}'.format(job))
                    tar.add(jobStdout,arcname='ppp.sh.o{0}'.format(job))
                    os.remove(jobStdout)
                    joblist.remove(job)
                else:
                    self.logger.info("Job %s isn't in the list, weird" %  (job))
            # Wait a minute
            time.sleep(60)
        #self.session.synchronize(joblist,drmaa.Session.TIMEOUT_WAIT_FOREVER,False)
        #for curjob in joblist:
        #    logger.info("Collecting job %s" % (curjob))
        #    retval = self.session.wait(curjob, drmaa.Session.TIMEOUT_WAIT_FOREVER)
        #    site = jobmap[retval.jobId]
        #    logger.info("Precise Point Positioning Job for site %s (id %s) finished with status %s" % (site,str(retval.jobId),str(retval.hasExited)))
        tar.close()

    def amb(self,networks,date,orb,ac,proj):
        """
        Run Ambiguity resolution for a network of sites
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        tarFileName = os.path.join(self.workingDirectory,'{0}-{1}-amb.tar.gz'.format(gipsyDate,proj))
        tar = tarfile.open(tarFileName,'w:gz')

        joblist = []
        for network in networks:
            jt = self.session.createJobTemplate()
            jt.jobEnvironment = self.gipsyEnvironment
            jt.workingDirectory = self.workingDirectory
            jt.remoteCommand = '/share/apps/cwu/bin/sm_amb.sh'
            arguments = ['-snet',network,'-date',gipsyDate,'-orb',orb,'-ac',ac,'-proj',proj,'-clean']
            if (proj in ('panga','pbo-week','pbo-supplemental','pbo-supplemental6')):
                self.logger.info('Adding VMF1 flag for network %s' % (network))
                arguments.extend(('-vmf1','-tropo0','vmf1'))
            jt.joinFiles = True

            jt.args = arguments

            jobid = self.session.runJob(jt)
            joblist.append(jobid)
            self.logger.info("Ambiguity Resolution job for network %s submitted with id %s" % (network,jobid))
            self.session.deleteJobTemplate(jt)

        self.session.synchronize(joblist,drmaa.Session.TIMEOUT_WAIT_FOREVER,False)
        for curjob in joblist:
            self.logger.info("Collecting job %s" % (curjob))
            retval = self.session.wait(curjob, drmaa.Session.TIMEOUT_WAIT_FOREVER)
            self.logger.info("Ambiguity Resolution Job %s finished with status %s" % (str(retval.jobId),str(retval.exitStatus)))
            # Package up the log files
            jobStdout = os.path.join(self.workingDirectory,'sm_amb.sh.o{0}'.format(retval.jobId))
            tar.add(jobStdout,arcname='sm_amb.sh.o{0}'.format(retval.jobId))
            os.remove(jobStdout)

    def getOrbits(self,date,ac=None,type=None):
        """
        Get orbit files
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/get_orbits.pl','-ac',ac,'-type',type,'-date',gipsyDate,'-purge']
        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE,cwd='/tmp')
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

    def cleanOrbitCaches(self,date,ac=None,type=None):
        """
        For a given day, delete the orbits in the cache

        :param date: Orbit date
        :param ac: Processing center (e.g. jgp, igs, code)
        :param type: Orbit type (e.g. final, rapid)
        """
        self.logger.info("Cleaning orbit cache")
        # Form orbit directory name
        orbitString = '{0:s}_{1:s}'.format(ac.lower(),type.lower())
        dateString = date.strftime("%Y-%m-%d")
        # Form the remote directory name
        target = '{0:s}/orbits/{1:s}/{2:s}'.format(self.gipsyEnvironment['CACHE_DIR'],orbitString,dateString)
        # Form the call to rm
        rmString = "'/bin/rm -rf {0:s}'".format(target)
        args = ['/opt/rocks/bin/rocks','run','host','compute',rmString]
        p = subprocess.Popen(' '.join(args),env=self.gipsyEnvironment,stdout=subprocess.PIPE,cwd='/tmp',shell=True)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

    def populateOrbitCaches(self,date,ac=None,type=None):
        """
        Copy the orbit cache for a particular date from the head node to all of the compute nodes

        :param date: Orbit date
        :param ac: Processing center (e.g. jgp, igs, code)
        :param type: Orbit type (e.g. final, rapid)
        """
        self.logger.info("Populating Orbit Cache on Head Node")
        # First, run populate_orbit_cache.pl to populate the cache on the head node
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/populate_orbit_cache.pl','-ac',ac,'-type',type,'-date',gipsyDate]
        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE,cwd='/tmp')
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

        # Second, use rocks iterate host to create the orbit cache directories
        self.logger.info("Creating Orbit Cache Directories on Compute Nodes")
        orbitString = '{0:s}_{1:s}'.format(ac.lower(),type.lower())
        target = '{0:s}/orbits/{1:s}'.format(self.gipsyEnvironment['CACHE_DIR'],orbitString)
        mkdirString = "'/bin/mkdir -p {0:s}'".format(target)
        args = ['/opt/rocks/bin/rocks','run','host','compute',mkdirString]
        p = subprocess.Popen(' '.join(args),env=self.gipsyEnvironment,stdout=subprocess.PIPE,cwd='/tmp',shell=True)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

        # Third, use rocks iterate host '' to scp the cache from the head node to the compute nodes
        self.logger.info("Populating Orbit Caches on Compute Nodes")
        orbitString = '{0:s}_{1:s}'.format(ac.lower(),type.lower())
        source = '{0:s}/orbits/{1:s}/{2:s}'.format(self.gipsyEnvironment['CACHE_DIR'],orbitString,gipsyDate)
        target = '%:{0:s}/orbits/{1:s}/.'.format(self.gipsyEnvironment['CACHE_DIR'],orbitString)
        scpString = "'/usr/bin/scp -r {0:s} {1:s}'".format(source,target)
        args = ['/opt/rocks/bin/rocks','iterate','host','compute',scpString]
        p = subprocess.Popen(' '.join(args),env=self.gipsyEnvironment,stdout=subprocess.PIPE,cwd='/tmp',shell=True)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

    def getSites2(self,project=None,network=None,date=None):
        """
        Return a list of sites for a given network.  If network is None, return
        all sites
        """
        sites = []
        # Get the location of the files
        year = date.year
        year2 = date.strftime("%y")
        doy = date.strftime("%j")

        archiveDirectory = '{base}/{year}/{doy}'.format(base=self.gipsyEnvironment['RNX_DIR_ARCHIVE'],year=year,doy=doy)
        pattern = '????{doy}?.{year2}*'.format(doy=doy,year2=year2)
        for root,dirs,files in os.walk(archiveDirectory):
            for filename in fnmatch.filter(files,pattern):
                sites.append(filename[0:4])

        return sites 

    def getSites(self,project=None,network=None):
        """
        Return a list of sites for a given network.  If network is None, return
        all sites
        """
        p = None
        if (network is not None):
            p = subprocess.Popen(['/share/apps/cwu/bin/proj_list','-proj',project,'-snet',network],
                    env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        else:
            p = subprocess.Popen(['/share/apps/cwu/bin/proj_list','-proj',project,'-sites'],
                    env=gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
        # Cut up stdout into a list of strings and use
        # filter to remove zero length strings
        sites = filter(len,stdout.split('\n'))
        return sites

    def repairQmFile(self,filename):
        """
        Check and, if necessary, repair a QM file. It is assumed
        that the qm file is not compressed.
        """
        import shutil
        import goa.qmreader
        
        # Load the QM file
        try:
            qmfile = goa.qmreader.load(filename)
        except QmreaderError:
            print "Invalid qmfile {0}".format(filename)

        if (qmfile.isNative() == 0):
                output_qmfile = tempfile.NamedTemporaryFile(delete=False)
		# dump the input qmfile to the output qmfile
		cmd = "dump_qm -n -d " + filename + " | qm_dump_to_qm " + output_qmfile.name
		os.system(cmd)
                shutil.copyfile(output_qmfile.name,filename)

    def getNetworks(self,project):
        """
        Return a list of networks
        """
        p = subprocess.Popen(['/share/apps/cwu/bin/proj_list','-proj',project,'-nets'],
                env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
        # Cut up stdout into a list of strings and use
        # filter to remove zero length strings
        sites = filter(len,stdout.split('\n'))
        return sites

    def bulkDownload(self,project,date):
        """
        Bulk download RINEX data
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/get_proj_rnx_data','-proj',project,'-sdate',gipsyDate,'-archive']

        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)


    def summary(self,project,date,ac,amb):
        """
        Generate a summary
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/summary','-proj',project,'-date',gipsyDate,'-ac',ac,
                '-fields','rms','-rfields','dpos,chi,rms','-nmax','20']
        if (amb):
            args.append('-amb')
        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()

        return stdout

    def mergeNetworks(self,project,date,ac,orb,makeSinex=False):
        """
        Merge ambiguity resolved networks
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/mrg-net','-proj',project,'-date',gipsyDate,'-ac',ac,
                '-orb',orb]

        if (makeSinex):
            args.append('-snx')
        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

        return stdout

    def makeSinex(self,project,date,ac,orb):
        """
        Make a SINEX file
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/make_pbo_pp_snx','-proj',project,'-date',gipsyDate,'-ac',ac,'-orb',orb]
        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
        self.logger.info(stdout)
        if (stderr is not None):
            self.logger.error(stderr)

        return stdout

    def upload(self,project,date,ac):
        """
        Upload SINEX files for a specific date
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        args = ['/share/apps/cwu/bin/upload_pbo','-proj',project,'-date',gipsyDate,'-ac',ac]

        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()

        return stdout

    def createNetwork(self,date,proj):
        """
        Create the network list for a project
        """
        gipsyDate = date.strftime("%Y-%m-%d")
        scriptName = '/share/apps/cwu/bin/mk_%s_nets' % (proj)
        args = [scriptName,gipsyDate]

        p = subprocess.Popen(args,env=self.gipsyEnvironment,stdout=subprocess.PIPE)
        (stdout,stderr) = p.communicate()
