#!/usr/bin/env python

import geodetic

def read(filename):
    fp = open(filename,'r')

    stations = []
    for line in fp:
        items = line.split()
        name = items[0].lower()
        longitude = float(items[1])
        if (longitude < 0.0):
            longitude = longitude + 360.0
        latitude = float(items[2])
        station = {'name':name, 'coordinate': geodetic.Point(latitude=latitude,longitude=longitude)}

        stations.append(station)
    fp.close()

    return stations

def calculateDistances(point,stations):
    """
    Create a list of the distance from a point to all stations in a list
    """
    distances = []
    for station in stations:
        distance = point.distance(station['coordinate'],units='km')
        distances.append({'name':station['name'],'distance':distance})

    return distances

def calculateBarycenter(stations):
    """
    Given a list of stations, calculate the barycenter
    """
    # I don't think this is strictly correct
    longitude = 0.0
    latitude = 0.0
    for station in stations:
        longitude = longitude + station['coordinate'].longitude
        latitude = latitude + station['coordinate'].latitude

    longitude = longitude/float(len(stations))
    latitude = latitude/float(len(stations))
    return geodetic.Point(latitude=latitude,longitude=longitude)

def removePoints(points,network):
    """
    Remove points whose name appears in the list networks
    """
    
    newPoints = []
    for point in points:
        if (point['name'] not in network):
            newPoints.append(point)

    return newPoints

def calculateClosestSites(point,sites,number=30):
    """
    Calculate the n closest sites to a point
    """
    def sortfunc(x,y):
        return cmp(x['distance'],y['distance'])
    distances = calculateDistances(point,sites)
    closestSites = sorted(distances,sortfunc)

    nClosest = [x['name'] for x in closestSites[0:number]]
    closestPoints = [x for x in sites if (x['name'] in nClosest)]
    return closestPoints

def generateNetworkName(i,base='PBO'):
    """
    Generate a name for the network
    """
    network_name = '{0}_{1:02d}'.format(base,i)
    return network_name


def formSubnets(common_points,points,networkSize=30,commonSites=5):
    """
    :param networkSize: The number of sites per sub-network
    :param commonSites: The number of common sites per sub-network
    """

    i = 0
    networks = [{'network name':'NA_A', 'sites':[x['name'] for x in common_points]}]
    while (len(points) > 0):
        # Calculate the barycenter of the remaining points
        center = calculateBarycenter(points)
        # Calculate the site closest to the barycenter
        closestToBarycenter = calculateClosestSites(center,points,number=1)
        # Recalculate distances to this new point
        closestSites = calculateClosestSites(closestToBarycenter[0]['coordinate'],points,
                                             number=(networkSize-commonSites))
        # Calculate the barycenter of this new network
        subnet_center = calculateBarycenter(closestSites)
        # Now, calculate the closest common sites to the subnet
        closestCommonSites = calculateClosestSites(subnet_center,common_points,number=commonSites)

        network_name = 'PBO_{0:2s}'.format(str(i))
        network = [x['name'] for x in closestSites]
        network.extend([x['name'] for x in closestCommonSites])
        networks.append({'network name': network_name, 'sites':network})

        points = removePoints(points,[x['name'] for x in closestSites])
        i = i + 1
    print networks

def main():
    common_points = read('na_loc')
    points = read('reanalysis_loc')
    formSubnets(common_points,points)

def matplotlibTest():
    import matplotlib.animation as manimation
    from mpl_toolkits.basemap import Basemap
    import matplotlib.pyplot as plt
    import numpy as np

    common_points = read('na_loc')
    points = read('reanalysis_loc')

    i = 0
    # Create an animation
    metadata = dict(title='Tremor Bursts', artist='Walter Szeliga')
    writer = manimation.FFMpegFileWriter(fps=30,bitrate=-1, metadata=metadata)
    fig = plt.figure(tight_layout=True)
    axes = fig.add_subplot(111)
    basemap = Basemap(projection='ortho',lat_0=45,lon_0=-100,resolution='l',ax=axes)
    basemap.drawcoastlines(linewidth=0.25)
    basemap.drawcountries(linewidth=0.25)
    basemap.fillcontinents(color='coral',lake_color='aqua')
    # draw the edge of the map projection region (the projection limb)
    basemap.drawmapboundary(fill_color='aqua')
    # draw lat/lon grid lines every 30 degrees.
    basemap.drawmeridians(np.arange(0,360,30))
    basemap.drawparallels(np.arange(-90,90,30))

    centerMarker, = axes.plot([],[],'go')
    nearestMarker, = axes.plot([],[],'ro')
    commonMarkers = []
    markers = []
    for j in range(30):
        marker, = axes.plot([],[],'ko')
        markers.append(marker)
    for j in range(5):
        marker, = axes.plot([],[],'k^')
        commonMarkers.append(marker)

    with writer.saving(fig,'foo.mp4',180):
        while (len(points) > 0):
            center = calculateBarycenter(points)
            x,y = basemap(center.longitude,center.latitude)
            centerMarker.set_data(x,y)
            # Calculate the site closest to the barycenter
            closestToBarycenter = calculateClosestSites(center,points,number=1)
            # Recalculate distances to this new point
            closestSites = calculateClosestSites(closestToBarycenter[0]['coordinate'],points,number=30)
            # Calculate the barycenter of this new network
            subnet_center = calculateBarycenter(closestSites)
            # Now, calculate the closest common sites to the subnet
            closestCommonSites = calculateClosestSites(subnet_center,common_points,number=5)

            network_name = 'PBO_{0:2s}'.format(str(i))

            for j in range(len(closestCommonSites)):
                coord = closestCommonSites[j]['coordinate']
                x,y = basemap(coord.longitude,coord.latitude)
                commonMarkers[j].set_data(x,y)

            for j in range(len(closestSites)):
                coord = closestSites[j]['coordinate']
                x,y = basemap(coord.longitude,coord.latitude)
                if (j == 0):
                    nearestMarker.set_data(x,y)
                else:
                    markers[j].set_data(x,y)
            points = removePoints(points,[x['name'] for x in closestSites])
            i = i + 1
            axes.set_title(network_name)
            writer.grab_frame()
            

if __name__ == "__main__":
    main()
