#!/usr/bin/env python

import os.path
import distutils.dir_util
import datetime
import logging
from email.mime.text import MIMEText
import smtplib
import timeutils
import utilities 
import rsync.rsync
import gipsy6

logger = logging.getLogger(__name__)

class PBOProcess(object):

    def __init__(self,project=None,ac=None,orb=None):
        """
        Create a PBO process object

        @param project The project name
        @param ac acquisition center name [igs,jgp]
        @param orb orbit quality [rapid,final]
        """
        self.project = project
        self.ac = ac
        self.orb = orb
        self.gipsy = gipsy6.Gipsy()
        self.gipsy.initialize()
        
    def shutdown(self):
        self.gipsy.shutdown()

    def clean(self,date,rinex=False,ppp=False,amb=False):
        """
        Remove results from the processing directory
        """
        status = True
        # Get the location of the files
        year = date.year
        doy = date.strftime("%j")
        ymd = date.strftime("%Y-%m-%d")

        if (rinex):
            # RINEX directory
            processingDirectory = '{rnx_dir}/{year}/{doy}'.format(rnx_dir=self.gipsy.gipsyEnvironment['RNX_DIR'],year=year,doy=doy)
            distutils.dir_util.remove_tree(processingDirectory)
            #utilities.remove()
        if (ppp):
            # PPP directory
            processingDirectory = '{ppp_dir}/{ac}/{year}/{ymd}'.format(ppp_dir=self.gipsy.gipsyEnvironment['PPP_DIR'],ac=self.ac,year=year,ymd=ymd)
            distutils.dir_util.remove_tree(processingDirectory)
            #utilities.remove()

        if (amb):
            # AMB directory
            processingDirectory = '{amb_dir}/{ac}/{proj}/{year}/{ymd}'.format(amb_dir=self.gipsy.gipsyEnvironment['AMB_DIR'],
                    ac=self.ac,proj=self.project,year=year,ymd=ymd)
            distutils.dir_util.remove_tree(processingDirectory)
            #utilities.remove()
            #  Combination directory
            #processingDirectory = '{raw_comb_dir}/{ac}/{proj}/{year}'.format(raw_comb_dir=self.gipsy.gipsyEnvironment['RAW_COMB_DIR'],ac=self.ac,
            #        proj=self.project,year=year,ymd=ymd)
            #distutils.dir_util.remove_tree(processingDirectory)
            #utilities.remove()
        #  Combination directory
        processingDirectory = '{raw_comb_dir}/{ac}/{proj}/{year}'.format(raw_comb_dir=self.gipsy.gipsyEnvironment['RAW_COMB_DIR'],ac=self.ac,
                proj=self.project,year=year,ymd=ymd)
        products = self.productFilenames(date,amb=amb)
        for product in products:
            os.remove(os.path.join(processingDirectory,product))


        return status

    def archive(self,date,rinex=True,ppp=True,amb=True):
        """
        Move the results of a processing run into an archive directory
        """
        status = True
        # Get the location of the files
        year = date.year
        doy = date.strftime("%j")
        ymd = date.strftime("%Y-%m-%d")

        if (rinex):
            # RINEX directory
            processingDirectory = '{rnx_dir}/{year}/{doy}'.format(rnx_dir=self.gipsy.gipsyEnvironment['RNX_DIR'],year=year,doy=doy)
            archiveDirectory = '{base}/{year}/{doy}'.format(base=self.gipsy.gipsyEnvironment['RNX_DIR_ARCHIVE'],year=year,doy=doy)
            rsync.rsync.copy_tree(processingDirectory,archiveDirectory,update=True)
            #distutils.dir_util.copy_tree(processingDirectory,archiveDirectory)
            #status = (status and utilities.move(processingDirectory,archiveDirectory,overwrite=True))

        if (ppp):
            # PPP directory
            processingDirectory = '{ppp_dir}/{ac}/{year}/{ymd}'.format(ppp_dir=self.gipsy.gipsyEnvironment['PPP_DIR'],ac=self.ac,year=year,ymd=ymd)
            archiveDirectory = '{base}/{ac}/{year}/{ymd}'.format(base=self.gipsy.gipsyEnvironment['PPP_DIR_ARCHIVE'],ac=self.ac,year=year,ymd=ymd)
            rsync.rsync.copy_tree(processingDirectory,archiveDirectory,update=True)
            #distutils.dir_util.copy_tree(processingDirectory,archiveDirectory)
            #status = (status and utilities.move(processingDirectory,archiveDirectory,overwrite=True))

        if (amb):
            # AMB directory
            processingDirectory = '{amb_dir}/{ac}/{proj}/{year}/{ymd}'.format(amb_dir=self.gipsy.gipsyEnvironment['AMB_DIR'],
                    ac=self.ac,proj=self.project,year=year,ymd=ymd)
            archiveDirectory = '{base}/{ac}/{proj}/{year}/{ymd}'.format(base=self.gipsy.gipsyEnvironment['AMB_DIR_ARCHIVE'],
                    ac=self.ac,proj=self.project,year=year,ymd=ymd)
            rsync.rsync.copy_tree(processingDirectory,archiveDirectory,update=True)
            #distutils.dir_util.copy_tree(processingDirectory,archiveDirectory)
            #status = (status and utilities.move(processingDirectory,archiveDirectory,overwrite=True))

        #  Combination directory
        processingDirectory = '{raw_comb_dir}/{ac}/{proj}/{year}'.format(raw_comb_dir=self.gipsy.gipsyEnvironment['RAW_COMB_DIR'],ac=self.ac,
                proj=self.project,year=year,ymd=ymd)
        archiveDirectory = '{base}/{ac}/{proj}/{year}'.format(base=self.gipsy.gipsyEnvironment['RAW_COMB_DIR_ARCHIVE'],ac=self.ac,
                proj=self.project,year=year)
        products = self.productFilenames(date,amb=amb)
        for product in products:
            rsync.rsync.copy_file(os.path.join(processingDirectory,product),
                                  os.path.join(archiveDirectory,product),update=True)
        #distutils.dir_util.copy_tree(processingDirectory,archiveDirectory)
        #status = (status and utilities.moveFiles(processingDirectory,archiveDirectory,overwrite=True))

        return status

    def retrieveArchivedSolutions(self,date,ppp=True,amb=True):
        year = date.year
        ymd = date.strftime("%Y-%m-%d")

        if (ppp):
            # PPP directory
            archiveDirectory = '{base}/{ac}/{year}/{ymd}'.format(base=self.gipsy.gipsyEnvironment['PPP_DIR_ARCHIVE'],ac=self.ac,year=year,ymd=ymd)
            workingDirectory = "{base}/{ac}/{year}/{ymd}".format(base=self.gipsy.gipsyEnvironment['PPP_DIR'],year=year,ac=self.ac,ymd=ymd)
            copied = rsync.rsync.copy_tree(archiveDirectory,workingDirectory,update=True)
            if ( len(copied) == 0 ):
                logger.info("0 files copied from {0}!".format(archiveDirectory))
            #distutils.dir_util.copy_tree(archiveDirectory,workingDirectory)
        if (amb):
            # AMB directory
            archiveDirectory = '{base}/{ac}/{proj}/{year}/{ymd}'.format(base=self.gipsy.gipsyEnvironment['AMB_DIR_ARCHIVE'],
                               ac=self.ac,year=year,ymd=ymd,proj=self.project)
            workingDirectory = "{base}/{ac}/{proj}/{year}/{ymd}".format(base=self.gipsy.gipsyEnvironment['AMB_DIR'],
                               year=year,ac=self.ac,ymd=ymd,proj=self.project)
            rsync.rsync.copy_tree(archiveDirectory,workingDirectory,update=True)
            #distutils.dir_util.copy_tree(archiveDirectory,workingDirectory)

    def productFilenames(self,date,amb=False):
        """
        Cook up the file names for SINEX, RMS and Met products
        :params amb:  Boolean, whether this is a ppp run or ambiguity resolved
        """
        projectCode = {'pbo-rapid': 'a',
                        'pbo-week': 'b',
                        'pbo-supplemental': 'c',
                        'pbo-supplemental6': 'f',
                        'pbo-reanalysis': 'e',
                        'panga': ''}
        # Stacov file yyyy-mm-dd.comb.stacov
        # met file CWU{gpswd}.{yyyymmdd}.e.met.gz
        # SINEX file CWU{gpswd}.{yyyymmdd}.e.snx
        # RMS file cwu{gpswd}.{yyyymmdd.e.rms
        ymd = date.strftime("%Y%m%d")
        gpswd = "%04d%d" % (timeutils.datetimeToGPSWeek(date))
        code = projectCode[self.project]

        stacov = "{date}.comb.r.stacov".format(date=date.strftime("%Y-%m-%d"))
        if (amb):
            stacov = "{date}.comb.stacov".format(date=date.strftime("%Y-%m-%d"))
        met = "CWU{gpswd}.{date}.{code}.met.gz".format(gpswd=gpswd,date=ymd,code=code)
        snx = "CWU{gpswd}.{date}.{code}.snx".format(gpswd=gpswd,date=ymd,code=code)
        rms = "cwu{gpswd}.{date}.{code}.rms".format(gpswd=gpswd,date=ymd,code=code)

        return (stacov,met,snx,rms)

    def checkArchivedSolutions(self,sites,date):
        """
        Check whether a solution for sites exist in the archives
        """
        notProcessed = set()
        year = date.year
        ymd = date.strftime("%Y-%m-%d")
        frame = 'igs08'
        if (self.ac == 'jgp' and date >= datetime.datetime(2012,10,7)):
            frame = 'igb08'

        archiveDirectory = '{base}/{ac}/{year}/{ymd}'.format(base=self.gipsy.gipsyEnvironment['PPP_DIR_ARCHIVE'],ac=self.ac,year=year,ymd=ymd)
        for site in sites:
            if (self.orb == 'final'):
                stacovFile = '{ymd}.{site}.{frame}.stacov'.format(ymd=ymd,site=site,frame=frame)
            elif (self.orb == 'rapid'):
                stacovFile = '{ymd}.{site}.{frame}.r.stacov'.format(ymd=ymd,site=site,frame=frame)
            solution = os.path.join(archiveDirectory,site,stacovFile)
            if (not os.path.exists(solution)):
                logger.info("Station {0} not previously processed, adding".format(site))
                notProcessed.add(site)

        return notProcessed

    def constrainNetworks(self,sites,networks):
        """
        Iterate over a list of networks and retain only the ones
        that a site in the list sites
        """
        neededNetworks = set()
        for network in networks:
            networkSites = set(self.gipsy.getSites(project=self.project,network=network))
            # If no sites from the list sites are in the list networkSites, then
            # the two sets are disjoint
            if (not networkSites.isdisjoint(sites)):
               logger.info("Flagging network %s for reprocessing" % (network))
               neededNetworks.add(network)

        return neededNetworks 

    def getOrbits(self,date):
        """
        Get orbits
        """
        self.gipsy.getOrbits(date,ac=self.ac,type=self.orb)

    def populateOrbitCaches(self,date):
        """
        Copy orbits to each node
        """
        self.gipsy.populateOrbitCaches(date,ac=self.ac,type=self.orb)

    def bulkPPPandAMB(self,date,snet,amb,local,update,wlpb):
        """
        Perform Precise Point Positioning and Ambiguity Resolution using an bulk processing scheme
        whereby all of sites are point positioned in parallel and the processing waits for them to 
        be completed before beginning ambiguity resolution.  

        @param date A date string
        @param proj A project name, e.g. pbo-week
        @param snet A network name e.g. NCAL_A
        @param orb The orbit quality e.g. final
        """

        startTime = "Start Time:{0:s}\n".format(str(datetime.datetime.now()))
        # Get a list of known networks
        networks = self.gipsy.getNetworks(self.project)
        if (snet is not None):
            networks = [snet]

        # Build a list of unique site names
        sites = set()
        for network in networks:
            newSites = set(self.gipsy.getSites(project=self.project,network=network))
            sites.update(newSites)

        # Do precise point positioning
        logger.info("Beginning Precise Point Positioning")
        # Check whether we are just processing new sites
        if (update):
            sites = self.checkArchivedSolutions(sites,date)
            #networks = self.constrainNetworks(sites,networks)
            # Retrieve solution files from the archive
            self.retrieveArchivedSolutions(date,ppp=True,amb=False)
        if (len(sites) == 0):
            logger.info("All existing RINEX files have been processed")

        # Point position remaining
        if (len(sites) != 0):
            self.gipsy.ppp(sites,date,self.orb,self.ac,self.project,local=True,wlpb=wlpb)

        # Do Ambiguity resolution
        if (amb):
            self.gipsy.amb(networks,date,self.orb,self.ac,self.project)

        mergeOutput = self.mergeNetworks(date,amb)
        summaryOutput = self.gipsy.summary(self.project,date,self.ac,amb)
        stopTime = "Stop Time:{0:s}\n".format(str(datetime.datetime.now()))

        return startTime+stopTime+mergeOutput+summaryOutput

    def checkRinexArchive(self,date):
        """
        Copy RINEX files from the archive to the processing area
        """
        status = True
        year = date.strftime("%Y")
        doy = date.strftime("%j")

        archiveDirectory = "{base}/{year}/{doy}".format(base=self.gipsy.gipsyEnvironment['RNX_DIR_ARCHIVE'],year=year,doy=doy)
        workingDirectory = "{base}/{year}/{doy}".format(base=self.gipsy.gipsyEnvironment['RNX_DIR'],year=year,doy=doy)

        #status = utilities.copyFiles(archiveDirectory,workingDirectory)
        rsync.rsync.copy_tree(archiveDirectory,workingDirectory,update=True)
        return status

    def createNetwork(self,date):
        self.gipsy.createNetwork(date,self.project)

    def cleanOrbitCaches(self,date):
        self.gipsy.cleanOrbitCaches(date,ac=self.ac,type=self.orb)

    def mergeNetworks(self,date,amb):
        """
        Merge the network stacov files.  If we aren't resolving ambiguities, then just
        make the SINEX file
        """
        if (amb):
       	    mergeOutput = self.gipsy.mergeNetworks(self.project,date,self.ac,self.orb,makeSinex=True)
        else:
            mergeOutput = self.gipsy.makeSinex(self.project,date,self.ac,self.orb)

        return mergeOutput

    def bulkDownload(self,date):
        if (self.project.startswith('pbo')):
            project = 'pbo-all'
        elif (self.project.startswith('panga')):
            project = 'panga-all'
        else:
            project = self.project
        
        self.gipsy.bulkDownload(project,date)

    def upload(self,date):
        results = self.gipsy.upload(self.project,date,self.ac)
        return results

    def mail_summary(self,date,amb,body,addressList):
        """
        Email the summary to someone
        """
        server = smtplib.SMTP('localhost')
        sender = 'gnet@geology.cwu.edu'
        # If the address list consists of one email address
        # convert it to a list
        if (type(addressList) is str):
            addressList = [addressList]
        level = 'PPP'
        if (amb):
            level = 'AMB'
        # Construct the message
        message = MIMEText(body)
        message['Subject'] = "[Processing]: %s %s DONE: %s" % (self.project,level,date.strftime("%Y-%m-%d"))
        message['From'] = sender
        message['To'] = ','.join(addressList)

        # Send the message
        server.sendmail(sender,addressList,message.as_string())
        server.quit()

def parseInputDate(date):
    startDate = None
    if (type(date) is str):
        startDate = datetime.datetime.strptime(date,"%Y-%m-%d")
    elif (isinstance(date,datetime.date)):
        startDate = date
    
    return startDate

def main(args):

    addressList = args.address

    pboProcess = PBOProcess(args.proj,args.ac,args.orb)
    # Parse input date
    startDate = parseInputDate(args.date)

    for i in range(args.ndays):
        currentDate = startDate + datetime.timedelta(days=i)
       
        pboProcess.getOrbits(currentDate) 
        pboProcess.populateOrbitCaches(currentDate)
        # Download all known sites for a given date
        if (not args.local):
            pboProcess.bulkDownload(currentDate)
        pboProcess.checkRinexArchive(currentDate)

        pboProcess.createNetwork(currentDate)
        # Process available RINEX
        results = pboProcess.bulkPPPandAMB(currentDate,args.snet,args.amb,args.local,args.update,args.wlpb)
   
        # Archive data 
        status = pboProcess.archive(currentDate,rinex=True,ppp=True,amb=args.amb)
        if (not status):
            results = results + "Archiving Failed"
        else:
            results = results + "Data Archived Successfully"

        # Clean working directory
        pboProcess.clean(currentDate,rinex=True,ppp=True,amb=args.amb)
        # Clean orbit caches
        pboProcess.cleanOrbitCaches(currentDate)

        # Upload results
        if (args.upload):
            uploadResults = pboProcess.upload(currentDate)
            results = results + uploadResults

        # Email a summary of the processing
        pboProcess.mail_summary(currentDate,args.amb,results,addressList)
    pboProcess.shutdown()

if __name__ == "__main__":
    import argparse
    import logging.config

    parser = argparse.ArgumentParser(description="PBO Processing script")
    parser.add_argument('--date','-d',help="Date to process",required=True)
    parser.add_argument('--ndays',help='Number of days to process',type=int,default=1)
    parser.add_argument('--wlpb',help="Use wide-lane phase biases in point positioning",action='store_true',default=False)
    parser.add_argument('--amb',help="Ambiguity resolution (implies ppp)",action='store_true',default=False)
    parser.add_argument('--orb',help="Orbit type",default='rapid',choices=['rapid','final'])
    parser.add_argument('--ac',help="Acquisition center",default='igs',choices=['igs','jgp'])
    parser.add_argument('--proj',help="Project name")
    parser.add_argument('--snet',help='Process a particular sub-network')
    parser.add_argument('--local',help='Use local RINEX files only',action='store_true',default=False)
    parser.add_argument('--update',help='Use existing solutions and only process new RINEX data',action='store_true',default=False)
    parser.add_argument('--upload',help='Use local RINEX files only',action='store_true',default=False)
    parser.add_argument('--address',help='Email address list',nargs=1,default=['walter@geology.cwu.edu'])

    args = parser.parse_args()
    # Set up logging
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s "
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': format
            },
         },
         'handlers': {
             'file': {
                 'level': 'DEBUG',
                 'class': 'logging.handlers.RotatingFileHandler',
                 'filename': '/var/log/pbo/pbo_process.log',
                 'formatter': 'standard',
                 'maxBytes': 5242880,
                 'backupCount': 5
             }
         },
         'loggers': {
             '': {
                 'handlers': ['file'],
                 'level': 'DEBUG',
                 'propagate': True
             }
         }
    })
    main(args)
