#!/usr/bin/env python

import os
import os.path
import copy
import shutil
import datetime
import logging
import urllib
from email.mime.text import MIMEText
import smtplib
import gipsy5
import timeutils
import sinex
import network
import pbo_process
import rsync

logger = logging.getLogger(__name__)

def partition(l, n):
    """
    Partition the input iterable (i.e. list, set, &c) into chunks of size n
    """
    return [l[i:i+n] for i in range(0, len(l), n)]

class PBOSupplemental(pbo_process.PBOProcess):
    """
    Create a PBO process object

    @param project The project name
    @param ac acquisition center name [igs,jgp]
    @param orb orbit quality [rapid,final]
    """

    def __init__(self,project=None,ac=None,orb=None):
        pbo_process.PBOProcess.__init__(self,project=project,ac=ac,orb=orb)

    def createSubnets(self,notProcessed,alreadyProcessed):
        """
        Create networks that contain the n-closest sites to the
        previously unprocessed sites.

        :param notProcessed: a set of site names that were previously
                             unprocessed
        :param alreadyProcessed: a set of site names that were previously
                                 processed
        """
        # Get a list with all of the known sites and locations
        all_sites = network.read('/share/apps/cwu/etc/proj_lists/pbo_reanalysis/na_loc')  
        all_sites.extend(network.read('/share/apps/cwu/etc/proj_lists/pbo_reanalysis/reanalysis_loc'))
       
        processedSites = [x for x in all_sites if (x['name'] in alreadyProcessed)]
        unprocessedSites = [x for x in all_sites if (x['name'] in notProcessed)]

        # Calculate the barycenter of the unprocessed sites
        center = network.calculateBarycenter(unprocessedSites)
        # Now add in the 6 closest previously processed sites
        nClosestSites = network.calculateClosestSites(center,processedSites,number=6)
        toProcess = partition(unprocessedSites,12)
        networks = []
        for (i,chunk) in enumerate(toProcess):
            network_name = network.generateNetworkName(i,base='PBO')
            chunk.extend(nClosestSites)
            networks.append({'name':network_name, 'sites':chunk})

        # Form networks from the n-closest sites that have already been
        # processed
        #networks = []
        #for site in unprocessedSites:
        #    logger.debug("Calculating 5 closest sites to {0}".format(site))
        #    nClosestSites = network.calculateClosestSites(site['coordinate'],all_sites,number=5)
        #    network_name = network.generateNetworkName(i,base='PBO')
        #    networks.append({'name': network_name,'sites': nClosestSites}) 

        #logger.debug("{0} subnets created".format(len(networks)))
        return networks

    def extractSites(self,networks):
        """
        Given a dictionary of sites and their networks, extract just the list of
        unique site names
        """
        sites = set()
        for net in networks:
            for site in net['sites']:
                sites.add(site['name'])

        return sites

    def printSubnets(self,subnets):
        """
        Print out an ASCII version of the subnets for this particular project
        :param subnets: A dictionary containing the networks and their sites
        :return: a list of netowork names
        """ 
        filename = '/share/apps/cwu/etc/proj_lists/{0}_net_sites'.format(self.project)
        fp = open(filename,'w')
 
        networks = []
        for net in subnets:
            networkString = net['name']
            networks.append(net['name'])
            sites = [x['name'] for x in net['sites']]
            networkString = networkString + ' ' + ' '.join(sites) 
            logger.info("Adding network {0}".format(networkString))
            fp.write(networkString + '\n')

        fp.close()

        return networks

    def sitesInSinex(self,date,cleanup=True):
        """
        For a particular date, get the CWU SINEX file and read the sites that we processed from it
        """
        try:
            sinexFile = sinex.downloadSINEX(date,'cwu','final','loose')
        except IOError:
            sinexFile = None
        if (not sinexFile):
            raise OSError("File not found for {0},cwu,final,loose")
        sinexParser = sinex.SINEX()
        sinexParser.parse(sinexFile)
        sites = [x['Site Code'].lower() for x in sinexParser.solutionEpochs]
        if (cleanup):
            os.remove(sinexFile)

        return set(sites)

    def bulkPPPandAMB(self,date,snet,amb,local,update,wlpb):
        """
        Perform Precise Point Positioning and Ambiguity Resolution using an bulk processing scheme
        whereby all of sites are point positioned in parallel and the processing waits for them to 
        be completed before beginning ambiguity resolution.  

        @param date A date string
        @param proj A project name, e.g. pbo-week
        @param snet A network name e.g. NCAL_A
        @param orb The orbit quality e.g. final
        """

        startTime = "Start Time:{0:s}\n".format(str(datetime.datetime.now()))
        # Get a list of known networks
        networks = self.gipsy.getNetworks(self.project)
        if (snet is not None):
            networks = [snet]

        # Build a list of unique site names
        sites = set()
        for network in networks:
            newSites = set(self.gipsy.getSites(project=self.project,network=network))
            sites.update(newSites)

        # Get our SINEX solutions and use these to constrain networks
        try:
            sinexSites = self.sitesInSinex(date)
        except OSError:
            msg = "No SINEX for day {0}, skipping".format(date)
            stopTime = "\nStop Time:{0:s}\n".format(str(datetime.datetime.now()))
            logger.error(msg)
            return startTime + msg + stopTime
            
        supplementalSites = sites.difference(sinexSites)
        supplementalOutput = "There are {0} sites in the SINEX file,".format(len(sinexSites))
        supplementalOutput += " {0} new sites will be included\n".format(len(supplementalSites))
        logger.info(supplementalOutput)
        #networks = self.constrainNetworks(supplementalSites,networks)
        subnets = self.createSubnets(supplementalSites,sinexSites)
        networks = self.printSubnets(subnets)

        # Do precise point positioning
        logger.info("Beginning Precise Point Positioning")
        # Check whether we are just processing new sites
        if (update):
            # See which sites need to be reprocessed
            #sites = self.checkArchivedSolutions(sites,date)
            sites = self.extractSites(subnets)
            tieSites = sites.difference(supplementalSites)
            supplementalOutput += "Tie Sites: "
            supplementalOutput += " ".join(str(e) for e in tieSites)
            supplementalOutput += "\n"
            # Retrieve solution files from the archive
            self.retrieveArchivedSolutions(date,ppp=True,amb=False)
        if (len(sites) == 0):
            logger.info("All existing RINEX files have been processed")

        # Point position remaining
        if (len(sites) != 0):
            self.gipsy.ppp(sites,date,self.orb,self.ac,self.project,local=local,wlpb=wlpb)

        # Do Ambiguity resolution
        if (amb):
            self.gipsy.amb(networks,date,self.orb,self.ac,self.project)

        mergeOutput = self.mergeNetworks(date,amb)
        summaryOutput = self.gipsy.summary(self.project,date,self.ac,amb)
        stopTime = "Stop Time:{0:s}\n".format(str(datetime.datetime.now()))

        return startTime+stopTime+supplementalOutput+mergeOutput+summaryOutput

def parseInputDate(date):
    startDate = None
    if (type(date) is str):
        startDate = datetime.datetime.strptime(date,"%Y-%m-%d")
    elif (isinstance(date,datetime.date)):
        startDate = date
    
    return startDate

def main(args):

    addressList = args.address

    pboProcess = PBOSupplemental(args.proj,args.ac,args.orb)
    # Parse input date
    startDate = parseInputDate(args.date)

    for i in range(args.ndays):
        currentDate = startDate + datetime.timedelta(days=i)
       
        pboProcess.getOrbits(currentDate) 
        pboProcess.populateOrbitCaches(currentDate)
        # Download all known sites for a given date
        if (not args.local):
            pboProcess.bulkDownload(currentDate)
        pboProcess.checkRinexArchive(currentDate)

        pboProcess.createNetwork(currentDate)
        # Process available RINEX
        results = pboProcess.bulkPPPandAMB(currentDate,args.snet,args.amb,args.local,args.update,args.wlpb)
   
        # Archive data 
        try:
            status = pboProcess.archive(currentDate,rinex=True,ppp=True,amb=args.amb)
            results = results + "Data Archived Successfully"
        except rsync.errors.RsyncFileError:
            results = results + "Archiving Failed"

        # Clean working directory
        try:
            pboProcess.clean(currentDate,rinex=True,ppp=True,amb=args.amb)
        except:
            logger.error("Error while cleaning directories for date {0}".format(currentDate))
        # Clean orbit caches
        pboProcess.cleanOrbitCaches(currentDate)

        # Upload results
        if (args.upload):
            uploadResults = pboProcess.upload(currentDate)
            results = results + uploadResults

        # Email a summary of the processing
        pboProcess.mail_summary(currentDate,args.amb,results,addressList)
    pboProcess.shutdown()

if __name__ == "__main__":
    import argparse
    import logging.config

    parser = argparse.ArgumentParser(description="PBO Supplemental Processing script")
    parser.add_argument('--date','-d',help="Date to process",required=True)
    parser.add_argument('--ndays',help='Number of days to process',type=int,default=1)
    parser.add_argument('--wlpb',help="Use wide-lane phase biases in point positioning",action='store_true',default=False)
    parser.add_argument('--noppp',help="Precise point positioning only",action='store_false',default=True)
    parser.add_argument('--amb',help="Ambiguity resolution (implies ppp)",action='store_true',default=False)
    parser.add_argument('--orb',help="Orbit type",default='rapid',choices=['rapid','final'])
    parser.add_argument('--ac',help="Acquisition center",default='igs',choices=['igs','jgp'])
    parser.add_argument('--proj',help="Project name",choices=['pbo-supplemental','pbo-supplemental6'])
    parser.add_argument('--snet',help='Process a particular sub-network')
    parser.add_argument('--local',help='Use local RINEX files only',action='store_true',default=False)
    parser.add_argument('--update',help='Use existing solutions and only process new RINEX data',action='store_true',default=False)
    parser.add_argument('--upload',help='Use local RINEX files only',action='store_true',default=False)
    parser.add_argument('--address',help='Email address list',nargs=1,default=['walter@geology.cwu.edu'])

    args = parser.parse_args()
    format = "%(asctime)s- %(name)s - %(levelname)s - %(message)s "
    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': format
            },
         },
         'handlers': {
             'file': {
                 'level': 'DEBUG',
                 'class': 'logging.FileHandler',
                 'filename': '/var/log/pbo/pbo-supplemental.log',
                 'formatter': 'standard'
             }
         },
         'loggers': {
             '': {
                 'handlers': ['file'],
                 'level': 'DEBUG',
                 'propagate': True
             }
         }
    })

    main(args)
