#!/opt/python/bin/python2.7

import os
import os.path
import datetime
import textwrap
from ftplib import FTP
import blq

locFile = '/share/apps/cwu/sta_info.abs/sta_loc'
tideFile = '/share/apps/goa-var/sta_info/ocnld_coeffs'
rinexDirectories = ('/share/apps/data/rinex','/reprocessing/rinex')

def getRinexSites(date):
    """
    Inspect the RINEX directory for a particular date and determine
    which sites were downloaded
    """
    sites = set()
    year = date.year
    doy = date.strftime("%j")
    for rinexDirectory in rinexDirectories:
        directory = rinexDirectory + "/{year}/{doy}".format(year=year,doy=doy)
        for (root,dirs,files) in os.walk(directory):
            for name in files:
                sites.add(name[0:4])
    return sites

def getKnownSites():
    """
    Read the current sta_loc file and return the set of known stations
    :return: a set of known stations
    """
    sites = set()
    fp = open(locFile,'r')
    for line in fp.readlines():
        site = line.split()[0]
        sites.add(site.lower())

    return sites

def getSiteLists(dir):
    """
    Retrieve the current lists of present missing stations from UNAVCO
    :return: The location of the downloaded files
    """
    filenames = ('pbo_analysis_site_data_for_yesterday_expected_found.txt',
    'pbo_analysis_site_data_for_yesterday_expected_not_found.txt',
    'pbo_analysis_site_data_for_yesterday_not_expected_pending.txt',
    'pbo_analysis_sites_cors.txt',
    'pbo_analysis_sites_sopac.txt',
    'pbo_analysis_site_data_for_yesterday_not_expected_removed.txt')

    ftp = FTP('data-out.unavco.org')
    ftp.login()
    ftp.cwd('pub/pickup/analysis_centers')
    for filename in filenames:
        ftp.retrbinary('RETR {0:s}'.format(filename), open(os.path.join(dir,filename), 'wb').write)
    ftp.quit()

    return [os.path.join(dir,x) for x in filenames]

def mergeFiles(filenames):
    """
    Merge the site lists from the expected, seen and missing stations files from the PBO processing
    :return: the set of sites that should be included in the PBO processing
    """
    sites = set()
    for filename in filenames:
        fileSites = open(filename,'r').readlines()
        for line in fileSites:
            if (line[0] == '#'):
                continue
            items = line.split()
            site = items[0]
            sites.add(site.strip().lower())

    return sites

def missingLocations(expectedSites,knownSites):
    missingSites = expectedSites.difference(knownSites)
    output = "The following sites are missing from sta_loc: "
    for site in missingSites:
        output = output + "{0:s}, ".format(site)
    print textwrap.fill(output[0:-2]),"\n"

def missingTides(knownSites):
    """
    Read the list of known sites from the BLQ-formatted ocean loading file and compare this with the
    list of sites whose position we know.
    """
    tideStations = set()
    tideReader = blq.BLQReader()
    tideReader.parse(tideFile)
    for station in tideReader.tideData:
        tideStations.add(station.name.lower())

    missingSites = knownSites.difference(tideStations)
    output = "The following sites are missing from ocnld_coeffs: "
    for site in missingSites:
        output = output + "{0:s}, ".format(site)
    print textwrap.fill(output[0:-2]),"\n"

def missingRinex(expectedSites,rinexSites):
    missingSites = expectedSites.difference(rinexSites)
    output = "RINEX files for the following sites could not be found: "
    for site in missingSites:
        output = output + "{0:s}, ".format(site)
    print textwrap.fill(output[0:-2]),"\n"

def main(date):
    """
    Compare the set of sites in sta_loc with the set of sites expected in the PBO processing
    """
    expectedSites = mergeFiles(['/share/apps/cwu/etc/proj_lists/panga/panga_loc','/share/apps/cwu/etc/proj_lists/panga/na_loc'])
    knownSites = getKnownSites()
    missingTides(knownSites)
    missingLocations(expectedSites,knownSites)
    if (date):
        rinexSites = getRinexSites(date)
        missingRinex(expectedSites,rinexSites)

if __name__ == "__main__":
    import sys
    date = None
    if (len(sys.argv) > 1):
        date = datetime.datetime.strptime(sys.argv[1],"%Y-%m-%d")
    main(date)
