#!/usr/bin/env python

import math
import datetime

def mdy2mjd(dt):
    """
    Convert a datetime.datetime object into a modified Julian day

    @param dt a datetime.datetime object
    @return the modified Julian date corresponding to the input datetime
    """
    MARCH_1 = 306
    DAYS_PER_4_YEARS = 1461
    DAYS_PER_400_YEARS = 146097

    y = dt.year
    m = dt.month
    d = dt.day

    if( m <= 2 ):
        adj = (14 - m) / 12
        y -= adj
        m += 12 * adj
    elif ( m > 14 ):
        adj = (m - 3) / 12
        y += adj
        m -= 12 *adj

    if ( y < 0 ):
        adj = (399 - y) / 400
        d -= DAYS_PER_400_YEARS * adj
        y += 400 * adj

    d += (m * 367 - 1094) / \
         12 + y % 100 * DAYS_PER_4_YEARS / \
         4 + (y / 100 * 36524 + y / 400) - MARCH_1
    # d is now the Rata Die day for m/d/y
    # Convert from Rata Die to Julian Day, then truncate and subtract
    # to get Modified Julian Day
    mjd = d+1721424
    mjd = mjd-2400000

    # Convert fraction of a day
    mjd = mjd + (dt.hour*3600 + dt.minute*60 + dt.second)/86400.0

    return mjd

def mjd2mdy(mjd):
    """
    Convert from modified Julian date to datetime.datetime

    @param mjd a modified Julian day
    @return a datetime.datetime object representing the modified Julian day
    """

    jd = int(math.floor(mjd))

    # Add one to jd because mjd = jd - 2400000.5 which rounds up to 2400001 
    jd = jd+2400000+1

    l = jd + 68569
    n = ( 4 * l ) / 146097

    l = l - ( 146097 * n + 3 ) / 4
    i = ( 4000 * ( l + 1 ) ) / 1461001
     
    l = l - ( 1461 * i ) / 4 + 31
    j = ( 80 * l ) / 2447
    d = l - ( 2447 * j ) / 80
    l = j / 11;
    m = j + 2 - ( 12 * l )
    y = 100 * ( n - 49 ) + i + l

    return datetime.datetime(year=y,month=m,day=d)


def jdToDecimalYear(jd):
    """
    Convert a Julian Day to a decimal year
    """
    jd2k = mdy2mjd(datetime.datetime(year=2000,month=1,day=1))
    jd2k = float(jd2k) - 0.5
    decimalYear = (float(jd) - jd2k)/365.25 + 2000.0000
    return decimalYear

def decimalYearToJD(epoch):
    """
    Convert a decimal year to Julian day
    """
    jd2k = mdy2mjd(datetime.datetime(year=2000,month=1,day=1))
    jd2k = float(jd2k) - 0.5
    jd = (epoch-2000.0000)*365.25 + jd2k
    return jd

def datetimeToDecimalYear(dt):
    """
    Convert a datetime.datetime object into a decimal year

    @param dt datetime.datetime object
    @return floating point decimal year
    """
    jd = mdy2mjd(dt)
    decimalYear = jdToDecimalYear(jd)
    return decimalYear

def decimalYearToDateTime(epoch):
    """
    Convert a floating point decimal year into a datetime.datetime object

    @param epoch [float] decimal year
    """
    jd = decimalYearToJD(epoch)
    dt = mjd2mdy(jd)
    return dt

def datetimeToGPSWeek(dt):
    """
    Convert a datetime.datetime object to GPS week and day

    @param dt datetime.datetime object
    @return a tuple containing the GPS week and day

    >>> import datetime
    >>> datetimeToGPSWeek(datetime.datetime(2013,10,30,12,45,0))
    (1764, 3)
    >>> datetimeToGPSWeek(datetime.datetime(2013,10,30,0,0,0))
    (1764, 3)
    >>> datetimeToGPSWeek(datetime.datetime(2013,10,29,23,59,59))
    (1764, 2)
    """
    # GPS week 1 day 1
    week1 = datetime.datetime(year=1980,month=1,day=6)
    td = mdy2mjd(dt) - mdy2mjd(week1)
    if ( td < 0 ):
        return (None,None)
    weeks = int(td)/7.0
    gpsWeek = int(weeks)
    gpsDay = round((weeks-gpsWeek)*7.0)

    return (gpsWeek,int(gpsDay))

def gpsWeekToDateTime(gpsTuple):
    """
    Take a tuple of GPS week and day and return a datetime.datetime object
    for the same instant.

    @param gpsTuple a tuple of GPS week, GPS day
    @return datetime.datetime object for the same day
    """
    if (gpsTuple[0] < 0):
        print "invalid GPS week number %d" % (gpsTuple[0])
        return None
    if (gpsTuple[1] < 0 or gpsTuple[1] > 6):
        print "invalid GPS day number %d" % (gpsTuple[1])
        return None

    mjdDelta = gpsTuple[0]*7+gpsTuple[1]
    # 44244 is the modified Julian day value for GPS day 0, 6 Jan. 1980
    dt = mjd2mdy(44244 + mjdDelta)

    return dt

def getSunday(date):
    """
    Given a date, return the Sunday at the beginning of the week.  If the day
    provided is a Sunday, return the same day.

    >>> import datetime
    >>> getSunday(datetime.date(2013,9,15))
    datetime.date(2013, 9, 15)
    >>> getSunday(datetime.date(2013,9,25))
    datetime.date(2013, 9, 22)
    """
    sunday = None
    # If the input date is already a Sunday, just return that
    if (date.isoweekday() == 7):
        sunday = date
    else:   
        # The Sunday of the week is isoweekday() days prior
        sunday = date - datetime.timedelta(days=date.isoweekday())

    return sunday
