package GOA::GPT;

use strict;
use lib "$ENV{GOA}/perl_pm";
use Exporter;
use GPS::DATES;

our @ISA = ('Exporter');

our @EXPORT = qw();

# From http://www.navipedia.net/index.php/Tropospheric_Delay
sub tropospheric_delay {
	my $P = shift; # Pressure in mbar
	my $T = shift; # Temperature in Kelvin
	my $e = shift; # Water vapor pressure in mbar
	my $beta = shift; # Temperature lapse rate in K/m
	my $lambda = shift; # Water vapor lapse rate in (dimensionless)
	my $lat = shift; # Station latitude (degrees)
	my $height = shift; # Station height in m

	# Constants (i.e. magic numbers)
	my $k1 = 77.604; # K/mbar refractive effect of induced molecular polarization of dry gases
	my $k2 = 382000.0; # K^2/mbar same-same, but combinded water vapor and dipole moment 
	my $Rd = 287.054; # J/(kg*K) specific gas constant for dry air
	my $g = 9.80665; # m/s^2 reference value for mid-latitude gravity at the surface
	$lat = $lat*3.1415926536/180.0; # Latitude in radians
	my $gm = 9.784*(1.0 - 2.66e-3*cos(2.0*$lat) - 2.8e-7*$height); # m/s^2 gravitational acceleration of the atmospheric column at the latitude and height of the antenna

        # Zero altitude vertical delay terms (meters)
	my $Tz0_dry = 1e-6*$k1*$Rd*$P/$gm;
	my $Tz0_wet = (1e-6*$k2*$Rd/(($lambda+1.0)*$gm - $beta*$Rd))*($e/$T);

	# Correct for station height
        my $Tz_dry = (1.0 - $beta*$height/$T)**($g/($Rd*$beta))*$Tz0_dry;
	my $Tz_wet = (1.0 - $beta*$height/$T)**(($lambda+1.0)*$g/($Rd*$beta)-1.0)*$Tz0_wet;

	return ($Tz_dry,$Tz_wet);
}

sub water_vapor_pressure {
	my $T = shift; # Temperature in Kelvin
	my $rh = shift; # Relative humidity [0-1.0]

	$T = $T-273.16; # Convert from Kelvin to centigrade
	# Water vapor pressure in mbar
	my $P = $rh*6.11*10.0**(7.5* $T / ($T + 2.373e2));

	return $P;
}

sub gpt {
        my $cal = shift;  # Date in modified Julian days
	my $lat = shift;  # Latitude in degrees
	my $lon = shift;  # Longitude in degrees
	my $height = shift; # ellipsoidal Height in meters

	my $mjd = get_date($cal,'mjd');

	# Calculate the pressure and temperature using the GPT
	my $output = `$ENV{GPS_HOME_DIR}/bin/gpt $mjd $lat $lon $height`;
	$output =~ s/^\s+//;

	my ($pressure,$temperature,$geoid_height) = split /\s+/,$output;
	# Convert output of GPT to correct units
	$temperature = $temperature + 273.16; # Kelvin = centigrade + 273.16
	my $beta = 6.5e-3; # Temerpature lapse rate in K/m
	my $rh = 0.5; # Relative humidity
	my $e = water_vapor_pressure($temperature,$rh);
	my $lambda = 3.0;
	my $height = $height - $geoid_height; # Orthometric height
        	
	my($Tdry,$Twet) = tropospheric_delay($pressure,$temperature,$e,$beta,$lambda,$lat,$height);

	return ($Tdry,$Twet);
}

1;
