package GOA::Geodesy;
use Carp;
use Exporter;
@ISA=('Exporter');
@EXPORT=qw( xyz2llh xyz2llh2 rad2deg dd2deg $r2d distance);

our $r2d = 45/atan2(1,1);
our $a_e = 6378136.3;               # m IERS Tech Note 13 pg 5
our $f_e = 1/298.257223563;               # km IERS Tech Note 13 pg 5
our $e2 = $f_e *(2 - $f_e);

my $deg2rad = atan2(1,1)/45;

my  %units = (
      km => 1,
      m  => 1_000,
      cm => 1_000_00,
      mm => 1_000_000
    );

sub distance {
   my %opt = (
       unit => 'km',
       lon1 => 0,
       lat1 => 0,
       lon2 => 0,
       lat2 => 0,
       @_,
   );
   
   if ( ! ( $opt{lon1} && $opt{lat1} 
           && $opt{lon2} && $opt{lat2} ) ) {
     croak "missing one or more lon1,lat1,lon2,lat2\n";
   } 
   $opt{lon1} += 360 if $opt{lon1} < 0.0 ; 
   $opt{lon2} += 360 if $opt{lon2} < 0.0 ; 
   my $r = $a_e * $units{$opt{unit}};
   my ($lon1,$lat1,$lon2,$lat2) = map { $_*$deg2rad } 
                                  @opt{qw/lon1 lat1 lon2 lat2/};
   # This hsin formula is copied from Geo::Distance
   # its precision is enough to obtain the network configuration
   # we are not interested in calculating the exact distance
   my $dlon = $lon2 - $lon1;
   my $dlat = $lat2 - $lat1;
   my $a = (sin($dlat/2)) ** 2  + 
           cos($lat1) * cos($lat2) * (sin($dlon/2)) ** 2;
   my $c = 2 * atan2(sqrt($a), sqrt(1-$a));
   return $c*$r # result in km by default
}


sub xyz2llh {
  my $xyz = @_ == 3              ? [@_]  : 
            @{$_[0]} == 3        ? shift : 
                                   die "Usage: xyz2llh(  \$x, \$y,\$z  ) \n".
                                       "       xyz2llh( [\$x ,\$y,\$z] ) \n";
  my ($x,$y,$z)=@$xyz;
  my ($z2, $p2, $p, $r, $mu, $phi, $lat, $lon, $h,$n);
  ($x, $y, $z) = ($x/$a_e, $y/$a_e, $z/$a_e);
  $z2 = $z**2;
  $p2 = $x**2 + $y**2;
  $p = sqrt($p2);
  $r = sqrt($p2 + $z2);
  $mu = atan2($z * (1 - $f_e + $e2 / $r), $p);
  $phi = atan2($z * (1 - $f_e) + $e2 * (sin($mu))**3,
              (1 - $f_e) * ($p - $e2 * (cos($mu))**3));
  $lat = $phi * $r2d;
  $lon = atan2($y, $x) * $r2d;
  $n = $a_e/sqrt((1.0 - $e2*sin($lat)**2));
  $h = $a_e * ($p * cos($phi) + $z * sin($phi)
              - sqrt(1 - $e2 * (sin($phi))**2));
  $lon = $lon + 360 if $lon < 0;
  return ( $lat , $lon , $h )
}


sub xyz2llh2 {
  my $xyz = shift;
  my ($x,$y,$z)=@$xyz;
  my $a = $a_e;
  my $f = $f_e;
  my $p = sqrt( $x*$x + $y*$y);
  my $r = sqrt($p*$p + $z*$z);
  my $lon = atan2($y,$x);
  my $lat = atan2($z/$p, 1.0 - $e2 );
  my $n = $a/sqrt(1.0 - $e2*(sin($lat))**2);
  my $h = $p/cos($lat) - $n;
  my $oldh = 0;
  my $iter = 0;
  my $num=$z/$p;
  my $den;
  while ( abs($h-$oldh) > 1e-12 ) {
    last if $iter++ > 5;
    $oldh=$h;
    $den= 1.0 - $e2*$n/($n+$h);
    $lat = atan2($num,$den);
    $n = $a/sqrt((1.0 - $e2*(sin($lat))**2));
    $h = $p/cos($lat) - $n;
  }
  $lat = $lat*$r2d;
  $lon = $lon * $r2d;
  $lon = $lon + 360 if $lon < 0;
  #$h = sprintf("%7.1f",$h);
  return ( $lat , $lon , $h )
}
sub dd2deg {
  my $dd = shift;
  my $deg = int($dd);
  #my $min = int(frac($dd)*60);
  my $min = int(abs($dd-$deg))*60;
  #my $sec = sprintf("%4.1f",($dd - $deg - $min/60.0 )*60);
  my $sec = abs($dd - $deg - $min/60.0 )*60;
  return ($deg, $min, $sec)
}


sub frac {
  my ($int,$f) = split(/\./, sprintf("%f", shift));
  return $f
}
sub rad2deg {
  my $rad= shift;
  return $rad*$r2d;
}
