package GOA::StaPos;
use Fortran::F90Format;
use Carp;
use GPS::DATES qw( get_date cal2sec sec2cal);
use GPS::StrUtil;
use GPS::Defaults;
use Matrix;
use File::Basename;

my @fields = qw( sta 
                 year month day hours min sec
                 duration
                 x y z 
                 vx vy vz 
                 comment
               );

my $fmt =  '1x,a4,'.
           '1x,i4,4(1x,i2)1x,f5.2,'.
           '1x,f10.2,'.
           '1x,3f15.4,'.
           '1x,3e15.8,'.
           '1x,a';
           
my @today = get_date('today');

my %ROW  = (   sta   => '',
               year     => 0,
               month    => 0,
               day      => 0,
               hours    => 0,
               min      => 0,
               sec      => 0,
               date     => '',
               duration => 1000001.00, #a bunch of years
               x        => 0.0,
               y        => 0.0,
               z        => 0.0,
               vx        => 0.0,
               vy        => 0.0,
               vz        => 0.0,
               comment  => "Generated on @today by VMS",
             );
my %par;
sub new {
  my $class = shift;
    my $self = { };
    my %opt = (
	  file => '',
      overwrite => 0,
      @_,
    );

    bless $self,$class;
     
    $self->{fmt} = Fortran::F90Format->new( fmt => $fmt );
    
    $self->{overwrite} = $opt{overwrite};

    %par = get_defaults( );
    
    $self->{file} = "$par{STA_INFO_DIR}/sta_pos";
   
    if ( $opt{file} ) {
	  $self->file($opt{file});
      $self->load(%opt);
	}
    
    return $self

}
sub overwrite {
  my $self = shift;
  my $o    = shift || 0;
  $self->{overwrite}  = $o ? 1 : 0;
  return $self->{overwrite}
}
sub file {
  my $self = shift;
  if ( @_ ) {
    my $f = shift;
	if ( -e $f ) {
	  $self->{file} = $f;
	}
	else {
	  die "sta_pos file $f doesn't exists \n";
	}
  }
  return $self->{file}
}

sub fmt {
  my $self = shift;
  return $self->{fmt}
}

sub load {
  my $self = shift;
  my %opt  = ( @_,
              );
  my $file = -e $opt{file} ? $opt{file} : $self->file;
  open( $fh,"<", $file) or die "Couldn't open $file $!\n";
  my ($h,$secs);
  while( <$fh> ) {
    chomp;
    my $h = {};   
    @{$h}{@fields}=$self->fmt->read($_);
    $secs = cal2sec( @{$h}{qw/year month day hours min sec/} );
    @{$self->{Pos}{ $h->{sta}   }
                   { $secs       }
                   }{ qw/duration 
                         x y z 
                         vx vy vz 
                         comment /  } = @{$h}{qw/ duration   
                                                   x y z 
                                                   vx vy vz 
                                                   comment / };
                                                   
                 
                 
  }
  return $self

}


sub add {
    my $self = shift;
    my %e    = ( %ROW,
                 @_,
                ); 

    if ( $e{date} ) {
      @e{ qw/year month day hours min sec/} = get_date($e{date});
    }

    foreach my $f ( keys %e ) { delete $e{$f} if ! exists $ROW{$f}; } 

    $self->set( @e{@fields} ); 
    
}

sub set {
   my $self = shift;
   my @vals = @_ == @fields ? @_ : croak "Incorrect number of fields\n";
   my $h = {};

   @{$h} {@fields} = @ROW{@fields};
   @{$h}{@fields}=@vals;
 
    $h->{sta}   = uc $h->{sta};
    my $ok = $h->{sta}      !~ /^\w{4}$/   ?  0 : 
             $h->{year}     !~ /^\d{4}$/   ?  0 :
             $h->{month}    !~ /^\d{1,2}$/ ?  0 :
             $h->{day}      !~ /^\d{1,2}$/ ?  0 :
                                              1 ;
    if ( $ok ) { # add new entry
       my $secs = cal2sec( @{$h}{qw/year month day hours min sec/} );
       if ( exists $self->{Pos}{$h->{sta}}{$secs} && ! $self->{overwrite} ) { 
             carp "$h->{sta} $h->{year} $h->{month} $h->{day} already exists\n".
                  "try  overwrite => 1 \n";
             return 0 
       }
       @{$self->{Pos}{ $h->{sta}   }
                       { $secs          } } #slice
                       { qw/duration
                            x y z 
                            vx vy vz 
                            comment /
                       } 
                        = @{$h}{ qw/duration 
                                    x y z 
                                    vx vy vz 
                                    comment /
                                };
   } #ok  
   else {
       carp "Not enough information to add a new row (sta,date,x,y,z)\n";
   } 
}

sub select {
    my $self = shift;
    my $sta  = uc shift || '';
    my $hsecs = $self->get($sta);
    my $rows = [];
    my ($y,$m,$d,$hours,$min,$sec);
    
    foreach my $ss ( sort { $b <=> $a } keys %{$hsecs} ) {
      ($y,$m,$d,$hours,$min,$sec)=sec2cal($ss);
      my $hh = $self->get($sta,$ss);
      next if ! $hh;
      my $h={};
      @{$h}{@fields} = ( $sta, 
                         $y,$m,$d,$hours,$min,$sec,
                         @{$hh}{ qw/duration
                                    x y z
                                    vx vy vz
                                    comment / }
                       );
      push @$rows,$h; 
    }
    
    return $rows
}

sub xyz {
  my $self = shift;
  my $sta  = @_ ? uc shift : die "xyz: You need to provide a valid 4 char. station name\n";
  my $x = 0;
  if ( ! exists $self->{Pos}{$sta} ) {
     warn "# xyz: no entry for station $sta\n";
     return Matrix->new([0,0,0]);
  }
  my ($sec) = sort { $b <=> $a } keys %{$self->{Pos}{$sta}};
  #my $p = [ @{$self->{Pos}{$sta}{$sec}}{qw/x y z/} ] ;
  my $xyz =  Matrix->new( [ @{$self->{Pos}{$sta}{$sec}}{qw/x y z/}   ] );
  #for my $i ( 0 .. 2 ) {
  #   $xyz->[$i] = $p->[$i]+0;
  #}
  return $xyz
}

sub get {
   my $sta   = '';
   my $secs  = '';
   my $self  = shift;
   my $sta   = shift if @_ ;
   my $secs  = shift if @_ ;
   my $s = $self->{Pos};
   if ( $sta ne '' ) {
       $s =   exists $s->{$sta} ? $s->{$sta} : '';
   }
   if ( $sta && $s ne '' && $secs ne '' ) {                                    
      $s = exists $s->{$secs} ? $s->{$secs} : '' ; 
   }
   #   $s = $s->{$sta}   if $sta  && exists $s->{$sta}; 
   #   $s = $s->{$secs}  if $secs && exists $s->{$secs}; 
   return $s 
}

sub print {
   my $self = shift;
   my %opt = (
        file => '',
        @_,
   );
   my $fh;
   if ( $opt{file} ) {
      open( $fh , "> $opt{file}") 
        or croak "Couldn't open $opt{file} $!\n";
   }
   else {
     $fh = \*STDOUT;
   }
   my $hsta = $self->get;
   foreach my $st ( sort keys %{$hsta} ) {
         my $hsecs = $self->get($st);
         my $ssp = 'begin';
         foreach my $ss ( sort { $b <=> $a } keys %{$hsecs} ) {
             my $hh = $self->get($st,$ss);
             
         $hh->{duration} = $ssp eq 'begin' ? 1000001.00 : ($ssp - $ss)/86400.0;
             my ($y,$m,$d,$hours,$min,$sec)=sec2cal($ss);
                   
             my @v = ( $st,  
                       $y,$m,$d,$hours,$min,$sec,
                       @{$hh}{ qw/duration
                                  x y z 
                                  vx vy vz 
                                 comment /   }
                      );
                #print "$ssp $ss\n"; 
                #print "@v\n";
                print $fh $self->fmt->write(@v);
                $ssp = $ss;
        }
   }
       
   close $fh if $opt{file};

}

1;









