package GOA::StaSvec;
use Fortran::F90Format;
use Carp;
use GPS::APC;
use GPS::DATES qw( get_date cal2sec sec2cal );
use GOA::Util;
use GPS::Defaults;
use File::Basename;

my @fields = qw( sta_to sta_from
                  year month day hours min sec
                  duration  pc_ant
                  x y z height sflag
                  iyear imonth iday
                  comment
               );

my $fmt =   '1x,a4,1x,a4,1x,'.        # sta_to sta_from
             'i4,4(1x,i2)1x,f5.2,'.    # year month day hours min sec
             '1x,f12.2,1x,a9,1x,'.     # duration  pc_ant
             '4f11.4,1x,a1,1x,'.       # x y z height sflag
             'i4,1x,i2,1x,i2,'.        # iyear imonth iday
             '1x,a';                   # comment

my @today = get_date('today');

my %ROW  = (   sta_to   => '',
               sta_from => '',
               year     => 0,
               month    => 0,
               day      => 0,
               hours    => 0,
               min      => 0,
               sec      => 0,
               date     => '',
               igs_code => '',
               duration => 946080000.00, #30 years
               pc_ant   => 'ZERO',
               x        => 0.0,
               y        => 0.0,
               z        => 0.0,
               height   => 0.0,
               sflag    => 'l',
               iyear    => $today[0],
               imonth   => $today[1],
               iday     => $today[2],
               comment  => "Generated on @today by VMS",
             );

sub new {
  my $class = shift;
    my $self = { };
    my %opt = (
      sta          => '',
      dir  => '',
      file => '',
      @_,
    );

    bless $self,$class;

    # sta_svec format
    # (1x,a4,1x,a4,1x,i4,4(1x,i2)1x,
    # f5.2,1x,f12.2,1x,a9,1x,4f11.4,
    # 1x,a1,1x,i4,1x,i2,1x,i2)
    # added a field for comments
    $self->{fmt} = Fortran::F90Format->new( fmt => $fmt );
    
    my $apc_file = 'igs_01.pcv,ant_info.003';

    our %par = get_defaults( );
    
    $apc_file = $self->abs_pc_file;

    $self->{overwrite} = 0;
    $self->{dir} = $par{STA_INFO_DIR};
    $self->{dir} = $opt{sta_info_dir} if $opt{sta_info_dir};
    $self->{file} = "$self->{dir}/sta_svec"; 
    
    $self->{goa_names} = 0;

    $self->load_apc();
    

	if ( $opt{file} ) {
	  $self->file($opt{file});
      $self->load(%opt) ; 
	}
    
    return $self

}
sub sta_lines_fh {
   my $self = shift;
   my $sta  = uc shift;
   my $sta_lines = join '', grep { /^ $sta/ } slurp( $self->file ); 
   open( my $sh , '<', \$sta_lines ) 
       or die "sta_lines_fh: Couldn't open string for reading $!\n";
   return $sh     
}

sub load_apc {
    my $self = shift;
    my %opt  = ( apc_file => $self->abs_pc_file, 
                 skip_freq_values => 'yes',
                @_,);    
    #my $apc_file = 'igs_01.pcv,ant_info.003';

    $self->{apc} = GPS::APC->new( file => $opt{apc_file},
                                  skip_freq_values => 'yes',
                                  );     
    $self->{goa_names} = $self->apc->goa_names();
}

sub overwrite {
  my $self = shift;
  my $o    = shift || 0;
  $self->{overwrite}  = $o ? 1 : 0;
  return $self->{overwrite}
}

sub file {
  my $self = shift;
  if ( @_ ) {
    my $f = shift;
	if ( -e $f ) {
	  $self->{file} = $f;
	}
	else {
	  die "pcenter file $f doesn't exists \n";
	}
  }
  return $self->{file}
}

sub fmt {
  my $self = shift;
  return $self->{fmt}
}
sub abs_pc_file {
   my $self = shift;
   my $atx_file = shift || '';
   if ( ! $atx_file ) {
      ($atx_file) = #map { basename $_ } 
                    reverse 
                    sort glob "$par{IGS_INFO_DIR}/igs0*.atx";
   }
   return $atx_file
}
sub sv {
   my $self     = shift;
   my $sta_to   = shift || '';
   my $sta_from = shift || $sta_to;
   my $date     = shift || '' ;
   my $sec = get_date($date,'secj2000');
   my $p   = $self->get($sta_to,$sta_from);
   if ( ! $p ) {
      croak "No sta_svec entry for $sta_to !\n";
   }
   my @ant_dates = sort { $b <=> $a } keys %$p;
   if ( $sec < $ant_dates[-1] ) {
        croak "$sta_to date: $date is before the first sta_svec entry !\n";
   }
   my $psec=$ant_dates[0];
   foreach my $s ( @ant_dates ) {
       $psec = $s;
       last if $s <= $sec;
   }
   
   return  $self->get($sta_to,$sta_from,$psec);
}

sub apc {
  my $self = shift;
  return $self->{apc}
}
sub goa_names {
  my $self = shift;
  return wantarray ? %{$self->{goa_names}} : $self->{goa_names}
}
sub igs2goa_name {
  my $self=shift;
  my $igs = uc shift || '';
  my $goa = '';
  my $goa_names = $self->goa_names;
  if ( $igs ) {
     if ( exists $goa_names->{$igs} ) {
       $goa = $goa_names->{$igs}
     }  
     else {
        carp "IGS code $igs not found\n";
     }
     return $goa
  }
}

sub load {
  my $self = shift;
  my %opt  = ( 
               @_,
              );
              
  my $file = -e $opt{file} ? $opt{file} : $self->file;
  open( my $fh,"<", $file) or die "couldn't open $file $!\n";
  if ( $opt{sta} ) {
     $fh =  $self->sta_lines_fh( $opt{sta} )
  } 
  my ($h,$secs);
  while( <$fh> ) {
    chomp;
    my $h = {};   
    @{$h}{@fields}=$self->fmt->read($_);
    $secs = cal2sec( @{$h}{qw/year month day hours min sec/} );
    #next if exists $self->{Svec}{$h->{sta_to}}{$h->{sta_from}}{$secs};
    @{$self->{Svec}{ $h->{sta_to}   }
                   { $h->{sta_from} }
                   { $secs          }
                   }{ qw/duration   pc_ant
                      x y z height sflag
                      iyear imonth iday
                      comment /           } = @{$h}{qw/duration   pc_ant
                                                    x y z height sflag
                                                    iyear imonth iday
                                                    comment /             };
                                                   
                 
    #print "$_\n";             
  }
  close $fh;
  return $self

}
sub igs2pc_ant {
  my $self = shift;
  my $igs  = uc shift;
  my $goa  = 'ZERO';
  my %goa_names =  $self->goa_names;
  my ($try0,$try1,$try2,$try3)=($igs)x4;
     $try1 =~ s/\w{4}$/    /; # no radome   
     $try2 =~ s/\w{4}$/NONE/; # no radome   
     $try3   =~ s/\w{4}$//;   # any radome   
  my @try = ($try0,$try1,$try2);   
     $try1 =~ s/([^A-Z0-9_ ])/\\$1/g; #scape funny characters   
     $try2 =~ s/([^A-Z0-9_ ])/\\$1/g; #scape funny characters   
     $try3 =~ s/([^A-Z0-9_ ])/\\$1/g; #scape funny characters   
     $try3 = trim($try3);   
  my @any;   
     @any = grep /$try3/,grep !/($try1|$try2)/, keys %goa_names;
     push @try,@any if @any;
     foreach my $t ( @try ) {
       if ( exists $goa_names{$t} ) {
         $goa=$goa_names{$t};
         last;
       }
     }

   return $goa    
}

sub add {
    my $self = shift;
    my %e    = ( %ROW,
                 @_,
                ); 
    
    $self->load_apc if ! $self->apc;
    
    if ( $e{date} ) {
      @e{ qw/year month day hours min sec/} = get_date($e{date});
    }

    # Set the "vector start date", whatever that is, to the antenna start date
    $e{iyear} = $e{year};
    $e{imonth} = $e{month};
    $e{iday} = $e{day};

    if ( $e{igs_code} ) {
       $e{pc_ant} = $self->igs2pc_ant( $e{igs_code} );
    }

    
    $e{sta_from} = $e{sta_to} if ! $e{sta_from} && $e{sta_to};

    foreach my $f ( keys %e ) { delete $e{$f} if ! exists $ROW{$f}; } 

    $self->set( @e{@fields} ); 
    
}

sub set {
   my $self = shift;
   my @vals = @_ == @fields ? @_ : croak "Incorrect number of fields\n";
   my $h = {};

   @{$h} {@fields} = @ROW{@fields};
   @{$h}{@fields}=@vals;
 
    $h->{sta_to}   = uc $h->{sta_to};
    $h->{sta_from} = uc $h->{sta_from};
    my $ok = $h->{sta_to}   !~ /^\w{4}$/   ?  0 : 
             $h->{sta_from} !~ /^\w{4}$/   ?  0 :
             $h->{year}     !~ /^\d{4}$/   ?  0 :
             $h->{month}    !~ /^\d{1,2}$/ ?  0 :
             $h->{day}      !~ /^\d{1,2}$/ ?  0 :
                                              1 ;
    if ( $ok ) { # add new entry
       my $secs = cal2sec( @{$h}{qw/year month day hours min sec/} );
       if ( exists $self->{Svec}{$h->{sta_to}}
                          { $h->{sta_from} }
                          { $secs          }     ) {
           #carp "$h->{sta_to} $h->{sta_from} $secs already exists\n";
           return if ! $self->{overwrite} 
       }
       @{$self->{Svec}{ $h->{sta_to}   }
                      { $h->{sta_from} }
                      { $secs          } } #slice
                      { qw/duration   pc_ant
                           x y z height sflag
                           iyear imonth iday
                           comment /
                      } 
                      = @{$h}{ qw/duration   pc_ant
                                  x y z height sflag
                                  iyear imonth iday
                                  comment /
                              };


    } #ok  
    else {
       carp "Not enough information to add a new row (sta_to,date,pc_ant)\n";
    } 
}
sub select {
    my $self = shift;
    my $sta_to = uc shift || '';
    my $sta_from = shift || $sta_to;
    my $hsecs = $self->get($sta_to,$sta_from);
    my $rows = [];
    my ($y,$m,$d,$hours,$min,$sec);
    foreach my $ss ( sort { $b <=> $a } keys %{$hsecs} ) {
      ($y,$m,$d,$hours,$min,$sec)=sec2cal($ss);
      my $hh = $self->get($sta_to,$sta_from,$ss);
      my $h={};
      @{$h}{@fields} = ( $sta_to, $sta_from, 
                $y,$m,$d,$hours,$min,$sec,
                @{$hh}{ qw/duration  pc_ant
                           x y z height sflag
                           iyear imonth iday
                          comment /         }
               );
      push @$rows,$h; 
    }
    return $rows
}

sub get {
   my $sta_to   = '';
   my $sta_from = '';
   my $secs     = '';
   my $self     = shift;
   my $sta_to   = shift if @_ ;
   my $sta_from = shift if @_ ;
   my $secs     = shift if @_ ;
   my $s = $self->{Svec};
      $s = $s->{$sta_to}   if $sta_to   && exists $s->{$sta_to}; 
      $s = $s->{$sta_from} if $sta_from && exists $s->{$sta_from}; 
      $s = $s->{$secs}     if $secs     && exists $s->{$secs}; 

   return $s 
}

sub print {
   my $self = shift;
   my %opt = (
        file => '',
        @_,
   );
   my ($fh);
   if ( $opt{file} ) {
      open( $fh , "> $opt{file}") 
        or croak "Couldn't open $opt{file} $!\n";
   }
   else {
     $fh = \*STDOUT;
   }
   my ($sec_prev); 
   my $hsta_to = $self->get;
   foreach my $st ( sort keys %{$hsta_to} ) {
       my $hsta_from = $self->get($st);
       foreach my $sf ( sort keys %{$hsta_from} ) {
            my $hsecs = $self->get($st,$sf);
            $ssp = 'begin';
            foreach my $ss ( sort { $b <=> $a } keys %{$hsecs} ) {
                my $hh = $self->get($st,$sf,$ss);
                $hh->{duration} = $ssp eq 'begin' ? 946080000.00 : $ssp - $ss;
                my ($y,$m,$d,$hours,$min,$sec)=sec2cal($ss);
                   
                my @v = ( $st, $sf, 
                          $y,$m,$d,$hours,$min,$sec,
                          @{$hh}{ qw/duration  pc_ant
                                     x y z height sflag
                                     iyear imonth iday
                                     comment /         }
                        );
                #print "$ssp $ss\n"; 
                #print "@v\n";
                print $fh $self->fmt->write(@v);
                $ssp = $ss;
            }
             
       }
   }
       
  close $fh if $opt{file};

}

1;









