package GOA::Stacov;
use Data::Dumper;
use Matrix;
use Fortran::F90Format;
use GPS::DATES;
use GOA::Util;
use Carp;

#millimeters to meters
$mm2m = 1e-3;

#Max. number of parameters, currently limited by having an I4 integer.
$MAX_PAR = 9999;

#Header line format
#$Hfmt='1X,I4,15x,A7';
$Hfmt="1X,I4,' PARAMETERS ON ',A7,'.'";

#Parameter line format.
#$pfmt='1X,I4,2X,A16,2X,E22.15,6X,E21.15'
$Pfmt="1X,I4,2X,A4,1X,A3,1X,A7,2X,E22.15,2X,'+-',2X,E21.15";

#Correlation coeffs. format
$Cfmt='1X,I4,2X,I4,2X,E22.15';

#$Afmt='1X,A16,2X,F7.4,2X,F10.4,2X,F10.4';
$Afmt="1X,A4,1X,'ANTENNA',1X,A2,1X,2X,F7.4,2X,F10.4,2X,F10.4,' !up north east (m)'";

#Fortran::F90Format global objects used for read and write.
#$header_fmt  =Fortran::F90Format->new( fmt => $Hfmt);
#$par_fmt     =Fortran::F90Format->new( fmt => $Pfmt);
#$cc_fmt      =Fortran::F90Format->new( fmt => $Cfmt);
#$antenna_fmt =Fortran::F90Format->new( fmt => $Afmt);

my %ParName;
@ParName{qw/STA VEL JMP AAS AAC ASS ASC/}=(1)x7;

sub new {
  my $class = shift;
  my $self = { npar => 0,
               date => '', #get_date('today','epoch'),
               par  => {},
               cc   => {},
               cov  => [],
     antenna_offset => {},
           comments => [],
               file => '',
               @_,
             };

  bless $self,$class;
  
  $self->load if $self->{file};
 
  return $self
}
#Fortran::F90Format global objects used for read and write.
sub header_fmt {
 my $self = shift;
 
 return $self->{header_fmt} if exists $self->{header_fmt};

 $self->{header_fmt} = Fortran::F90Format->new( fmt => $Hfmt);
 
 return $self->{header_fmt}
}
sub par_fmt {
  my $self = shift;
  return $self->{par_fmt} if exists $self->{par_fmt};
  $self->{par_fmt} = Fortran::F90Format->new( fmt => $Pfmt);
}

sub cc_fmt {
 my $self = shift;
 
 return $self->{cc_fmt} if exists $self->{cc_fmt};
 
 $self->{cc_fmt}  = Fortran::F90Format->new( fmt => $Cfmt);
 
 return $self->{cc_fmt}
}

sub antenna_fmt {
 my $self = shift;
 return $self->{antenna_fmt} if exists $self->{antenna_fmt};
 
 $self->{antenna_fmt} = Fortran::F90Format->new( fmt => $Afmt);

 return $self->{antenna_fmt}
}

sub file {
  my $self=shift;
  return $self->{file}
}

sub date {
  my $self = shift;
  if ( @_ ) {
     $self->{date} = get_date(shift,'epoch');
  }
  return $self->{date}
}

sub npar {
  my $self = shift;
  if ( @_ ) {
     my $n = shift;
     croak "Too many parameters! $n > $MAX_PAR\n" if $n > $MAX_PAR;
     $self->{npar} = $n;
  }
  return $self->{npar};
}

sub comments {
  my $self=shift;
  if ( @_ ) {
    for my $c ( @_ ) {
       if ( $c =~ /^\s+\!.+/ ) {
         push @{$self->{comments}},$c;
       }
       else {
           croak "The text:\n$c\n is not a valid comment for a stacov file\n";
       }
    }
  }
  return wantarray ? @{$self->{comments}} : $self->{comments}
}

sub par {
   my $self=shift;
   if ( ! @_  ) {
     return $self->{par}
   }
   elsif ( @_ == 1 && trim($_[0]) =~ /^[A-Z0-9_]{4}$/i ) {
      my @param = ();
      my $site = uc shift;
      #foreach my $ipar ( map { sprintf ("% 4d",$_) } 1 .. $self->npar ) {
      foreach my $ipar (  1 .. $self->npar ) {
         next if $site && $self->{par}{$ipar}{site} ne $site ;
         my $h=$self->{par}{$ipar};
         $h->{ipar} = $ipar;
         push @param, $h
      }
      return @param
   }
   elsif ( @_ == 1 && trim($_[0]) =~ /^\d{1,4}$/ ) {
      my $i = shift;
      return $self->{par}{$i};
   }
   my %opt = ( ipar     => 0,
               site     => '', # 4 char. station name: 'ALBH'
               pname    => 'STA', # 3 char. parameter name 'STA' or 'VEL'
               cname    => '',
               xyz_cname => [qw/X Y Z/],
               xyz      => [], # xyz array ref./Matrix object 
                               # with xyz values
               xyz_sig  => [], # xyz array ref. or Matrix object 
                               # with xyz parameter sigmas 
               xyz_cc   => 0,  # hash of correlation coefficients  
               @_,
               );
   my $site = uc $opt{site};
   my $pname = uc $opt{pname};
   my $xyz  = ref $opt{xyz} eq 'Matrix' ? $opt{xyz} :
              ref $opt{xyz} eq 'ARRAY'        ? $opt{xyz}      :
              croak "par: Error. xyz Not an object or array ref on input\n";
   my $xyz_sig = ref $opt{xyz_sig} eq 'Matrix' ? $opt{xyz_sig} :
                 ref $opt{xyz_sig} eq 'ARRAY'        ? $opt{xyz_sig}      :
                 croak "par: Error. xyz_sig Not an object or array ref on input\n";
   my $cc = $opt{xyz_cc};
   croak "Invalid cc_par input\n" if  $cc && ref $cc ne 'HASH'; 
   croak "Invalid site name $site \n" if length($site) != 4;
   croak "Invalid parameter name $opt{pname} \n" if length($opt{pname}) != 3;
  
   if ( ! @$xyz && ! @$xyz_sig ) {
      croak "No values or sigmas to be set for $self->{file}\n";
   }

   my @c = map { sprintf "%-7s",$_ } @{$opt{xyz_cname}};
   my @f = qw(site pname cname val sig);
   
   my $k=0;
   my $npar = $self->npar;
   my @range  =    map { $_->{ipar} }
                   sort { $a->{cname} cmp $b->{cname} } 
                   grep { $_->{pname} eq $pname }
                   $self->par($site);
   
   if ( ! @range ) {
      $k= $npar;
      @range = ( 1..3 );
   }
   my $i=-1;
   my $ipar;
   my @par=();
   foreach my $j ( @range ) {
      $i++;
      $ipar=$k+$j;
      push @par,$ipar if $cc;
      $self->{par}{$ipar}{site}  = $site;
      $self->{par}{$ipar}{pname} = $pname;
      $self->{par}{$ipar}{cname} = $c[$i];
      $self->{par}{$ipar}{val}   = $xyz->[$i]     if defined $xyz->[$i];
      $self->{par}{$ipar}{sig}   = $xyz_sig->[$i] if defined $xyz_sig->[$i];
      $self->{npar}++ if $k+$j > $npar;
   }
   if ( $cc ) { 
      my ($i,$j,$pi,$pj,$rho);
      for my $i ( 2 .. 3 ) {
           $pi = $par[$i-1];
           for my $j ( 1 .. $i-1 ) {
             $pj = $par[$j-1];
             $rho = exists $cc->{$i}{$j} ? $cc->{$i}{$j} : 
                    exists $cc->{$j}{$i} ? $cc->{$j}{$i} :
                                           0.0;
             $self->{cc}{$pi}{$pj}=$rho;
           }
      }
   }          
   return $self->{par}
}


sub antenna_offset {
   my $self = shift;
   if ( ! @_  ) {
      return $self->{antenna_offset}
   }
   elsif ( @_ == 1 ) {
     my $sta = uc shift;
     if (  exists $self->{antenna_offset}{$sta} ) {
       my $a = $self->{antenna_offset}{$sta};
       return wantarray ? @{$a}{qw/up north east/}: $a
     }
     else {
       carp "No antenna offset for site $sta\n"; 
       return 0
     }
   }
   my %opt = (
         site =>  '',
         dtype => 'LC',
         up    => 0.0,
         north => 0.0,
         east  => 0.0,
      @_,
   );
   my @f = qw/dtype up north east/;
   my $site = $opt{site} or croak "No site name for antenna off. \n";
   @{$self->{antenna_offset}{$site}}{@f} = @opt{@f};   
   return $self->{antenna_offset}
}

sub ipar {
   my $self  = shift;
   croak "ipar: No integer parameter supplied\n" if ! @_;
   my $i = $_[0] =~ /^\d{1,4}$/ ? shift : 
            croak "$_[0] is not an integer\n";
 
   if ( ! exists $self->{par}{$i} ) {
     croak "$ip is not in this stacov parameter list\n";
   }
   return $self->{par}{$i} 
}

sub pname {
  my $self = shift;
  return $self->ipar(@_)->{pname} 
}
sub cname {
  my $self = shift;
  return $self->ipar(@_)->{cname}
}

sub site {
  my $self = shift;
  return $self->ipar(@_)->{site}
}

sub val {
  my $self = shift;
  my $i = shift;
  if ( ! @_ ) {
    return $self->ipar($i)->{val}
  }
  if ( exists  $self->{par}{$i} ) {
    return $self->{par}{$i}{val} =  shift;  
  }
  else {
     croak "$i is not in the parameter list\n";
  }

}

sub sig {
  my $self = shift;
  my $i = shift;
  if ( ! @_ ) {
    return $self->ipar($i)->{sig}
  }
  if ( exists  $self->{par}{$i} ) {
    return $self->{par}{$i}{sig} =  shift;  
  }
  else {
     croak "$i is not in the parameter list\n";
  }

}

sub sites {
   my $self = shift;
   my @sites =   map  { $self->{par}{$_}{site} }
                 grep { $self->{par}{$_}{cname} eq 'X      ' }
                 grep { $self->{par}{$_}{pname} eq 'STA' }
                 (  1 .. $self->npar  );
   return wantarray ? @sites : \@sites
}

sub xyz_vel {
   my $self = shift;
   return $self->xyz(shift,'VEL');
}

sub xyz_vel_sig {
   my $self = shift;
   return $self->xyz(shift,'VEL','sig');
}

sub xyz_sig {
   my $self = shift;
   return $self->xyz(shift,'STA','sig');
}

sub xyz {
   my $self = shift;
   my $site = uc shift  || '';
   my $pname = uc shift || 'STA';
   my $vname = lc shift || 'val'; # or 'sig'
   my @param;
   if ( ! $site ) {
       foreach my $ipar (  1 ..$self->npar ) {
          my $h = $self->{par}{$ipar};
          next if $h->{pname} ne $pname ;
          $h->{ipar} = $ipar; 
          push @param,$h;  
       }
   }
   else {
     @param = grep { $_->{pname} eq $pname } $self->par($site);
   }
   if ( ! @param ) {
     carp "$site not found in $self->{file}\n";
     return 0
   }
   
   @param = grep { $_->{pname} eq $pname } @param;
   
   if ( ! @param ) {
     carp "No $pname parameter for $site  in $self->{file}\n";
     return 0
   }
   
   if ( @param < 3 ) {
      my $npar = @param;
      carp "Missing parameters $npar < 3\n";
      return 0
   }
   
   my $xyz = Matrix->new([ map { $_->{$vname} } @param ]); 
   return $xyz

#   my $i=0;
#   foreach my $h ( @param ) {
#       $xyz->[$i] = $h->{$vname};
#       $i++;
#   }
#   return  $xyz 
}

sub cc {
   my $self = shift;
   my ($i,$j,$cc);
   if ( ! @_ ) {
      return $self->{cc}
   }
   elsif ( @_ == 1 ) {
     croak "cc cannot have only 1 argument\n".
           "Use cc_sta to supply 1 or 2 site names\n";
   }
   elsif ( @_ == 2 ) {
     $i =  shift;
     $j =  shift;
     $cc = exists $self->{cc}{$i} && exists $self->{cc}{$i}{$j} ? $self->{cc}{$i}{$j} : 
           exists $self->{cc}{$j} && exists $self->{cc}{$j}{$i} ? $self->{cc}{$j}{$i} :
                                                                  0.0;
     return $cc
   }
   elsif ( @_== 3 ) {
     $i =  shift;
     $j =  shift;
     $cc = shift;
     $self->{cc}{$i}{$j} = $cc;
     return $cc
   }
}

sub cc_sta {
   my $self = shift;
   if ( ! @_ ) {
       return $self->cc
   }
   my $sta_i =  uc shift;
   my $sta_j =  uc shift || $sta_i;
   my %par = ();
   my @sites =  $sta_i eq $sta_j ? ($sta_i) : ($sta_i,$sta_j);
   foreach my $site (  @sites ) {
      my @param=$self->par($site);
      next if ! @param ;
      push @{$par{$site}},   map { $_->{ipar} }
                             sort { $a->{cname} cmp $b->{cname} }
                             grep { $_->{pname} eq 'STA' } 
                             @param;
   }
   if ( ! @{$par{$sta_i}} && ! @{$par{$sta_j}} ) {
      croak "Site $sta_i|$sta_j are not in $self->{file}\n";
   }
   my ($i,$j,$ii,$jj)=(0)x4;
   my $cc_sta = {};
   my $cc =  $self->{cc};
   for $ii ( 2 .. 3  ) {
       $i = $par{$sta_i}[$ii-1];
       for $jj ( 1 .. $ii-1 ) {
          $j = $par{$sta_j}[$jj-1];
          $cc_sta->{$ii}{$jj}= exists $cc->{$i} && exists $cc->{$i}{$j} ? $cc->{$i}{$j} : 
                               exists $cc->{$j} && exists $cc->{$j}{$i} ? $cc->{$j}{$i} :
                                                                          0.0;
       }
   }
   return $cc_sta
}


sub cc2cov {
   my $self = shift;
   my $n=$self->npar();  
   my $c = Matrix->new([]);
   my ($sig_i,$sig_j,$rho,$i,$j);
   for my $ii ( 0 .. $n-1 ) {
     $i=$ii+1;
     $sig_i = $self->sig($i); 
     for my $jj ( 0 .. $ii-1 ) {
        $j=$jj+1;
        $sig_j = $self->sig($j); 
        $rho   = $i == $j ? 1.0 : $self->cc($i,$j);
        $c->[$ii][$jj]=$rho*$sig_i*$sig_j;
        if ( $j != $i ) {
          $c->[$jj][$ii]=$rho*$sig_i*$sig_j;
        }
     }
   }
  $self->{cov}=$c;
  return $self->{cov}
}

sub cov {
   my $self = shift;
   if ( ! @_ ) {
     return $self->cc2cov();
   }
   my $what = @_ ? ref $_[0] : '';
   if ( $what eq 'Matrix' || $what eq 'ARRAY' ) {
       my $cov = shift;
       if ( @{$cov} == @{$cov->[0]} ) {
           $self->{cov} = $cov;
           $self->cov2cc;
           return $cov 
       }
       else {
         die "cov: Error. Input array/matrix is not square\n";
       }
   }
   my $sta_i   =  uc shift  || '';
   my $sta_j   =  uc shift  || $sta_i;
   my $pname   =  uc shift  || 'STA';
   my $cname   =  uc shift  || '.';
   my %par = ();
   my @sites =  $sta_i eq $sta_j ? ($sta_i) : ($sta_i,$sta_j);
   my $kk=0;
   foreach my $site (  @sites ) {
      my @param=$self->par($site);
      die "Site $site not in $self->{file}\n" if ! @param ;
      push @{$par{$site}},   map { $_->{ipar} }
                             sort { $a->{cname} cmp $b->{cname} }
                             grep { $_->{cname} =~ /$cname/ } 
                             grep { $_->{pname} eq $pname   } 
                             @param;
   }
   my $c = Matrix->new([]);
   my ($sig_i,$sig_j,$rho,$ok);
   my ($i,$j,$ii,$jj,$ki,$kj,$ci,$cj)=(0)x8;
   #my $sig = $self->xyz_sig;
   for $ii ( 0 .. 2 ) {
       $i = $par{$sta_i}[$ii];
       for $jj ( 0 .. 2 ) {
            $j = $par{$sta_j}[$jj];
            $rho   = $i == $j ? 1.0 : $self->cc($i,$j);
            #$c->[$ii][$jj]=$rho*$sig->[$i-1]*$sig->[$j-1];
            $c->[$ii][$jj]=$rho*$self->sig($i)*$self->sig($j);
       }
   }          
   return $c

}

sub cov2sig {
   my $self = shift;
   my $sta  = shift || '';
   my $c = $self->{cov};
   my $n = @$c;
   my @ip = ( 0 .. $n-1 );
   my %isig =();
   if ( $sta ) {
      @ip = map { $_->{ipar}-1 } $self->par($sta);
   }
   my @sig = (0.0)x(@ip);
   my $sxx = 0.0;
   my $k=0;
   for my $i ( @ip ) {
     $sxx=sqrt $c->[$i][$i];
     $self->{par}{$i+1}{sig} = $sxx ;
     $sig[$k++]=$sxx ; 
   }
   return wantarray ? @sig : \@sig 
}

sub cov2cc {
  my $self = shift;
  my $c    = shift || $self->{cov}; 
  if ( @$c != @{$c->[0]} ) {
    die "cov2cc: input array is not square \n";
  }
  
  my $n=@$c;
  my @sig = $self->cov2sig;

  for my $i ( 0 .. $n-1 ) {
    for my $j ( 0 .. $i-1 ) {
       $self->{cc}{$i+1}{$j+1} = $c->[$i][$j]/$sig[$i]/$sig[$j];
    }
  }
  
  return $self->{cc}
}


sub baseline {
  my $self = shift;
  my $sta1 = uc shift || '';
  my $sta2 = uc shift || '';
  
  if ( ! $sta1 || ! $sta2 ) {
    die "You need more than 1 site to compute baseline! \n";
  }
  #if ( $sta1 eq $sta2 ) {
  #  return 0.0
  #}
  my $xyz1 = $self->xyz($sta1)->transpose;
  my $xyz2 = $self->xyz($sta2)->transpose;
  
  return $xyz1->subtract($xyz2);

}

sub baseline_len {
   my $self=shift;
   return $self->baseline(@_)->absolute
}

sub baseline_cov {
  my $self=shift;
  my $sta1= uc shift || '';
  my $sta2= uc shift || $sta1;
  
  if ( ! $sta1 && ! $sta2 ) {
    die "You need at least 1 site to compute baseline covariance! \n";
  }
  
  my $Caa = $self->cov( $sta1,$sta1 );
  my $Cbb = $self->cov( $sta2,$sta2 );
  my $Cab = $self->cov( $sta1,$sta2 )* -2; #->multiply_scalar(-2);

   
  return $Caa->add($Cbb)->add($Cab);
  
}



sub load {
   my $self = shift;
   my %opt = (  file => $self->{file},
                @_,
              );
   croak "No file name! use 'file => filename' \n" if ! $opt{file};
   
   open(my $fh,"<",$opt{file}) or croak "Couldn't open $opt{file} : $!\n";
   
   ($self->{npar},$self->{date})=$self->header_fmt->read( uc readline($fh) ); 
   
   $ipar=1;
   while ( <$fh> ) {
      ($ipar,$site,$pname,$cname,$x,$sx) = $self->par_fmt->read($_);
      #add 0 to force conversion from string to number 
      $ipar+=0;
      $self->{par}{$ipar++}={ site => $site , pname => $pname,
                              cname=> $cname, val   =>$x ,sig=>$sx };
      last if $. == $self->{npar}+1;
   }   
   while ( <$fh> ) {
      if ( /ANTENNA/ ) {
        ($site,$lc,$up,$north,$east)=$self->antenna_fmt->read($_);
        $self->{antenna_offset}{$site}={site=>$site, dtype => $lc,
                                        up =>$up,north=>$north,east =>$east};
        next;
      }
      if ( /^\s+\d+\s+\d+/) {
        ($ic,$jc,$cc)=$self->cc_fmt->read($_); 
        #add 0 to convert from string to number 
        $self->{cc}{$ic+0}{$jc+0}=$cc;
      }
      if ( /^\s\!/ ) {
         push @{$self->{comments}},$_;
      }
   } 
   close($fh);
}

sub print {
  my $self = shift;
  my %opt  = (
           fh => \*STDOUT,
           file => '',
           @_,
  );
  my $fh = $opt{fh};
  if ( $opt{file} ) {
     open( my $gh,"> $opt{file}" ) or croak "Couldn't open $opt{file} $!\n";
     $fh = $gh;
  }
  $self->print_header($fh);
  $self->print_par($fh);
  $self->print_cc($fh);
  $self->print_antenna($fh);
  $self->print_comments($fh);
}
sub print_comments {
  my $self = shift;
  my $fh   = shift || \*STDOUT;
  foreach my $c  ( @{$self->{comments}} ) {
    print $fh $c;
  } 
}
sub print_header {
  my $self = shift;
  my $fh   = shift || \*STDOUT;
  print $fh $self->header_fmt->write($self->{npar},$self->{date});
}

sub print_par {
  my $self = shift;
  my $fh   = shift || \*STDOUT;
  my @f=qw/site pname cname val sig/;
  my @p;
  foreach my $ipar ( 1..$self->npar ) {
     @p=@{ ${$self->{par}}{$ipar} } {@f};
     $p[2] = sprintf("%-7s",trim($p[2])); # format cname correctly
     print $fh $self->par_fmt->write($ipar,@p);
  }
}
sub print_cc {
  my $self = shift;
  my $fh   = shift || \*STDOUT;
  foreach my $ic ( sort {$a <=> $b } keys %{$self->{cc}} ) {
     foreach my $jc ( sort {$a <=> $b} keys %{${$self->{cc}}{$ic}} ) {
         print $fh $self->cc_fmt->write($ic,$jc, $self->{cc}{$ic}{$jc}); 
     }
  }
  
}

sub print_antenna {
  my $self = shift;
  my $fh   = shift || \*STDOUT;
  my @f=qw/dtype up north east/; 
  foreach my $s ( $self->sites() ) {
     next if ! exists $self->{antenna_offset}{$s};
     my @p = ($s);
     push @p,@{ ${$self->{antenna_offset}}{$s} } {@f};
     print $fh $self->antenna_fmt->write(@p); 
  }
}

sub remove {
   my $self= shift;
   my %opt = (
          site  => '',
          pname => '',
          full => 0,
          @_,
   );
   
   my %seen = ();
   my %sta=();
   my %par_name=();
   my @sites=();
   my @pnames = ();
   
   if ( ref $opt{site} eq 'ARRAY' &&  @{$opt{site}}   ) {
      @sites = map { uc $_ } @{$opt{site}};
      @sta{@sites}=@sites;
   }
   elsif ( $opt{site} ) {
      @sites  = grep { ! $seen{$_}++ } split( /,/ , uc $opt{site} );
      @sta{@sites}=@sites;
   }
   
   if ( ref $opt{pname} eq 'ARRAY' &&  @{$opt{pname}}  ) {
      @pnames = map { uc $_ } @{$opt{pname}};
      @par_name{@pnames}=@pnames;
   }
   elsif ( $opt{pname} ) {
      %seen = ();
      @pnames  = grep { ! $seen{$_}++ } split(/,/,uc $opt{pname});
      @par_name{@pnames}=@pnames;
   }
   
   my @ip=();
   if ( @sites ) {
      push @ip, grep{ exists $sta{$self->{par}{$_}{site}} } keys %{$self->{par}};
   }
   if ( @pnames ) { 
     push @ip, grep { exists $par_name{$self->{par}{$_}{pname}} } keys %{$self->{par}};
   }
   %seen=();
   @ip = grep { ! $seen{$_}++ } @ip;
   
   if ( @ip ) {
     $self->del_par(@ip); 
     $self->reorganize($opt{full});
   }
   #$self->rebuild;
   return $self
}



sub del_par {
    my $self=shift;
    $self->{npar}=$self->{npar} - scalar @_;
    my %seen;
    foreach my $ipar ( @_ ) {
         $seen{ $self->{par}{$ipar}{site} }++;
         delete $self->{par}{$ipar};    
    }
    foreach my $ic ( @_ ) {
       foreach my $jc ( keys %{${$self->{cc}}{$ic}} ) {
         delete $self->{cc}{$ic}{$jc}; 
         delete $self->{cc}{$jc}{$ic}; 
       }
    }
    foreach my $s ( keys %seen ) {
       delete $self->{antenna_offsets}{$s};
    } 
}

sub rebuild {
   my $self=shift;
   my $new_ip=1;
   my $new_ic=1;
   my $new_jc=1;

   my %par = ( par => {} );
   my %cc  = ( cc  => {} );

   my %ip; 
   my @f=qw/site pname cname val sig/;

   foreach my $ipar (   keys %{$self->{par}} ) {
       $ip{$ipar}=$new_ip;
       @{$par{par}{$new_ip}}{@f}=@{$self->{par}{$ipar}}{@f};         
       $new_ip++;
   }
  

   foreach my $ic ( keys %{$self->{cc}} ) {
       foreach my $jc (  keys %{${$self->{cc}}{$ic}} ) {
            if  ( $ip{$ic}  &&  $ip{$jc} ) {
              $cc{cc}{$ip{$ic}}{$ip{$jc}} = $self->{cc}{$ic}{$jc}; 
            }
       }
    }

    $self->{par}=$par{par};
    $self->{cc} = $cc{cc};

    return $self
}

sub reorganize {
  my $self = shift;
  my $full = shift || 0;
  my %par = ( par => {} );
  my %cc  = ( cc  => {} );

  my @ip=(undef);
  my @f=qw/site pname cname val sig/;
  my @iplist = keys %{$self->{par}};
  my %seen=();
  my @sites =   sort 
                grep { ! $seen{$_}++ }
                map  { $self->{par}{$_}{site} }
                @iplist;
                
  %seen=();
  my @pn    =  grep { ! $seen{$_}++ }
               map { $self->{par}{$_}{pname} }
               @iplist;
  my @pnames=(); 
  foreach my $p ( qw/STA VEL JMP AAS AAC ASS ASC/ ) {
      push @pnames,$p if exists $seen{$p};  
  }             
  my $cal = get_date($self->date,'cal'); 
  my (@param,@isite,$c,$d);
  foreach my $pn ( @pnames ) {
     foreach my $s ( @sites ) {
          push @ip, map  { $_->[0] }
                    sort { $a->[1] cmp $b->[1] }
                    map  { [ $_->[0], 
                             $_->[2] eq '  ' ? "$cal $_->[1]" :
                             get_date("$_->[2]:$_->[3]:00000",'cal') . " $_->[1]" ] }
                    map  { [ $_, unpack 'a1x1a2a3',$self->{par}{$_}{cname} ]  } 
                    grep { $self->{par}{$_}{pname} eq $pn }
                    grep { $self->{par}{$_}{site}  eq $s  }
                    @iplist;
     }
  }
  
  my $npar = @iplist;
  for my $i ( 1 .. $npar ) {
       $par{par}{$i} = $self->{par}{$ip[$i]}; 
       #@{$par{par}{$i}}{@f}=@{$self->{par}{$ip[$i]}}{@f};
       #print $self->par_fmt->write($i,@{$par{par}{$i}}{@f});
  }
  
  my ($ic,$jc,$rho);
  
  foreach my $i ( 2 .. $npar ) {
       $ic = $ip[$i];
       foreach my $j (  1 .. $i-1 ) {
	      $jc=$ip[$j];
          $rho = exists $self->{cc}{$ic} && exists $self->{cc}{$ic}{$jc} ? $self->{cc}{$ic}{$jc} : 
                 exists $self->{cc}{$jc} && exists $self->{cc}{$jc}{$ic} ? $self->{cc}{$jc}{$ic} :
                                                                           0;
          $cc{cc}{$i}{$j} = $rho if $rho || $full;    
	   }
  }
  $self->{par} = $par{par}; 
  $self->{cc}  = $cc{cc};

  return $self
}






