package GPS::Antex;
use Carp;
use Fortran::F90Format;
use GPS::StrUtil;
use Scalar::Util qw(looks_like_number);
use Cwd;

our %atx = (
 'ANTEX VERSION / SYST' =>  { fmt      => 'F8.1,12X,A1,39X,A20',
                              fields   => [   qw/format_version 
                                              satellite_system/
                                          ],
                              format_version    => 1.3,
                              satellite_system => 'M', # GPS      => 'G'
                                                       # GLONASS  => 'R',
                                                       # GALILEO  => 'E', 
                                                       # Compass  => 'C',
                                                       # QZSS     => 'J',
                                                       # SBAS     => 'S',
                                                       # MIXED    => 'M'
                            },
 'PCV TYPE / REFANT'    =>  { fmt          => 'A1,19X,A20,A20,A20',
                              fields       => [qw/pcv_type 
                                                 ref_ant_type 
                                                 serial_num /
                                              ],
                              #pcv_type     => 'R', #relative
                              pcv_type     => 'A', #absolute
                              #ref_ant_type => 'AOA/M_T', # or Blank
                              ref_ant_type => '', # or Blank
                              serial_num   => ''  # Blank: all 
                            },
 
 'COMMENT'              =>  { fmt     =>'A60,A20', 
                              fields => ['atx_comment'],
                              atx_comment => '',
                            },
 'END OF HEADER'        =>  { fmt =>'60X,A20' ,
                              fields => ['eoh'], 
                              eoh => 'END OF HEADER' 
                            },
 'START OF ANTENNA'     =>  { fmt => '60X,A20',
                              fields => ['start_of_ant'],
                              start_of_ant => 'START OF ANTENNA',
                            },
 'TYPE / SERIAL NO'     =>  { fmt => 'A20,A20,A10,A10,A20',
                              fields => { receiver => [qw/ant_type 
                                                          serial_num
                                                          blank
                                                          blank
                                                      /],
                                          satellite => [qw/sat_ant_type
                                                           sat_code_nn
                                                           sat_code_nnn
                                                           cospar_id 
                                                        /],
                                        },
                              ant_type      => '',
                              serial_num    => '',
                              blank         => '',
                              sat_ant_type  => '', #
                              sat_code_nn   => '', # PRN number
                              sat_code_nnn  => '', # GPS number
                              cospar_id     => '',
                            },
 'METH / BY / # / DATE' =>  {fmt    => 'A20,A20,I6,4X,A10,A20',
                             fields => [qw/calibration_method 
                                           agency 
                                           ant_calibrated_num
                                           date/
                                       ],
                             calibration_method => '', #CHAMBER,ROBOT,FIELD
                                                       #COPIED,
                                                       #CONVERTED,BLANK
                             agency => '',
                             ant_calibrated_num => '',
                             date               => '', #DD-MMM-YY
                            },
 'DAZI'                 =>  { fmt  =>'2X,F6.1,52X,A20', 
                              fields => ['dazi'],
                              dazi => '', # 5.0 or 0.0 for non az. dep.
                            },
 'ZEN1 / ZEN2 / DZEN'   =>  { fmt => '2X,3F6.1,40X,A20' ,
                              fields => [qw/zen1 zen2 dzen/ ],
                              zen1 => '',
                              zen2 => '', #zen2 > zen1
                              dzen => '',
                            },
 '# OF FREQUENCIES'     =>  { fmt => 'I6,54X,A20',
                              fields => ['num_freq'], 
                              num_freq => ''
                            },
 'VALID FROM'           =>  { fmt => '5I6,F13.7,17X,A20', 
                              fields => [qw/ syyyy smm sdd shh smin ssec/],
                              syyyy => '',
                              smm => '',
                              sdd   => '',
                              shh  => '',
                              smin => '',
                              ssec => '',
                            },
 'VALID UNTIL'          =>  { fmt => '5I6,F13.7,17X,A20', 
                              fields => [qw/ eyyyy emm edd ehh emin esec/],
                              eyyyy => '',
                              emm => '',
                              edd   => '',
                              ehh  => '',
                              emin => '',
                              esec => '',
                            },
 'SINEX CODE'           =>  { fmt => 'A10,50X,A20', 
                              fields => ['snx_code'],
                              snx_code => '' 
                            },
 'START OF FREQUENCY'   =>  { fmt => '3X,A1,I2.2,54X,A20',
                              fields   => [qw/sat_sys freq_num/],
                              sat_sys  => '', #G : GPS
                                              #R : GLONASS
                                              #E : Galileo

                              freq_num => '', #G01 -> L1
                                              #G02 -> L2
                                              #G05 -> L5
                                              #R01 -> G1
                                              #R02 -> G2
                                              #E01 -> E1
                                              #E05 -> E5a
                                              #E07 -> E5b
                                              #E08 -> E5 (E5a+E5b)
                                              #E06 -> E6
                                              #
                            },
 'NORTH / EAST / UP'    =>  {fmt => '3F10.2,30X,A20',
                             fields => [qw/north east up/],
                             north => '',
                             east  => '',
                             up    => '',
                            },
 'VALUES NOAZI'         =>  { fmt => "3X,A5,15F8.2",# default 15 values;
                                                    # will get the total number
                                                    # on the fly
                              fields => [ qw/noazi values_noazi/ ],
                              noazi  => 'NOAZI',
                              values_noazi => [], 
                            },
 'VALUES DAZI'          =>  { fmt => "F8.1,15F8.2", # default 15 values;
                                                    # will get the total number
                                                    # on the fly
                              fields => [ qw/dazi values_dazi/],
                              dazi  => '',
                              values_dazi => [], 
                            },
 'END OF FREQUENCY'     =>  { fmt => '3X,A1,I2.2,54X,A20',
                              fields   => [qw/sat_sys freq_num/],
                              sat_sys  => '', #G : GPS
                                              #R : GLONASS
                                              #E : Galileo

                              freq_num => '', #G01 -> L1
                                              #G02 -> L2
                                              #G05 -> L5
                                              #R01 -> G1
                                              #R02 -> G2
                                              #E01 -> E1
                                              #E05 -> E5a
                                              #E07 -> E5b
                                              #E08 -> E5 (E5a+E5b)
                                              #E06 -> E6
                            },
# 'START OF FREQ RMS'    =>  '3X,A1,I2,54X',
# 'NORTH / EAST / UP'    =>  '3F10.2,30X',
# 'RMS VALUES NOAZI'     =>  "3X,A5,${m}F8.2",
# 'RMS VALUES DAZI'      =>  "F8.1,${m}F8.2",
# 'END OF FREQ RMS'      =>  '3X,A1,I2,54X',
 'END OF ANTENNA'       =>  { fmt => '60X,A20,A20' ,
                              fields => ['eoa'], 
                              eoa => 'END OF ANTENNA' 
                            },
);

my %freq_code = (
    #GPS
    L1    =>  'G01',
    G01   => 'G01',
   'G 1'  => 'G01',
    L2    =>  'G02',
    G02   => 'G02',
   'G 2'  => 'G02',
    L5    => 'G05', 
    G05 => 'G05', 
    #GLONASS
    G1 =>  'R01',
    R01 =>  'R01',
    G2 =>  'R02',
    R02 =>  'R02',
    #Galileo
    E1   => 'E01',
    E01  => 'E01',
    E5a  => 'E05',
    E05  => 'E05',
    E5b  => 'E07',
    E07  => 'E07',
    E5   => 'E08', # E5 = (E5a+E5b)
    E08  => 'E08', # E5 = (E5a+E5b)
    E6   => 'E06',
    E06  => 'E06',
);

sub new {
   my $class = shift;

   my $file   = @_ ? shift : '' ;

   my $skip_freq =  @_ ? 1 : 0;

   my $self  = { fmt => 0 , file => '' };
   
   bless $self,$class;
   
   $self->file($file) if $file;
   
   $self->skip_freq_values(0);
  
   $self->fmt();
   
   if ( $self->file ) {
      $self->read_atx();
   }

   return $self
}

sub fmt {
  my $self = shift;
  my $tag  = @_ ? uc shift : '';
  my $fmt  = @_ ? shift : 0 ;
  if ( ! $self->{fmt} ) {
    $self->{fmt}={};
    foreach my $k ( keys %atx ) {
      $self->{fmt}{$k} = Fortran::F90Format->new( fmt => $atx{$k}{fmt} );
    } 
  }
  
  if ( $tag && ! $fmt ) {
     if ( exists $self->{fmt}{$tag} ) {
        return $self->{fmt}{$tag} 
     }
  }
  elsif ( $tag && $fmt ) {
    if ( ref $fmt eq 'Fortran::F90Format' ) {
      return $self->{fmt}{$tag} = $fmt;
    }
  }  
  return $self->{fmt}
}
 
sub skip_freq_values {
  my $self = shift;
  if ( @_ ) {
    $self->{skip_freq_values} = $_[0] ? 1 : 0;
  }
  return $self->{skip_freq_values}
}

sub file {
  my $self = shift;
  if ( @_ ) {
     if ( -e $_[0] ) {
       $self->{file}= $_[0];
     }
     else {
       die "Antex file $f doesn't exists\n";
     }
  }
  return $self->{file}
}

sub read_atx {
  my $self = shift;  
  my $file = shift || $self->file;
  open( my $fh , "< $file" ) 
       or croak "Couldn't open $file $!\n";
  my ($line,$block) = ('')x2;
  #read header
  $self->read_atx_header($fh);
  #read antenna
  $self->read_atx_antennae($fh);
  
}  

sub list {
  my $self = shift;
  my @list = sort keys %{$self->{antennae}};
  return wantarray ? @list : \@list;
}

sub show {
 my $self = shift;
 my $ant  = shift || 0;
 foreach my $e ( @{$self->{header}} ) {
     foreach my $t (grep !/COMMENT/i, keys %{$e} ) {
        print "{header}{$t} = $e->{$t}\n";
     }
 }

 foreach my $a ( $self->list() ) {
   next if $ant && $a !~ /$ant/;
   foreach my $e ( grep {! /COMMENT/} sort keys %{$self->{antennae}{$a}} ) {
     if ( $e eq 'frequency' ) {
      foreach my $ft ( sort keys %{$self->{antennae}{$a}{$e}} ) {
        foreach my $t ( sort keys %{$self->{antennae}{$a}{$e}{$ft}} ) {
          if ( $t =~ /(north|east|up)/ ) {
            my $vals = $self->{antennae}{$a}{$e}{$ft}{$t};
            print "{antennae}{$a}{$e}{$ft}{$t} = $vals\n";
            next;
          } 
          foreach my $azi ( sort keys %{$self->{antennae}{$a}{$e}{$ft}{$t}} ) { 
             my $vals = $self->{antennae}{$a}{$e}{$ft}{$t}{$azi};
             for( my $i=0; $i< @$vals ; $i++) {
                 print "{antennae}{$a}{$e}{$ft}{$t}{$azi}".
                          " = $vals->[$i]\n";
             }
          }  
        } 
     }
    }
    else {
       print "{antennae}{$a}{$e} = $self->{antennae}{$a}{$e}\n";
    }
  }
  #last if ! $ant;
 }

}


sub vals {
  my $self = shift;
  my %opt = (
     antenna   => '',
     freq_code => '',
     azi_angle => 'NOAZI',
     @_,
  );
  $opt{antenna} = uc $opt{antenna};
  if ( ! exists $self->{antennae}{$opt{antenna}} ) {
    croak "Invalid antenna type $opt{antenna}\n";
  }

  $opt{freq_code} = uc $opt{freq_code};  
  
  if ( ! exists $freq_code{$opt{freq_code}} ) {
    my %seen;
    my @codes = sort grep { ! $seen{$_} } keys %freq_code;
    croak "Invalid frequency code = |$opt{freq_code}| \nvalid codes:\n@codes\n";
  }  
  else {
   $opt{freq_code} = $freq_code{$opt{freq_code}};
  }

  $opt{azi_angle} = uc trim($opt{azi_angle}); 
  my $azi_type = $opt{azi_angle} eq 'NOAZI'                 ? 'noazi':
                 looks_like_number($opt{azi_angle}) &&
                 $opt{azi_angle} >= 0.0             && 
                 $opt{azi_angle} <= 360.0                   ? 'dazi' :
                 croak "Invalid azimuth angle |$opt{azi_angle}|\n".
                 "azi_angle = 'NOAZI' or 0.0 <= azi_angle <= 360.0\n";                                                    

  my $azis = $self->{antennae}
                    {$opt{antenna}}
                    {frequency}
                    {$opt{freq_code}}
                    {$azi_type};

  #print "|$opt{antenna}|$opt{freq_code}|$azi_type|$opt{azi_angle}|\n";
  #print 'hazis ',join "|",keys %{$azis},"\n";
  if ( $opt{azi_angle} ne 'NOAZI' ) {
    ($opt{azi_angle}) = grep { $opt{azi_angle} == trim($_) } keys %$azis; 
  }

  return  $self->{antennae}
                 {$opt{antenna}}
                 {frequency}
                 {$opt{freq_code}}
                 {$azi_type}
                 {$opt{azi_angle}}
  
}

sub read_atx_header {
  my $self = shift;
  my $fh   =  shift;
  my $block = 'header';
  my $line = '';
  while ( $line = <$fh> ) { 
    push @{$self->{$block}},$self->read_atx_line($line); 
    last if $line =~ /END OF HEADER/;
  } 
}
sub read_atx_antennae {
  my $self = shift;
  my $fh   =  shift;
  my $block = 'antennae';
  my ($sline,$line) = '';
  my ($field,$ant_type)=('','');
  while ( $line = <$fh> ) { 
    chomp $line;
    #print "reading|$line\n";
    if ( $line =~ /START OF ANTENNA/ ) {
       $sline = $line; 
       next;
    }
    if ( $line =~ /END OF ANTENNA/ ) {
       $self->add_entry( $line, $ant_type); 
       $self->add_entry( $sline, $ant_type); 
       next;      
    }
    if ( $line =~ /TYPE \/ SERIAL NO/ ) {
       $ant_type = uc unpack 'a20',$line;
       $field = $ant_type =~ /^(BLOCK I|GLONNASS)/ ? 'satellite' : 'receiver';
       my $entry = $self->read_atx_line($line,$field);
       if ( $field eq 'satellite' ) {
          #my $a = sprintf '%-12s',trim($entry->{sat_ant_type});
          #my $a = trim($entry->{sat_ant_type});
          #$ant_type = pack 'A20',"$a $entry->{sat_code_nn} $entry->{sat_code_nnn}";
          $ant_type = join (' ',map { trim($entry->{$_}) } qw(sat_ant_type sat_code_nn sat_code_nnn) );
       }
       $self->add_entry( $line, $ant_type ,$field); 
    }
    elsif ($line =~ /START OF FREQUENCY/ ) {
       $self->read_atx_freq( $fh,$ant_type,$line);
    }
    elsif ( $ant_type ) {
       $self->add_entry( $line, $ant_type); 
    }
  }
}
sub read_atx_freq {
    my $self = shift;
    my $fh   = shift;
    my $ant_type = shift;
    my $line = shift;
    my $entry = {};
    my $block = 'antennae',
    #start of freq, reading first freq code
    $self->add_entry( $line, $ant_type); 
    my $freq = "$self->{antennae}{$ant_type}{sat_sys}".
               "$self->{antennae}{$ant_type}{freq_num}"; 
    #North/East/Up
    $line = <$fh>;
    $entry = $self->read_atx_line($line,$field,$tag);
    my $neu = [ map { trim($_) } @{$entry}{qw/north east up/} ]; 
    @{$self->{antennae}{$ant_type}{frequency}{$freq}}{qw/north east up/}=@$neu;
    #$self->add_entry( $line, $ant_type); 
    while ( $line = <$fh> ) {
      next if $line !~ /END OF FREQUENCY/ && $self->{skip_freq_values};
      chomp $line;
      # print "reading $line\n";
      if ( $line =~ /END OF FREQUENCY/ ) {
        $self->add_entry( $line, $ant_type); 
        last;
      }
      else {
        my $n =  (    $self->{antennae}{$ant_type}{zen2} 
                   -  $self->{antennae}{$ant_type}{zen1} ) 
                / $self->{$block}{$ant_type}{dzen} 
                + 1 ;
        my $tag = 'VALUES DAZI';
        if ( $line =~ /NOAZI/ ) {
           $tag = 'VALUES NOAZI';
           $atx{$tag}{fmt} = $self->fmt_values_noazi($n);
        }
        else { 
           $tag = 'VALUES DAZI';
           $atx{$tag}{fmt} = $self->fmt_values_dazi($n);
        }
        $self->fmt( $tag , Fortran::F90Format->new( fmt => $atx{$tag}{fmt} ) );
        my $vals =[ map { trim($_) } $self->fmt($tag)->read($line) ];
        #print "n= $n dazi values ". scalar @$vals."\n";
        my $azi = trim( shift @$vals );
        my ($f) = $self->read_fields($tag);
        #push @{$self->{antennae}{$ant_type}{frequency}{$freq}{$f}{$azi}},$vals;       
        $self->{antennae}{$ant_type}{frequency}{$freq}{$f}{$azi}=$vals;       
      }
    }

}
sub read_atx_line {
   my $self = shift;
   my $line = shift;
   my $sub_fields = shift || '';
   my $tag = shift || unpack 'x60a20',$line;
   my $entry = {};
   chomp ($line);
   #print "reading $tag ";
   # print unpack('x60a20',$line)."\n"; 
   $tag=trim($tag);
   if ( exists $atx{$tag} ) {
     my @fields = $self->read_fields($tag,$sub_fields); 
     #print "found $tag @fields\n";
     @{$entry}{@fields} = $self->read_vals($line,$tag);
   }
   else {
    carp "couldn't find tag |$tag| in \%atx \n";
   }
   return $entry
}

sub add_entry {
   my $self = shift;
   my $line = shift;
   my $ant_type = shift;
   my $field = shift||'';
   my $tag   = shift || '';
   my $entry = {};
   $entry = $self->read_atx_line($line,$field,$tag); 
   @{$self->{antennae}{$ant_type}}{ keys %$entry} = values %$entry; 
}

sub read_fields {
   my $self = shift;
   my $tag  = shift;
   my $sub_fields = shift || '';

   if ( $sub_fields ) {
     return  @{$atx{$tag}{fields}{$sub_fields}}
   }
   else {
     return  @{$atx{$tag}{fields}}
   }
}

sub fmt_values_noazi { 
   my $self = shift;
   my $m = shift || 19; 
   return "3X,A5,${m}F8.2"
}
sub fmt_values_dazi { 
   my $self = shift;
   my $m = shift ||19; 
   return "F8.1,${m}F8.2"
}
sub read_vals {
   my $self = shift;
   my $line = shift; 
   my $tag  = shift;
   my ( $v,$t );
   my @vals = ();
  #print $line;
  if ( $tag =~ /(VALUES NOAZI|VALUES DAZI)/ ) {
      $self->fmt($tag, Fortran::F90Format->new( fmt => $atx{$tag}{fmt} ) );
  }
  #print "$tag : $line\n";
  @vals = $self->fmt($tag)->read($line);
  #print "$tag : $atx{$tag}{fmt}: @vals\n"; #, $fmt{$tag}->read($line),"\n";
  #}
  #else {
  #  carp "couldn't parse tag |$tag|\n";
  #}
  return @vals
}

1;
