package GPS::Geodesy;
use Carp;
use Exporter;
@ISA=('Exporter');
@EXPORT=qw( xyz2llh xyz2env env_rot rad2deg dd2deg $r2d $Pi distance);
#our $Pi  = 4*atan2(1,1);     # Pi/4  aka 45 degrees
our $Pi = 3.1415926535898;
our $r2d = 180/$Pi;
our $a_e = 6378136.3;               # m IERS Tech Note 13 pg 5
our $f_e = 1/298.257223563;         # km IERS Tech Note 13 pg 5
our $e2 = 2*$f_e  - $f_e*$f_e;

my $deg2rad = $Pi/180;

my  %units = (
      km => 1e-3,
      m  => 1,
      cm => 100,
      mm => 1000
    );

sub distance {
   my %opt = (
       unit => 'km',
       lon1 => 0,
       lat1 => 0,
       lon2 => 0,
       lat2 => 0,
       @_,
   );
   
   if ( ! ( $opt{lon1} && $opt{lat1} 
           && $opt{lon2} && $opt{lat2} ) ) {
     croak "missing one or more lon1,lat1,lon2,lat2\n";
   } 
   $opt{lon1} += 360 if $opt{lon1} < 0.0 ; 
   $opt{lon2} += 360 if $opt{lon2} < 0.0 ; 
   my $r = $a_e * $units{$opt{unit}};
   my ($lon1,$lat1,$lon2,$lat2) = map { $_*$deg2rad } 
                                  @opt{qw/lon1 lat1 lon2 lat2/};
   # Vincenty formula from Wikipedia
   # its precision is enough to obtain the network configuration
   # we are not interested in calculating the exact distance
   my $dlon = $lon2 - $lon1;
   my $dlat = $lat2 - $lat1;
   my $sindlat = sin($dlat);
   my $sindlon = sin($dlon);
   my $cosdlon = cos($dlon);
   my $sin_dlat2 = sin($dlat/2);
   my $sin_dlon2 = sin($dlon/2);
   my $coslat2 = cos($lat2);
   my $coslat1 = cos($lat1);
   my $sinlat1 = sin($lat1);
   my $sinlat2 = sin($lat2);
   my $num = sqrt( ($coslat2*$sindlon)**2 + 
                   ($coslat1*$sinlat2 - $sinlat1*$coslat2*$cosdlon)**2 );
   my $den = $sinlat1*$sinlat2 + $coslat1*$coslat2*$cosdlon;
   
   my $c = atan2($num,$den);

   return $c*$r # result in m by default
}

sub env_rot {
  my ($lat,$lon) = (shift,shift);
  ($lat,$lon) = ($deg2rad*$lat,$deg2rad*$lon);
  my $rot = ();
  my ($sinlat,$coslat)=(sin($lat),cos($lat));
  my ($sinlon,$coslon)=(sin($lon),cos($lon));
  
  $$rot[0][0]=-$sinlon;          
  $$rot[0][1]= $coslon;          
  $$rot[0][2]= 0.0;              
  $$rot[1][0]= -$sinlat*$coslon; 
  $$rot[1][1]= -$sinlat*$sinlon; 
  $$rot[1][2]= $coslat;          
  $$rot[2][0]= $coslat*$coslon;  
  $$rot[2][1]= $coslat*$sinlon;  
  $$rot[2][2]= $sinlat;

  return $rot
 
}

sub xyz2env {
  my $rxyz = shift;
  my $xyz  = shift; # 
  my $cov = @_ ? shift : undef ;  # covariance xyz
  my $rot = @_ ? shift : undef ;  # rotation matrix
  if ( ! $rot ) {
     my ($lat,$lon,$h)=   xyz2llh($rxyz);
     $rot = env_rot($lat,$lon);
  }
  my @nev = (0)x3;
  my ( $i,$j,$k)=(0)x3;
  for my $i ( 0 .. 2 ) {
    for my $j ( 0 .. 2 ) {
      $nev[$i]+=$$rot[$i][$j] * $$xyz[$j];
    }
  }

 my @c = ();

 if ( $cov ) { 
   for $i ( 0 .. 2 ) {
      for $j ( 0 .. 2) {
         $c[$i][$j]=0.0;
         for $k ( 0 .. 2 ) {
            $c[$i][$j]+= $$rot[$i][$k]*$cov->[$k][$j];
         }
      }
   }
   for $i ( 0 .. 2 ) {
      for $j ( 0 .. 2) {
         $cov->[$i][$j]=0.0;
         for $k ( 0 .. 2 ) {
            $cov->[$i][$j]+= $c[$i][$k]*$$rot[$j][$k];
         }
      }
   }

   
 }
 
  return (\@nev,$cov)
}


sub xyz2llh {
  my $xyz = @_ == 3              ? [@_]  : 
            @{$_[0]} == 3        ? shift : 
                                   die "Usage: xyz2llh(  \$x, \$y,\$z  ) \n".
                                       "       xyz2llh( [\$x ,\$y,\$z] ) \n";
  my ($x,$y,$z)=@$xyz;
  my ($z2, $p2, $p, $r, $mu, $phi, $lat, $lon, $h,$n);
  ($x, $y, $z) = ($x/$a_e, $y/$a_e, $z/$a_e);
  $z2 = $z**2;
  $p2 = $x**2 + $y**2;
  $p = sqrt($p2);
  $r = sqrt($p2 + $z2);
  $mu = atan2($z * (1 - $f_e + $e2 / $r), $p);
  $phi = atan2($z * (1 - $f_e) + $e2 * (sin($mu))**3,
              (1 - $f_e) * ($p - $e2 * (cos($mu))**3));
  $lat = $phi * $r2d;
  $lon = atan2($y, $x) * $r2d;
  $n = $a_e/sqrt((1.0 - $e2*sin($lat)**2));
  $h = $a_e * ($p * cos($phi) + $z * sin($phi)
              - sqrt(1 - $e2 * (sin($phi))**2));
  $lon = $lon + 360 if $lon < 0;
  return ( $lat , $lon , $h )
}


sub xyz2llh2 {
  my $xyz = shift;
  my ($x,$y,$z)=@$xyz;
  my $a = $a_e;
  my $f = $f_e;
  my $p = sqrt( $x*$x + $y*$y);
  my $r = sqrt($p*$p + $z*$z);
  my $lon = atan2($y,$x);
  my $lat = atan2($z/$p, 1.0 - $e2 );
  my $n = $a/sqrt(1.0 - $e2*(sin($lat))**2);
  my $h = $p/cos($lat) - $n;
  my $oldh = 0;
  my $iter = 0;
  my $num=$z/$p;
  my $den;
  while ( abs($h-$oldh) > 1e-12 ) {
    last if $iter++ > 5;
    $oldh=$h;
    $den= 1.0 - $e2*$n/($n+$h);
    $lat = atan2($num,$den);
    $n = $a/sqrt((1.0 - $e2*(sin($lat))**2));
    $h = $p/cos($lat) - $n;
  }
  $lat = $lat*$r2d;
  $lon = $lon * $r2d;
  $lon = $lon + 360 if $lon < 0;
  #$h = sprintf("%7.1f",$h);
  return ( $lat , $lon , $h )
}
sub dd2deg {
  my $dd = shift;
  my $deg = int($dd);
  #my $min = int(frac($dd)*60);
  my $min = int(abs($dd-$deg))*60;
  #my $sec = sprintf("%4.1f",($dd - $deg - $min/60.0 )*60);
  my $sec = abs($dd - $deg - $min/60.0 )*60;
  return ($deg, $min, $sec)
}


sub frac {
  my ($int,$f) = split(/\./, sprintf("%f", shift));
  return $f
}
sub rad2deg {
  my $rad= shift;
  return $rad*$r2d;
}
