package GPS::NetMaker;
use Carp;
use GPS::DataTable;
our $a_e = 6378.137; # +- 0.002 km
our $deg2rad = atan2(1,1)/45;
our %units = (
      km => 1,
      m  => 1_000,
      cm => 1_000_00,
      mm => 1_000_000
    );

sub new {
   my $class = shift;
   my $self  = {};
   my %opt = ( @_, );
   bless $self,$class;
   return $self->init(%opt);
}
sub init {
   my $self = shift;
   my %opt = (
       sites        => '',
       common_sites => '',
       max_net      => 36,
       max_snet     => 33,
       sub_net_fields => 'name,lon,lat',
       sub_net_cfmt   =>  '%4s %-9.4f %-9.4f %-9s',
       @_, 
      ); 
  our $points = GPS::DataTable->new(  data       => [],
                                 fields_in  => $opt{sub_net_fields},
                                 cfmt       => $opt{sub_net_cfmt}
                               );
  our $MAX_NET  = $opt{max_net};
  our $MAX_SNET = $opt{max_snet};
  our $NCOMMON = $MAX_NET - $MAX_SNET;
  carp  "Warning number of common sites is 0\n" if $NCOMMON == 0;
  croak "max_snet cannot be bigger than max_net\n" if $NCOMMON < 0;
  $self->{sub_net} = {};
  return $self 
}

sub make_nets {
   my $self = shift;
   my %opt  = (
       sites           => '',
       common_sites    => [],
       prefix          => '',
       num_count       =>  0,
       sub_net         => '',
       @_,
   );
    croak "No data for sites! \n"        if ! $opt{sites};
    carp  "Warning: No data for common sites\n" if ! $opt{common_sites};
    #fix longitud when assigning from external sources
    my $sites  =  $self->fix_lon( points => $opt{sites} );
    my $csites =  $self->fix_lon( points => $opt{common_sites} );
    my $pts=[];
    my $k= $opt{num_count} ? 1 : 'A'; # initialize counter as 
                                      # numeric or alphanumeric
    my $sn = "$opt{prefix}_$k";
    
    $sn = $opt{sub_net}  if $opt{sub_net} ;
    $points->{data} = $sites;
    my ($p,$psn,@snkeys);
    while ( @{$points->{data}} ) {
       #get a copy of the data
       $psn = $points->get_data();
       $p = $self->find_first( points => $psn );
       $psn = $self->closest( source => $p, points => $psn );
       if ( @$psn ) {
         @$psn = @$psn > $MAX_SNET ? @{$psn}[0..$MAX_SNET-1] : @$psn;
         $sn = "$opt{prefix}_$k";
         $self->{sub_net}{$sn} = $points->add_col( 
                                           data => $psn,
                                           field => 'sub_net' , 
                                           value => $sn
                                 );
         push @snkeys,$sn;
         $k++;
       }
       #remove the entries comparing the name only
       $points->remove( entries => $psn ,name => 1); 
    }    
    if ( @{$opt{common_sites}} ) {
       my $cs = $opt{common_sites};
       foreach my $sn ( @snkeys ) {
          $psn=[];
          $p = $self->get_baricenter( sub_net => $sn );
          $psn = $self->closest( source => $p, points => $cs );
          @$psn = @$psn > $NCOMMON ? @{$psn}[0..$NCOMMON-1] : @$psn;
          # obtain a copy of the common points otherwise the sub_net field
          # will be overwritten since is passed by reference
          $psn = $points->get_data( data => $psn);
          $psn = $points->add_col( data => $psn, 
                                   field => 'sub_net' , 
                                   value => $sn
                                  );
          #add common points to the sub-network
          push @{$self->{sub_net}{$sn}}, @$psn; 
       }
    } 
    return $self 
}
sub get_names {
   my $self = shift;
   my %opt = ( 
         sub_nets => 1,
         sites    => 0,
         sorted   => 0,
         @_, 
      );
   my @snames = sort keys %{$self->{sub_net}};
   my @names;
   my @sites;
   if ( $opt{sites} ) {
      foreach my $sn ( @snames ) {
          @sites = map { $_->{name} } @{$self->{sub_net}{$sn}};
          @sites = sort @sites if $opt{sorted};
          push @names,join ' ',($sn,@sites) ;  
      } 
   }
   else { 
     @names = @snames; 
   }
   return @names
}
sub print {
  my $self = shift;
  my %opt = ( 
      sub_net    => '',
      fields     => 'name,lon,lat,sub_net',
      @_, 
     );
  my $sn = $opt{sub_net};
  delete $opt{sub_net};
  if ( exists $self->{sub_net}{$sn} ) {
    $points->{data} = [];
    $points->{data} = $self->{sub_net}{$sn};
    $points->sort ( field => 'name'  );
    $points->print( %opt );
  }
}
sub sub_net {
  my $self = shift;
  my $sn  = shift || '';
  if ( exists $self->{sub_net}{$sn} ) {
    $points->{data} = [];
    $points->{data} = $self->{sub_net}{$sn};
    #$points->sort ( field => 'name'  );
    return $points->get_col( col => 'name' );
  }
  else {
    carp "Sub-net $sn doesn't exists !\n";
  }
}

sub distance {
   my $self = shift;
   my %opt = (
       unit => 'km',
       lon1 => 0,
       lat1 => 0,
       lon2 => 0,
       lat2 => 0,
       @_,
   );
   if ( ! ( $opt{lon1} && $opt{lat1} 
           && $opt{lon2} && $opt{lat2} ) ) {
     croak "missing one or more lon1,lat1,lon2,lat2\n";
   } 
   $opt{lon1} += 360 if $opt{lon1} < 0.0 ; 
   $opt{lon2} += 360 if $opt{lon2} < 0.0 ; 
   my $r = $a_e * $units{$opt{unit}};
   my ($lon1,$lat1,$lon2,$lat2) = map { $_*$deg2rad } 
                                  @opt{qw/lon1 lat1 lon2 lat2/};
   # Vincenty formula from Wikipedia
   # its precision is enough to obtain the network configuration
   # we are not interested in calculating the exact distance
   my $dlon = $lon2 - $lon1;
   my $dlat = $lat2 - $lat1;
   my $sindlat = sin($dlat);
   my $sindlon = sin($dlon);
   my $cosdlon = cos($dlon);
   my $sin_dlat2 = sin($dlat/2);
   my $sin_dlon2 = sin($dlon/2);
   my $coslat2 = cos($lat2);
   my $coslat1 = cos($lat1);
   my $sinlat1 = sin($lat1);
   my $sinlat2 = sin($lat2);
   my $num = sqrt( ($coslat2*$sindlon)**2 +
                   ($coslat1*$sinlat2 - $sinlat1*$coslat2*$cosdlon)**2 );
   my $den = $sinlat1*$sinlat2 + $coslat1*$coslat2*$cosdlon;

   my $c = atan2($num,$den);

   return $c*$r # result in m by default

}
sub fix_lon {
   my $self = shift;
   my %opt  = (
       sub_net => '',
       points  => [],
       @_,
   );
   my $pts=[]; my $sn;
   $sn = $opt{sub_net} if $opt{sub_net};
   if ( exists $self->{sub_net}{$sn} ) {
        @$pts = @{$self->{sub_net}{$sn}};
   }
   @$pts = @{$opt{points}} if @{$opt{points}};
   foreach my $h ( @$pts ) {
      $h->{lon} += 360 if $h->{lon} < 0.0;
   }
   return $pts
}
sub closest {
    my $self = shift;
    my %opt = (
       sub_net  => '', # name of the sub-net
       rev      => 0,  # reverse sorting
       inc_src  => 0,  # include source
       source   => [],
       points   => [], # points to consider 
       max_dist => 0,
       min_dist => 0,
       @_,
    );
    croak "You need to supply a source ! \n" if ! @{$opt{source}}; 
    croak "You need to supply points ! \n"   if ! @{$opt{points}};
    my $pts=[]; 
    my $sn='';
    my ($src) = @{$opt{source}};
    push @$pts,$src  if $opt{inc_src};
    
    $sn = $opt{sub_net}  if $opt{sub_net} ;
    if ( exists $self->{sub_net}{$sn} ) {
        @$pts = @{$self->{sub_net}{$sn}};
    }
    
    $pts = $points->get_data( data => $pts );
    
    #fix longitud when assigning from external sources
    $pts = $self->fix_lon( points => $opt{points} ) if @{$opt{points}};
  
    my $dist = [];
    foreach my $p ( @$pts ) {
         $d_sp = $self->distance( 
                           lon1 => $src->{lon},
                           lat1 => $src->{lat},
                           lon2 => $p->{lon},
                           lat2 => $p->{lat},
                 );
         push @$dist, [ $d_sp, $p ]; # prepare for Schwartzian transform
    }
    # distance is in first element of the array, but needs de-referencing
    my %sort_dist = ( ascending  => sub { $a->[0] <=> $b->[0]},
                      descending => sub { $b->[0] <=> $a->[0]}
                    );
    my $comp_func = $opt{rev} ? $sort_dist{descending} : $sort_dist{ascending};
    @$dist = sort $comp_func @$dist;
    #filter min and max distances
    @$dist = grep { $_->[0] >= $opt{min_dist} } @$dist;
    #do only if we have a max distance
    @$dist = grep { $_->[0] <= $opt{max_dist} } @$dist if $opt{max_dist};
    #print map { "$_->[0] $_->[1]{name}\n"} @$dist;
    # retrive only the original entry stored in the second element.
    @$dist = map { $_->[1] } @$dist;
    return $dist 
}

sub get_baricenter {
  my $self = shift;
  my %opt = (
      sub_net => '',
      points  => [],
      @_,
  );
  my $pts=[]; 
  my $sn = $opt{sub_net};
  if ( exists $self->{sub_net}{$sn} ) {
      $pts = $self->{sub_net}{$sn} ;
  }
  $pts = $opt{points} if @{$opt{points}};
  my $n = @$pts; # number of elements
  my $b = [{name => 'B_C' ,lon => 0 ,lat => 0 , sub_net => $sn }];
  foreach my $p ( @$pts ) {
    $p->{lon} += 360 if $p->{lon} < 0.0;
    $b->[0]{lon} += $p->{lon};
    $b->[0]{lat} += $p->{lat};
  }
  $b->[0]{lon} /= $n;
  $b->[0]{lat} /= $n;
  return $b
}

sub get_center {
  my $self = shift;
  my %opt = (
      sub_net => '',
      points  => [],
      @_,
  );
  my $pts=[]; my $sn='';
  $sn = $opt{sub_net} if $opt{sub_net};
  if ( exists $self->{sub_net}{$sn} ) {
       $pts = $self->{sub_net}{$sn};
  }
  $pts = $opt{points} if @{$opt{points}};
  my $lim = $self->get_limits( points => $pts );
 
  my $c = [{lon => 0 ,lat => 0 , name => "C_$sn" }];
  $c->{lon}  = ($lim->{lonmin}+$lim->{lonmax})/2;
  $c->{lat}  = ($lim->{latmin}+$lim->{latmax})/2;
  
  return $c
}
sub get_limits {
  my $self = shift;
  my %opt = (
      sub_net => '',
      points  => [],
      @_,
  );
  my $pts=[]; my $sn='';
  $sn = $opt{sub_net} if $opt{sub_net};
  if ( exists $self->{sub_net}{$sn} ) {
      $pts = $self->{sub_net}{$sn} ;
  }
  $pts = $opt{points} if @{$opt{points}};
  $pts = $self->fix_lon( points => $pts );
  $points->{data} = $pts;  
  my $n = @$pts;  
  my ( $lonmin,$latmin,$lonmax,$latmax)=(0)x4;
  if ( $n ) {  
    $points->sort( field => 'lon' ) if $n > 1; 
    $lonmin = $points->{data}[0]{lon}; 
    $lonmax = $points->{data}[$n-1]{lon}; 

    $points->sort( field => 'lat' ) if $n > 1; 
    $latmin = $points->{data}[0]{lat}; 
    $latmax = $points->{data}[$n-1]{lat}; 
  }
  else {
    carp "No points to get limits\n"; 
  }  
  return { lonmin => $lonmin,
           lonmax => $lonmax,
           latmin => $latmin,
           latmax => $latmax  } 
}

sub find_first {
   my $self = shift;
   my %opt = (
      sub_net => '',
      points  => [],
      @_,
  );
  my $pts=[]; my $sn ='';
  $sn = $opt{sub_net}  if $opt{sub_net} ;
  if ( exists $self->{sub_net}{$sn} ) {
     $pts = $self->{sub_net}{$sn} ;
  }
  $pts = $opt{points} if @{$opt{points}};
  my $first=[];
  if (  @$pts ) {
     my $c = $self->get_baricenter( points => $pts );
     $first = $self->closest ( source => $c , points => $pts );
     @$first = @{$first}[0 .. 0]; 
  }
  return $first
}












