package GPS::Orb;
use Carp;
#use Data::Dumper;
use Net::FTP;
use GPS::IGS::Sum;
use GPS::DATES qw( get_date cal2sec sec2cal gps2utc jd2cal m2month );
use GPS::PRN_GPS;
use GPS::Defaults;
use GOA::Util;
use File::Copy;
use File::Path;
use File::Basename;
use Cwd;
my %par = get_defaults();

sub new {
   my $class = shift;
   my %opt = (  type => '',
				frame => 'itrf05',
				date => '',
				hh => '',
				@_,
               );
   my $self = { type => '', 
                date => '', 
				hh=>'', 
                t1 => undef,
                t2 => undef,
				interp_300 => '',
				cache_age => 0,   #days
				log_h => undef,
                rate => 300,
				goa_files => [qw(
				                  GPS_antcal_ref
								  tpeo.nml
								  Shadow
								  pos
								  TDPfile
                                  frame
				               )],
			  };
   bless $self, $class;
   $self->log_h;
   $self->date($opt{date});
   $self->config("$ENV{GPS_HOME_DIR}/etc/orbits_config.txt");
   $self->type($opt{type});
   $self->cache_age($opt{cache_age});
   $self->rate($opt{rate});
   $self->t1($opt{t1});
   $self->t2($opt{t2});
   $self->hh($opt{hh});
   
   return $self
}

sub igs_bad_sats {
   my $self = shift;
   my $file = shift;
   my $pg   = shift || 'gps';
   my $flags = uc shift || 'X';
   my $s = GPS::IGS::Sum->new($file);
   my %seen=();
   my @sats = grep { ! $seen{$_}++ }
              map { $_->{$pg} } 
              grep { $_->{flags} =~ /$flags/i } $s->sats($self->date);
   
    return @sats
}

sub t1 {
   my $self = shift;
   if ( $_[0] ) {
     my $t = shift;
     $self->{t1} = get_date($t,'secj2000');
   }
   return $self->{t1};
}
sub t2 {
   my $self = shift;
   if ( $_[0] ) {
     my $t = shift;
     $self->{t2} = get_date($t,'secj2000');
   }
   return $self->{t2};
}

sub rate {
   my $self = shift;
   if ( @_ ) {
     my $r = shift;
     $self->{rate} = $r =~ /^\d+$/ && $r >= 1 ? $r : 300;
   }
   return $self->{rate};
}

sub log_h {
  my $self = shift;
  
  if ( ! defined $self->{log_h} ) {
    my $f = shift || 'get_orbits_pl.log';
    open ( my $log_h , '>', $f ) or croak "Couldn't open $f for writing $!\n";
    $self->{log_h} = $log_h;
  }

  return $self->{log_h}
}

sub cache_age {
  my $self = shift;
  if ( @_ ) {
    my $age = shift;
	if ( $age >= 0 ) {
      $self->{cache_age} = $age;
	}
	else {
	  $self->print_log("Negative cache_age, cache_age set to 0\n");
	}
  }
  return $self->{cache_age}
}

sub interp_300 {
   my $self = shift;
   if ( @_ ) {
      $self->{interp_300} = shift ;
   }
   return $self->{interp_300}
}

sub goa_files {
 my $self = shift;
 @files = @{$self->{goa_files}};
 if ( $self->type =~ /(igs|fid)/ ) {
   @files = grep { ! /XFILE/ } @files;
 }
 return wantarray ? @files : \@files;
}

sub hh {
  my $self = shift;
  if ( @_ ) {
    my $hh = shift;
    if ( $hh =~ /^(00|06|12|18)$/ ) {  
      $self->{hh} = $hh; 
    }
  }
  return $self->{hh}
}

sub date {
   my $self = shift;
   if ( @_ ) {
     my $date = shift;
     my ($y,$m,$d,$hh) = get_date($date);
     $self->{date} = "$y-$m-$d";
     $self->hh($hh);
     if ( $self->{date} !~ /^\d{4}-\d{2}-\d{2}$/ ) {
        croak "Invalid date $date in date\n";
     }
   }
   elsif ( $self->{date} ) {
      return $self->{date}
   }
   else {
     croak "orbit date is undefined!\n";
   }
}

sub type {
  my $self = shift;
  my $t='';
  if ( @_ ) {
     $t=lc shift; 
     croak "Invalid type |$t| \n" if ! exists $self->{products}{$t};
     $self->{type} = $t; 
  }
  croak "No product type assigned\n" if ! $self->{type};
  return $self->{type}
}

sub products {
  my $self = shift;
  my $t = $self->type(@_);
  my @p =  $self->fill(@{$self->{products}{$t}});
  return wantarray ? @p : \@p;
}

sub save2archive {
  my $self = shift;
  my ($archive_dir) = $self->archive;
  if ( ! -d $archive_dir ) {
    mkpath($archive_dir) or die "Couldn't make dir $archive_dir $!\n";
  }
  my $here = cwd;
  my ($i,$ok)=(0,0);
  my @files = map { basename $_ } $self->products;
  foreach my $f (@files) {
    if ( -e "$here/$f" ) {
      copy("$here/$f","$archive_dir/$f") ; 
	  if ( -e "$archive_dir/$f" ) {
	    $i++;
	  }
	  else {
	    $self->print_log( "Couldn't copy $f to $archive_dir\n" );
	  }
	}
  }
  $ok = 1 if $i == @files;
  return $ok
}

sub copy_from_archive {
  my $self = shift;
  my ($archive_dir) = $self->archive;
  if ( ! -d $archive_dir ) {
    $self->print_log( "Can't copy from $archive_dir, dir doesn't exists! $!\n");
	return 0
  }
  my $here = cwd;
  my @files = map { basename $_ } $self->products;
  my $i=0;
  my $ok =0;
  foreach my $f ( @files ) {
    if ( -e "$archive_dir/$f" ) {
      copy("$archive_dir/$f","$here/$f");
	  if ( -e "$here/$f" ) {
	     $i++;
	   }
	   else {
	     $self->print_log("Couldn't copy $f from $archive_dir $!\n");
	   }
	}
  }
  $ok = 1 if $i == @files;
  return $ok
}


sub archive {
  my $self = shift;
  my ($ac) = split(/_/,$self->type);
  my $t = $self->type =~ /(rapid|final|ultra|hr)/ ? $1 : '';
  if (  exists $self->{archive}{$ac}{$t} ) {
    return $self->fill($self->{archive}{$ac}{$t});
  }
  else {
    $self->print_log( "Undefined [archive $ac ] for $t in config\n");
  }
}

sub fill {
  my $self = shift;
  my @lines = @_;
  my $cal = $self->date;
  my $hh  = $self->hh;
  my $yyyy = unpack 'a4',$cal;
  my $yy   = unpack 'x2a2',$yyyy;
  my $doy = (split (' ',get_date($cal,'doy')))[1];
  my $epoch= get_date($cal,'epoch');
  my $gpsw = get_date($cal,'gpsw');
  my $gdow = get_date($cal,'gdow'); #gps day of week starts on sundays
  foreach my $f ( @lines ) {
          $f =~  s/\{yyyy\}/$yyyy/g;
          $f =~  s/\{yy\}/$yy/g;
          $f =~  s/\{doy\}/$doy/g;
          $f =~  s/\{cal\}/$cal/g;
          $f =~  s/\{epoch\}/$epoch/g;
          $f =~  s/\{frame\}/$frame/g;
          $f =~  s/\{wwww\}/$gpsw/g;
          $f =~  s/\{d\}/$gdow/g;
          $f =~  s/\{hh\}/$hh/g;
  }
  return  @lines;
} 

sub config {
 my $self = shift;
 my $f = shift;
 open(my $fh, '<', $f ) or die "Couldn't open $f $!\n";
 my ($var1,$var2, $val,$href );
 while ( <$fh> ) {
   chomp;                  # no newline
   s/#.*//;                # no comments
   s/^\s+//;               # no leading white
   s/\s+$//;               # no trailing white
   next unless length;     # anything left?
   last if /\[end\]/i;
   if ( $_ =~/\{([A-Z_]+)\}/ ) {
	  my $d = $1;
	  if ( exists $par{$d} ) {
		  my $v = $par{$d};
          $_ =~ s/\{$d\}/$v/g;
	  }
   }
   if ( /^\[(.*)\]/ ) {
      ($var1,$var2) =  split ' ',lc $1;
	  $val = undef;
	  $self->{$var1}{$var2} = $val;
      next;
   }
   if ( /\s*=\s*/ ) {
     my ($hvar, $hvalue) = split(/\s*=\s*/, $_, 2);
     $hvalue =~ s/^('|")//;
     $hvalue =~ s/('|")$//;
	 $self->{$var1}{$var2}{$hvar}=$hvalue;
   }
   else {
	 push @{$self->{$var1}{$var2}},$_;
   }
 }
 close($fh);
 return $self
}

sub source {
  my $self = shift;
  return wantarray ? %{$self->{source}} : $self->{source}
}

sub ftp {
   my $self = shift;
   my %source = $self->source;
   my %opt = (  ftp_passwd => "$ENV{USER}\@geology.cwu.edu",
                force => 0,
				@_,
              );
   my $ok = 0;
   my $log_h = $self->{log_h};
   my $type = $self->type; #save for later;
   my ($ac) = split(/_/,$type);
   foreach my $s ( sort grep { /^$ac/ } keys %source ) {
      my $host = $source{$s}{host};
      my ($dir)  = $self->fill($source{$s}{dir});
      print $log_h "Connecting to $s host:$host dir:$dir\n";
      my $ftp = Net::FTP->new ( $host,
                                Timeout => 60,
                                Debug   => 0,
                                Passive => 1
                               )
                or print $log_h "Can't connect to  $host $@\n";
      next if ! $ftp; 

      $ftp -> login ( "anonymous" , $opt{ftp_passwd} )
              or print $log_h  "Couldn't login into $host\n";
      $ftp -> binary ()
              or print $log_h  "Couldn't set type to binary in $host\n";
      $ftp -> cwd ( $dir )
                 or print $log_h  "Couldn't change to dir $dir\n";
       
      $ok = 0;
	  my @fnames = $self->products;
      if ( exists $source{products} ) {
	    @fnames = $self->products($source{products});
	  }
      foreach my $f ( @fnames ) {
          my $local_file = basename $f;
          print $log_h "Getting $f : ";
          if ( -e $local_file && ! $opt{force} ) {
             $ok++;
             print $log_h " got it\n";
             next;
          }

          $local_file = $ftp -> get ( "$dir/$f" );
          
          if ( defined $local_file ) {
             $ok ++;
             print $log_h " got it\n";
          }
          else {
             print $log_h "Couldn't get $f \n";
         }
      }
     
      $ftp -> quit()
              or print $log_h "Couldn't quit from $host\n";
     
      $ok = $ok == @fnames ;
      
      last if $ok;
   }
   
   #close $log_h; 
   #restore type in case it changed due to different sources
   $self->type($type) if $type ne $self->type;  
   return $ok
}
sub save2cache {
  my $self = shift;
  my $cal = $self->date;
  my $type = $self->type;
  my $cache_dir = "$par{CACHE_DIR}/orbits/$type/$cal";
  if ( ! -d $cache_dir ) {
    mkpath($cache_dir);
  }
  my $here=cwd;
  my ($i,$ok) =(0,0);
  my @goa_files = $self->goa_files;
  foreach my $f ( $self->goa_files ) {
    if ( -e "$here/$f" ) {
      copy ("$here/$f","$cache_dir"); #/$f"); 
	  if ( -e "$cache_dir/$f" ) {
	    $i++;
	  }
	  else {
  	    $self->print_log("Couldn't copy $f to $cache_dir $!\n");
	  }
	}
  }
  $ok = $i == @goa_files;
  return  $ok
}
sub print_log {
  my $self = shift;
  my $log_h = $self->log_h;
  print $log_h @_;
}
sub copy_from_cache {
  my $self = shift;
  my $cal = $self->date;
  my $type = $self->type;
  my $cache_dir = "$par{CACHE_DIR}/orbits/$type/$cal";
  if ( ! -d $cache_dir ) {
    return 0
  }
  my $here=cwd;
  my ($i,$ok) =(0,0);
  my @goa_files = $self->goa_files;
  foreach my $f ( @goa_files ) {
    if ( -e "$cache_dir/$f" ) {
      copy ("$cache_dir/$f","$here/$f") or
	    $self->print_log("Couldn't copy $f from $cache_dir $!\n");
	   $i++ if -e "$here/$f";
	}
  }
  $ok=1 if $i == @goa_files;
  return $ok
}

sub clear_cache {
   my $self = shift;
   my $age = shift || $self->{cache_age};
   my $type = $self->type;
   my $cache_dir = "$par{CACHE_DIR}/orbits/$type";
   if ( ! -d $cache_dir ) {
      return 0
   }
   opendir(my $dh,$cache_dir) or die "Couldn't open $cache_dir $!\n";
   my @dirs = grep { /^\d{4}-\d{2}-\d{2}$/ } readdir($dh);
   foreach my $d ( @dirs ) {
      if ( -M "$cache_dir/$d" > $age ) {
	    rmtree("$cache_dir/$d");
	  }
   }
}

sub jgp2goa_files {
   my $self = shift;
   my @files = map { basename($_) } $self->products;
   foreach my $f ( @files ) {
	 if ( $f =~ /\.x\./ ) {
	   `zcat $f > XFILE`;
	 }
     if ( $f =~ /\.pos\./ ) {
	    `zcat $f > pos`;
		symlink('pos','peci');
	 }
	 if ( $f =~ /\.ant\./ ) {
	    `zcat $f > GPS_antcal_ref`;
	 }
     if ( $f =~ /\.eo\./ ) {
	    `zcat $f >  tpeo.nml`;
	 }
     if ( $f =~ /\.shad\./ ) {
	    `zcat $f > Shadow`;
	 }
     if ( $f =~ /\.frame\./ ) {
       `zcat $f > frame`;
     }
     if ( $f =~ /\.tdp\./ ) {
	    my $lines;
	    open( my $sh,'>',\$lines);
        print $sh grep { ! /YAW_RATE|SAT BIAS/ } `zcat $f`;
		my $tv2c = "tv2c.pl - "; 
		if ( $self->type =~ /_hr/ ) {
		   my $r = 30;
		   $r = 300 if  $self->interp_300;
           $tv2c = "tv2c.pl -y - | bound_tv_clocks -t $r "; 
		}
		system("zcat $f | $tv2c > TDPfile");
		open(my $fh,">> TDPfile") or die "Couldn't open TDPfile $!\n";
		print $fh $lines;
		close $fh;
	 }
   }
}

sub code2goa_files {
  my $self = shift;
  my $type = $self->type;
  my $date = $self->date;
  my @fnames =  $self->products;

  my $igs = $type =~ /final/ ? 'igs' :
            $type =~ /rapid/ ? 'igr' :
			$type =~ /ultra/ ? 'igu' :
			$type =~ /pred/  ? 'igp' : die "Unknown type $type\n";
  foreach my $f ( map { basename $_ } @fnames ) {
     my $g = $f;
	 $g =~ s/COD/$igs/;
	 $g =~ s/SUM/sum/;
	 $g =~ s/EPH(_R|_Pi|)/sp3/;
	 $g =~ s/ERP(_R|_Pi|)/erp/;
	 $g =~ s/CLK(_R|_05S|)/clk/;
     if ( $type =~ /final/ && $g =~ /CLK.Z/ ) {
       next if  (grep { /CLK_05S/ } @fnames);
     }
  	 rename($f,$g);
     if ( $type ne 'code_final' ) {
	   `compress $g`;
	 }
  }
  my $igs_type = $type;
  $igs_type =~ s/code/igs/;
  $self->type( $igs_type );
  $self->igs2goa_files;
  $self->type( $type );

}

sub igs2gps_antcal_ref {
  my $self=shift;
  my $sp3 =shift || croak "No sp3 file input !\n";
  open(my $fh, '<', $sp3 ) or croak "Couldn't open $sp3 for reading $!\n";
  my ($pcv,$frame);
  while( <$fh> ) {
    if ( $. == 1 ) {
      $frame = unpack('x46a5',$_);
    }
    next if ! /PCV:(IGS\d\d_\d\d\d\d)/;
	$pcv=lc $1;
	last;
  }
  close $fh;
  $pcv="igs05_1402" if ! $pcv;
  my $antcal ="SAT $pcv.atx \n";
     $antcal.="GRN $pcv.atx \n";
  open(my $fh,'>','GPS_antcal_ref') or croak "Couldn't open GPS_antcal_ref for writing $!\n";
  print $fh $antcal;
  close $fh;
  
  open(my $fh,'>','frame') or croak "Couldn't open frame for writing $!\n";
  print $fh $frame;
  close $fh;

  return $antcal
}
sub igs2goa_files {
  my $self = shift;
  my $date = $self->date; #$opt{date};
  my @fnames_z = $self->products;
  my $here = cwd;
  my @fnames   =  map { s/\.Z$//; "$here/$_" }  @fnames_z; 
    
  #my ($bdate,$adate) = get_ba_dates($date);
   
  `gzip -df @fnames_z 2> /dev/null `; 
    
  my @sp3    = grep /\.sp3$/ ,@fnames; 
  
  my ($eci,$tpnml,$shad,$tdp,$ant_cal)  = map { "$here/$_" } qw( pos tpeo.nml Shadow TDPfile GPS_antcal_ref );
    
  $self->igs2gps_antcal_ref($sp3[0]);
    
  my $bdate = get_date($date,'char',-3);
  my $edate = get_date($date,'char', 3);
  $ENV{TPNML} = $tpnml;
  #system("tp_nml $bdate $adate 1> $tpnml 2> tp_nml.err");
  `bula2tpnml -day1 $bdate -day2 $edate 1> $tpnml 2> /dev/null `; #bula2tpnml.err`; 
  
  my $EpochEO = eo_epoch($date);
  #eop($EpochEO);
  open ( my $fh, '>>' , $tpnml ) or die "Unable to open $tpnml $!\n";

  print $fh <<EOT
  \$earth_orientation
         
         EOTIDE = \'IERS03\'
         EpochEO = \"$EpochEO\"
         
  \$end
EOT
;
   
   
  my $start =  get_date($date,'secj2000');
  my $end   =  get_date($date,'secj2000',1);
    

  $self->spx2eci( $start, $end, \@sp3, $eci );

  my @clk = $self->type =~ /ultra/ ? @sp3 : glob( "*.clk");

  $self->clk2tdp($start,$end,\@clk,$tdp);

  dummy_shadow($shad,$date);
    
  #     tdp_gps_phase_ctr.pl -igs -o tdp_gps_pc
  #     cat tdp_gps_pc >> $tdp
  #     tv2c.pl $tdp.$$  > $tdp
  system("
       gps_sat_offset -ac igs -date $date -tdp $tdp -ant GPS_antcal_ref -abs_pc   
      \\rm -f $tdp.$$ @fnames
  ");
    
}

sub spx2eci {
    my $self = shift;
    my $start = shift ; # get_date($cal,'secj2000',-0.125);
    my $end   = shift || $start; # get_date($cal,'secj2000',1.125);
    my $sp3   = shift;
    my $eci   = shift || 'peci';
    my $here   = cwd;   

    my $tmp_dir = "$here/spx2eci_tmp.$$";

    mkdir $tmp_dir or die "Coulnd't make $tmp_dir $!\n";

    chdir $tmp_dir or die "$!\n";

    `spx2eci @$sp3 2> spx2eci.err`;

    $self->filter_eci_files($start,$end); 

    `cat gps*.eci > $eci 2> cat_gps_eci.err `;

    `rm -rf $tmp_dir`;

     

     chdir $here ;
}

sub ba_flinn {
   my %opt = (
       sdate   => '',
       edate   => '',
       type    => '',
       latest  => 0,
       gwdelay => 2,
       @_,
   );
   my $cal   = get_date('today','cal');
   my $yesterday = get_date($cal,'jd',-1) ;
   my $gpsw  = get_date($cal,'gpsw') - $opt{gwdelay} ;
   my $bd    = get_date($opt{sdate},'jd');
   my $ad    = get_date($opt{edate},'jd');
   my $gad   = get_date($opt{edate},'gpsw');
   my $atype  = $opt{type};
   $bd--; 
   $ad++; 
   $opt{latest} =  $atype eq 'final' &&  $gad <   $gpsw      ?  0 : 
                   $atype eq 'final' &&  $gad ==  $gpsw      ?  1 : 
                   $atype eq 'final' &&  $gad >   $gpsw      ? -1 : 
                   $atype eq 'rapid' &&  $ad  <   $yesterday ?  0 : 
                   $atype eq 'rapid' &&  $ad  ==  $yesterday ?  1 : 
                                                               -1 ;
                   
   if ( $opt{latest} < 0 ) {
       $ad--;
       #carp "No all products available for igs to flinn conversion\n";
   }
   if ( $opt{latest} >= 0 ) {
        $atype = 'rapid' if $atype eq 'final' ;
   }
   
   return ($bd,$ad,$atype)
   
}

sub dummy_shadow{
  my $f = shift;              # name of shadow file
  my $cal = shift || '';
  my $here = cwd;
  #if ( -e "$par{CACHE_DIR}/orbits/jgp_rapid/$cal/Shadow" ) {
  #  copy("$par{CACHE_DIR}/orbits/jgp_rapid/$cal/Shadow",$here);
  #}
  #else {
    open(F, ">$f") or die "Can't open $f $!\n";
    print F "                                      \n"; # enough blanks for compress
    print F "NUMBER OF GPS SATELLITES IN ECLIPSE: 0\n";
    close F;
  #}
}

sub eo_epoch {
  my $cal = shift;
  my $start = get_date($cal,'secj2000',-3);
  my $end   = get_date($cal,'secj2000',3);
  my $eo = ($start+$end)/2;
  $eo = `gps2utc $eo`; chomp $eo; 
  my $d = sprintf "%4d-%02d-%02d:%02d:%02d:%7.5f",sec2cal($eo );
  return get_date($d,'eop')
}

sub eop {
  my $EpochEO = shift;
  open(my $fh, ">> tpeo.nml") or die "Couldn't open tpeo.nml for writing $!\n";
  my $eo ="\n";
  $eo .= '$earth_orientation'."\n";
  $eo .= qq(
 !        Received from Jim Ray on 10/18/2000
 !

    EpochEO = \"$EpochEO\"
    Extended_UT1_model= .true.
    Extended_XPM_model= .true.
    Extended_YPM_model= .true.
! Ext. UT1 parameter nominals (units = seconds of time)    ->     IERS
  UMD0 =  -7.20e-6,  16.17e-6,    ! U M2 COS ,  U M2 SIN        COS, -SIN
          -0.04e-6,   7.59e-6,    ! U S2 COS ,  U S2 SIN        COS, -SIN
          -1.54e-6,   3.80e-6,    ! U N2 COS ,  U N2 SIN        COS, -SIN
          -0.38e-6,   1.96e-6,    ! U K2 COS ,  U K2 SIN        COS, -SIN
           8.64e-6,  17.71e-6,    ! U K1 COS ,  U K1 SIN        SIN,  COS
         -12.10e-6, -16.05e-6,    ! U O1 COS ,  U O1 SIN       -SIN, -COS
          -2.86e-6,  -5.16e-6,    ! U P1 COS ,  U P1 SIN       -SIN, -COS
          -2.45e-6,  -5.03e-6     ! U Q1 COS ,  U Q1 SIN       -SIN, -COS
! Ext. XPM parameter nominals (units = radians)
  XMD0 = -.1357e-09,     1.6000e-09,      ! X M2 COS ,  X M2 SIN
          .3103e-09,      .7030e-09,      ! X S2 COS ,  X S2 SIN
         -.0630e-09,      .2763e-09,      ! X N2 COS ,  X N2 SIN
          .0824e-09,      .1745e-09,      ! X K2 COS ,  X K2 SIN
         -.7369e-09,      .3782e-09,      ! X K1 COS ,  X K1 SIN
          .6448e-09,     -.2376e-09,      ! X O1 COS ,  X O1 SIN
          .2424e-09,     -.1212e-09,      ! X P1 COS ,  X P1 SIN
          .1261e-09,     -.0291e-09       ! X Q1 COS ,  X Q1 SIN
! Ext. YPM parameter nominals (units = radians)
  YMD0 =  .9502e-09,     -.1794e-09,      ! Y M2 COS ,  Y M2 SIN
          .4218e-09,     -.2860e-09,      ! Y S2 COS ,  Y S2 SIN
          .1600e-09,     -.0533e-09,      ! Y N2 COS ,  Y N2 SIN
          .1067e-09,     -.0873e-09,      ! Y K2 COS ,  Y K2 SIN
         -.3782e-09,     -.7369e-09,      ! Y K1 COS ,  Y K1 SIN
          .2376e-09,      .6448e-09,      ! Y O1 COS ,  Y O1 SIN
          .1212e-09,      .2424e-09,      ! Y P1 COS ,  Y P1 SIN
          .0291e-09,      .1261e-09       ! Y Q1 COS ,  Y Q1 SIN

);
  $eo .= '$end'."\n";


 print $fh $eo;  

}

sub filter_eci_files {
    my $self = shift;
    #my $cal = get_date( shift, 'cal');
    #my $start =  get_date($cal,'secj2000',-0.125);
    #my $end   =  get_date($cal,'secj2000',1.125);
    my $start = shift ; # get_date($cal,'secj2000',-0.125);
    my $end   = shift || $start; # get_date($cal,'secj2000',1.125);
    my $hh = 3.5;
    $start = ( $self->t1 ? $self->t1 : $start) - $hh*3600; # minus 6.5 hours
    $end   = ( $self->t2 ? $self->t2 : $end  ) + $hh*3600; # plus 6.5 hours
    
    my @fnames_z = $self->products;
    my ($sum_file)   =  map { s/\.Z$//; "../$_" }  grep { /sum/ }  @fnames_z; 
    my @bad_sats = map{ sprintf("gps%02d.eci",$_) } $self->igs_bad_sats($sum_file);
    if ( @bad_sats ) {
       unlink @bad_sats;
       $self->print_log( "removing bad satellites:|@bad_sats|\n");
    }

    #my $start   = shift;        # GPS sec past J2000 take all records >=
    #my $end     = shift;        # GPS sec past J2000 take all records <=

    my $tmp_eci = "eci$$";
    unlink $tmp_eci if(-e $tmp_eci);

    #my $st_utc = `gps2utc $start`; chomp $st_utc;
    #my $end_utc = `gps2utc $end`; chomp $end_utc;
    my $st_utc =  gps2utc( $start);
    my $end_utc = gps2utc($end);
    
    foreach my $g ( glob "gps*.eci" ){
        open( T, ">$tmp_eci");
        open( G, "<$g" ) or die "Unable to open $g";
        while(<G>){
            my @l = split(' ');
            # fix GPS32 for now
            #next if $l[0] == 32 || $l[0] eq '00';
            next if $l[0] eq '00';
            my $t = cal2sec(@l[1..6]);
            print T $_ if( $t >= $st_utc && $t <= $end_utc );
        }
        close T;
        close G;
        rename $tmp_eci, $g or die "Unable to rename $tmp_eci to $g ";
    }
}



sub filter_eci_files_ {
    #my $cal = get_date( shift, 'cal');
    #my $start =  get_date($cal,'secj2000',-0.125);
    #my $end   =  get_date($cal,'secj2000',1.125);
    my $start = shift ; # get_date($cal,'secj2000',-0.125);
    my $end   = shift || $start; # get_date($cal,'secj2000',1.125);
    
    #my $start   = shift;        # GPS sec past J2000 take all records >=
    #my $end     = shift;        # GPS sec past J2000 take all records <=

    my $tmp_eci = "eci$$";
    unlink $tmp_eci if(-e $tmp_eci);
    


    #my $st_utc = `gps2utc $start`; chomp $st_utc;
    #my $end_utc = `gps2utc $end`; chomp $end_utc;
    my $st_utc =  gps2utc( $start);
    my $end_utc = gps2utc($end);
    
    foreach my $g ( glob "gps*.eci" ){
        open( T, ">$tmp_eci");
        open( G, "<$g" ) or die "Unable to open $g";
        while(<G>){
            my @l = split(' ');
            # fix GPS32 for now
            #next if $l[0] == 32 || $l[0] eq '00';
            next if $l[0] eq '00';
            my $t = cal2sec(@l[1..6]);
            print T $_ if( $t >= $st_utc && $t <= $end_utc );
        }
        close T;
        close G;
        rename $tmp_eci, $g or die "Unable to rename $tmp_eci to $g ";
    }
}
sub clk2tdp_jpl {
    my $start = shift; 
    my $end   = shift; 
    my $clk_files = shift;
    my $tdp  = shift || 'TDPfile';
    my $sl = 299792.458;
    my $pg = GPS::PRN_GPS->new();
    open(my $th , "> $tdp.$$") or croak "Couldn't open $tdp $!\n";    
    foreach my $clk ( @$clk_files ) { 
      $clk =~ s/\.Z//;
      open(my $ch , "< $clk") or croak "Couldn't open clock file $clk $!\n";
      while ( <$ch> ) {
        next if ! /^AS G(\d\d)/ ;
        my $prn = $1;
        $prn =~ s/^0//;
        my @l = split ' ',$_;
        my $clk_bias = $l[9]*$sl;
        my $sig      = $l[10]*$sl;
        my $secd = cal2sec(@l[2..7]);
        next unless( $secd >= $start && $secd <= $end );
        my $gps = $pg->gps( prn =>$prn , secj2000 => $secd);
        my $name = sprintf "GPS%02d",$gps;
        printf $th  "%15.4f %17.10E  %14.9f    %10.3E SAT BIAS%s\n",
                    $secd,0,$clk_bias,$sig,$name;
     }
     
     close($ch);
   }
   
}

sub clk2tdp {
    my $self = shift;
    my $start = shift; 
    my $end   = shift; 
    my $clk_files = shift;
    my $tdp  = shift || 'TDPfile';
    my %TDP = ();
    my $rate = $self->rate;
    my $t1 = $self->t1 ? $self->t1 : get_date($self->date,'secj2000');
    my $t2 = $self->t2 ? $self->t2 : get_date($self->date,'secj2000',1);
    foreach my $clk ( @$clk_files ) { 
      $clk =~ s/\.Z//;
    #  print STDERR "working on $clk\n";
      print `sat_clk -i $clk -dt $rate -o $tdp -ti $t1 -tf $t2 -int `;
    }
}

sub clk2tdp_ {
    my $start = shift; 
    my $end   = shift; 
    my $clk_files = shift;
    my $tdp  = shift || 'TDPfile';
    my $sl = 299792.458;
    my $pg = GPS::PRN_GPS->new();
    my %TDP = ();
    open(my $th , "> $tdp.$$") or croak "Couldn't open $tdp $!\n";    
    foreach my $clk ( @$clk_files ) { 
      $clk =~ s/\.Z//;
      open(my $ch , "< $clk") or croak "Couldn't open clock file $clk $!\n";
      while ( <$ch> ) {
        next if ! /^AS G(\d\d)/ ;
        my $prn = $1;
        $prn =~ s/^0//;
        my @l = split ' ',$_;
        my $clk_bias = $l[9]*$sl;
        #my $sig      = $l[10]*$sl;
        my $sig      = ! $l[10] ? 2.0e-5 : $l[10]*$sl;
        my $secd = cal2sec(@l[2..7]);
        #my $secd = `cal2sec @l[2..7] 2> /dev/null`;
        next unless( $secd >= $start && $secd <= $end );
        my $gps = $pg->gps( prn =>$prn , secj2000 => $secd);
        next if $gps eq '00';
        my $name = sprintf "GPS%02d",$gps;
        push @{$TDP{$name}}, sprintf( "%15.4f %17.10E  %14.9f    %10.3E SAT BIAS%s\n", $secd,0,$clk_bias,$sig,$name);
     }
     close($ch);
   }
   my %seen;
   foreach my $s ( sort keys %TDP ) {
        print $th ( grep { ! $seen{$_}++ } sort @{$TDP{$s}} );
   } 
}
sub get_ba_dates{
    my $jd = get_date(shift,'jd');
    my ($by,$bm,$bd) = jd2cal($jd - 10);
    my ($ay,$am,$ad) = jd2cal($jd + 10);
    my $bmm = m2month($bm); 
    my $amm = m2month($am); 
    #my $b = sprintf("%02d-%3s-%4d",$bd,$bmm,$by);
    #my $a = sprintf("%02d-%3s-%4d",$ad,$amm,$ay);
    my $b = sprintf("%02d-%3s-%4d",$bd,$bmm,$by);
    my $a = sprintf("%02d-%3s-%4d",$ad,$amm,$ay);
    return ( $b,$a)
    #return ( jd2cal($jd-10) , jd2cal($jd +10 ) )
}



1;

