package GPS::Rinex;
use Fortran::F90Format;
use Data::Dumper;
use GOA::Util;
use Carp;

my  %Header = ( 
 'RINEX VERSION / TYPE' => { fmt =>'F9.2,11X,A1,19X,A1,19X,A20',
                             fields => [ qw( rnx_version rnx_type hlabel )  ],
                            },
 'PGM / RUN BY / DATE ' => { fmt => 'A20,A20,A20,A20',
                             fields => [ qw( progname run_by date hlabel ) ],
                            },
 'COMMENT             ' => { fmt => 'A60,A20',
                             fields => [ qw( comment hlabel  ) ],
                           },
 'MARKER NAME         ' => { fmt => 'A60,A20',
                             fields => [qw( marker_name hlabel )],
                           },
 'MARKER NUMBER       ' => { fmt => 'A20,A20', 
                             fields => [ qw( marker_num hlabel ) ] 
						   },
 'OBSERVER / AGENCY   ' => { fmt => 'A20,A40,A20', 
                             fields => [ qw( observer  agency hlabel )] 
						   },
 'REC # / TYPE / VERS ' => { fmt => '3A20,A20', 
                             fields => [ qw( rcvr_num rcvr_type rcvr_version hlabel) ],
						   },
 'ANT # / TYPE        ' => { fmt => '2A20,A20', 
                             fields => [ qw(ant_num ant_type hlabel)],
						   },
 'APPROX POSITION XYZ ' => { fmt => '3F14.4,A20',
                             fields => [qw( appr_pos_x appr_pos_y appr_pos_z hlabel )],
                            },
 'ANTENNA: DELTA H/E/N' => { fmt => '3F14.4,A20', 
                             fields => [qw( ant_height ant_east ant_north  hlabel)],
							},
 'WAVELENGTH FACT L1/2' => { fmt => '2I6,I6,7(3X,A1,I2),A20', 
                             fields => [qw( wf1 wf2 z  )],
							},
 '# / TYPES OF OBSERV ' => { fmt => 'I6,9(4X,A2),A20', 
                             fmt_cont => '6X,9(4X,A2),A20', 
                             fields => [qw(num_obs 9*obs  hlabel)], #to be redefined at  runtime
							},
 'INTERVAL            ' => { fmt => 'F10.3,50X,A20', 
                             fields => [qw(interval hlabel)],
							},
 'TIME OF FIRST OBS   ' => { fmt => '5I6,F13.7,5X,A3,9X,A20', 
                             fields => [ qw(syyyy smonth sday shour smin ssec stime_sys hlabel )],
							},         #optional
 'TIME OF LAST OBS    ' => { fmt => '5I6,F13.7,5X,A3,9X,A20', 
                             fields => [ qw(eyyyy emonth eday ehour emin esec etime_sys hlabel )],
							},         #optional
 'RCV CLOCK OFFS APPL ' => 'I6,54X,A20',                    
 'LEAP SECONDS        ' => 'I6,54X,A20',     
 '# OF SATELLITES     ' => 'I6,54X,A20',
# 'PRN / # OF OBS      ' => '3X,A1,I2,9I6,A20',
# 'PRN_NEXT_LINE'        => '6X,9I6',
 'END OF HEADER       ' => { fmt => '60X,A20', fields=> [qw( blank hlabel )] }
                   );
my %Observ = ( epoch => { fmt  => '1X,I2.2,4(1X,I2),F11.7,2X,I1,I3,12(A1,I2),F12.9',
                          fmt_cont => '32X,12(A1,I2)',
						  fields => [qw( year month day hour min sec
						                 epoch_flag 
										 num_sat
										 prn * |num_sat > 12 ? num_sat-12 : num_sat|
									  )], # to redefined at runtime
                         },
			   obs => { fmt => '5(F14.3,I1,I1)' ,
			            fields => [ qw( observations lli signal_strength ) ], #to be redefined at run time
			           },

);

our $C =  299792458;
our $F1 = 1575.42;
our $F2 = 1227.60;
our $Fw = $F1 - $F2;
our $L1 = $C/($F1*1e6);
our $L2 = $C/($F2*1e6);
our $Lw = $C/($Fw*1e6);
our $ff1_ff2 = ($F1*$F1 - $F2*$F2);
our $f1mf2 = ($F1 - $F2);
our $f1pf2 = ($F1 + $F2);
our $ff1 = $F1*$F1/$ff1_ff2; # 2.54572778016316
our $ff2 = $F2*$F2/$ff1_ff2; # 1.54572778016316

sub new {
  my $class = shift;
  my $self  = {};
  bless $self,$class;
  $self->init();
  return $self 
}

sub fmt {
  my $self = shift;
  my $f =  shift;
  return $self->{fmt}{$f}
}
sub fmt_cont {
  my $self = shift;
  my $f =  shift;
  return $self->{fmt_cont}{$f}
}

sub init {
  my $self = shift;
  foreach my $k ( keys %Header ) {
     $self->{fmt}{$k} = Fortran::F90Format->new( fmt =>	 $Header{$k}{fmt} );   
	 if ( exists $Header{$k}{fmt_cont} ) {
       $self->{fmt_cont}{$k} = Fortran::F90Format->new( fmt => $Header{$k}{fmt_cont} );   
	 }
  }
  foreach my $k ( keys %Observ ) {
     $self->{fmt}{$k} = Fortran::F90Format->new( fmt =>	 $Observ{$k}{fmt} );   
	 if ( exists $Observ{$k}{fmt_cont} ) {
       $self->{fmt_cont}{$k} = Fortran::F90Format->new( fmt => $Observ{$k}{fmt_cont} );   
	 }
  }

  return $self
}

sub load {
  my $self = shift;
  my $rnx  = shift;
  
  open(my $fh,"< $rnx") or croak "Can't open $rnx $!\n";

  my (@values,$hlabel,$hvalue,$no,$no2);
  my @obs=();
  my %ot=();
  my ($num_obs,$num_obs_cont);
  while ( <$fh> ) {
   chomp;
   $_ = sprintf("%-80s",$_);
   ($hvalue,$hlabel)=unpack("a60a20",$_); 
   #print "$hvalue |$hlabel|",length($hlabel),"\n";
   if ( exists $Header{$hlabel} ) { 
      
	  @values = $self->fmt($hlabel)->read($_); 
      
	  if ( $hlabel =~ /TYPES OF OBSERV/ ) {
	     $num_obs = shift @values;
		 if ( $num_obs <= 9 ) {
		    @obs = @values[0..$num_obs-1];
		 }
		 else {
		    $num_obs_cont = $num_obs-9;
		    @obs = @values;
		    $_ = readline($fh);
            $_ = sprintf("%-80s",$_);
	        @values = $self->fmt_cont($hlabel)->read($_); 
			push @obs,@values[0..$num_obs_cont-1];
		 }
		 $self->{num_obs}=$num_obs;
		 @{$self->{obs}} = @obs;
         @ot{@obs}=@obs;
         #if ( exists $ot{P1} ) {
         #  $ot{P1P2}= exists $ot{P2} ? 1 : 0;
         #}
         # elsif ( exists $ot{C1} ) {
         #  $ot{C1P2}= exists $ot{P2} ? 1 : 0;
         #}

		# print "$n_obs ", join(',',@obs), "\n";
	  }
      #print $self->fmt($hlabel)->write(@values); #,"$hlabel\n";
	}
	else {
	  carp "$hlabel is not in Header definition\n";
	}
    last if /END OF HEADER/; 
	if ( $. > 300 ) {
	  croak "rinex header > 300 lines !\n";
	}
  }
  
  my ($yy,$y,$m,$d,$hh,$mm,$ss,$eflag,$num_sat,$num_rec,$clk_offset);
  
  my ($ttag,$stype,$snum,$sat,@sats,@tmp,$i,$j,$sst,$lli,$o_data,$ho);

  $num_obs = $self->{num_obs}; 

  while ( <$fh> ) {
    chomp;
	# read epoch line
    $_ = sprintf("%-80s",$_);
	@tmp = $self->fmt('epoch')->read($_); 
    ($yy,$m,$d,$hh,$mm,$ss,$eflag,$num_sat)= @tmp[0..7];
    
	$eflag = trim($eflag) + 0;
     
    if ( $eflag > 1 ) {
	   #skip post header records
	   for ( 1 .. $num_sat ) {
	      readline($fh);
	   }
	   next;
	}

    @values = @tmp[8..$#tmp-1];
	$clk_offset=$tmp[-1];
	$y = $yy < 50 ? 2000+$yy : 1900+$yy;
	$ttag = sprintf("%4d-%02d-%02dT%02d:%02d:%-11.7f",$y,$m,$d,$hh,$mm,$ss);


	#read continuation lines in case of more than 12 sats
    $no=$num_sat > 12 ? $num_sat - 12 : 0;
	while ( $no ) {
	  $_ = readline($fh);
      chomp;
      $_ = sprintf("%-80s",$_);
	  @tmp = $self->fmt_cont('epoch')->read($_); 
	  if ( $no <= 12 ) {
	     @tmp = @tmp[0..(2*$no)-1];
		 $no=0;
	  } 
	  else {
	    $no = $no-12; 
	  }
	  push @values,@tmp;
	}
	
	#assign sats
	@sats = ();
	for my $i ( 0 .. $num_sat-1 ) {
	   $stype= shift @values;
	   $snum = shift @values;
	   $sat = sprintf('%1s%02d',trim($stype),trim($snum));
	   push @sats,$sat;
	}
    
	# read observations
	foreach my $s ( @sats ) {
      $_ = readline($fh);
      chomp;
      $_ = sprintf("%-80s",$_);
  	  @values = $self->fmt('obs')->read($_);
	  $no = $num_obs > 5 ? $num_obs - 5 : 0 ;
	  while ( $no ) {
	     $_ = readline($fh);
         chomp;
         $_ = sprintf("%-80s",$_);
	     @tmp = $self->fmt('obs')->read($_); 
		 if ( $no <= 5 ) {
		    @tmp=@tmp[0..(3*$no)-1];
			$no=0;
	     }
	     else {
		   $no = $no - 5;
	     }
		 push @values,@tmp;
      }#while
    
	  #assign obs
      $h = {}; 
	  $i = 0;
	  for my $j ( 0 .. $num_obs-1 ) {
         $o_name = $obs[$j];
	     $o_data = $values[$i];
		 $lli = $values[$i+1];
		 $sst = $values[$i+2];
         $o_data = $o_name eq 'L1' ? $o_data*$L1 :
                   $o_name eq 'L2' ? $o_data*$L2 :
                                     $o_data;
	     if ( $o_data ) {
   	       $h->{$o_name}={ val => $o_data, # observation: 
                                            # Phase: units of cycles of carrier 
                                            # Code: meters
                                                                
		                    lli => $lli ,   # loss of lock indicator 0-7:
                                            # 0 or blank: OK or not known
                                            # Bit 0 set: lost of lock between previous and current observation
                                            #            cycle slip possible
                                            # Bit 1 set : Opposite wavelength factor to the one defined
                                            #             for the satellite by a previous WAVELENGTH FACT L1/2 line.
                                            # Bit 2 set : Observation under Antispoofing ( may suffer from increased noise)
                                                
							sst => $sst };  # signal strength 1-9:
                                            # 1: minimum possible signal strength
                                            # 5: treshold for good S/N ratio
                                            # 9: maximum possible signal strength
                                            # 0 or blank: don't know, don't care
		 }
		 else {
		   #carp "Incomplete record at $ttag, trying to read $obs[$j] j=$j for sat
		   #$s , |@values |\n";
		 }
		 $i+=3;
	  }
      $l1=$h->{L1}{val};
      $l2=$h->{L2}{val};
      $c1= exists $h->{C1} ? $h->{C1}{val} : 0;
      $p1= exists $h->{P1} ? $h->{P1}{val} : 0;
      $p2= exists $h->{P2} ? $h->{P2}{val} : 0;
      $p21 = $p2 && $p1 ? $p2 - $p1 : 0;
      $h->{LC}{val} = $ff1*$l1 - $ff2*$l2 ;
      $h->{LW}{val} = ($F1*$l1 - $F2*$l2)/$f1mf2;
      if ( $p1 && $p2 ) {
        $h->{PC}{val} = $ff1*$p1 - $ff2*$p2;
        $h->{PW}{val} = $h->{LW}{val}  - ($F1*$p1 - $F2*$p2)/$f1pf2;
        $h->{L1P1}{val}= $l1 - $p1 + 2*$p21*$ff2;
        $h->{L2P2}{val}= $l2 - $p2 + 2*$p21*$ff1;
      }
      if ( $c1 && $p2 ) {
        $h->{PC}{val} = $ff1*$c1 - $ff2*$p2;
      }
      if ( $p1 && ! $p2 ) {
        $h->{L1_P1}{val}= $l1 - $p1;
        $h->{L2_P1}{val}= $l2 - $p1;
      }
      if ( ! $p1 &&  $p2 ) {
        $h->{L1_P2}{val}= $l1 - $p2;
        $h->{L2_P2}{val}= $l2 - $p2;
      }
      if ( $c1 && ! $p1  ) {
        $h->{L1_C1}{val}= $l1 - $c1;
        $h->{L2_C1}{val}= $l2 - $c1;
      }
      $self->{data}{$ttag}{$s}=$h;
	} #sats
    
  }

  
}

1;


